/*
 * Network block device - make block devices work over TCP
 *
 * Note that you can not swap over this thing, yet. Seems to work but
 * deadlocks sometimes - you can not swap over TCP in general.
 * 
 * Copyright 1997 Pavel Machek <pavel@atrey.karlin.mff.cuni.cz>
 * 
 * (part of code stolen from loop.c)
 *
 * 97-3-25 compiled 0-th version, not yet tested it 
 *   (it did not work, BTW) (later that day) HEY! it works!
 *   (bit later) hmm, not that much... 2:00am next day:
 *   yes, it works, but it gives something like 50kB/sec
 * 97-4-01 complete rewrite to make it possible for many requests at 
 *   once to be processed
 * 97-4-11 Making protocol independent of endianity etc.
 * 97-9-13 Cosmetic changes
 * 98-5-13 Attempt to make 64-bit-clean on 64-bit machines
 * 99-1-11 Attempt to make 64-bit-clean on 32-bit machines <ankry@mif.pg.gda.pl>
 * 2000-01-03 A bunch of changes to the protocol to add support for
 *            gzipped data and versioning. the gzip stuff is turned off
 *            by default as it doesn't win on high-bandwidth links 
 *            <asun@cobalt.com>.
 *
 * 2000-03-23 even more changes. turned nbd client into a kernel thread
 *            and changed interface for setup/removal. <asun@cobalt.com>
 */

#define PARANOIA
#include <linux/config.h>
#include <linux/major.h>

#include <linux/module.h>

#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/fs.h>
#include <linux/malloc.h>
#include <linux/stat.h>
#include <linux/errno.h>
#include <linux/file.h>
#include <linux/ioctl.h>

#if defined(CONFIG_COBALT_GEN_III) || defined(CONFIG_COBALT_GEN_V)
#define NBD_REBOOT_TEARDOWN 1
#endif

#ifdef NBD_REBOOT_TEARDOWN
#include <linux/notifier.h>
#include <linux/reboot.h>
#endif

#include <asm/segment.h>
#include <asm/uaccess.h>
#include <asm/types.h>

#define MAJOR_NR NBD_MAJOR
#include <linux/nbd.h>

#define USE_GZIP 0

#if USE_GZIP
/* mondo yucko. zlib.c should be in the toplevel lib directory. */
#include "../net/zlib.c"
#undef next
#endif

#define LO_MAGIC 0x68797548

static int nbd_blksizes[MAX_NBD];
static int nbd_blksize_bits[MAX_NBD];
static int nbd_sizes[MAX_NBD];
static u64 nbd_bytesizes[MAX_NBD];

static struct nbd_device nbd_dev[MAX_NBD];
static spinlock_t nbd_lock = SPIN_LOCK_UNLOCKED;

#if 0
#define DEBUG( s ) printk( s ) 
#else
#define DEBUG( s )
#endif

#ifdef PARANOIA
static int requests_in;
static int requests_out;
#endif

#ifndef MIN
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#endif

#define NBD_TIMEOUT          (HZ*2)
#define NBD_MAX_PEND         (NR_REQUEST >> 1)
#define NBD_BLOCK_NONE        0  /* nonblocking mode */
#define NBD_BLOCK_ALL         1  /* block via scheduling */

#define NOISE_RATE  (2*HZ)
static inline int nbd_noisy(void)
{
  static unsigned long lasttime = 0;
  static unsigned int messages = 0;

  if ((long) (jiffies - lasttime) > NOISE_RATE) {
    lasttime = jiffies;
    if (messages) {
      printk("NBD: skipped %u kernel messages.\n", messages);
      messages = 0;
    }
    return 1;
  }

  ++messages;
  return 0;
}

#if USE_GZIP
/* zlib.c should have a default kernel zalloc/zfree */
static void *zalloc(void *arg, unsigned int items, unsigned int size)
{
	return kmalloc(items*size, GFP_KERNEL);
}

static void zfree(void *arg, void *ptr)
{
	kfree(ptr);
}
#endif

static int nbd_open(struct inode *inode, struct file *file)
{
	int dev;
	struct nbd_device *nbdev;

	if (!inode)
		return -EINVAL;
	dev = MINOR(inode->i_rdev);
	if (dev >= MAX_NBD)
		return -ENODEV;

	nbdev = &nbd_dev[dev];
	nbd_dev[dev].refcnt++;
	if (!test_bit(NBD_INITIALISED, &nbdev->flags)) {
		nbdev->queue_sem = MUTEX;
		nbdev->read_sem = MUTEX;
		nbdev->write_sem = MUTEX;
		set_bit(NBD_INITIALISED, &nbdev->flags);
	}
	MOD_INC_USE_COUNT;
	return 0;
}

/*
 * Send or receive packet. to coalesce packets, we stick a header [+
 * data] into an iovec. if you just want to send a header, it's
 * better to use the second buffer.
 *
 * CAVEATS: this assumes that the appropriate thread semaphore is locked.
 *          if a partial read/write gets performed, we don't release the
 *          semaphore.
 */
static int nbd_xmit(const int send, struct nbd_device *lo, 
		    void *header, unsigned int hsize,
		    void *buf, unsigned int size, unsigned int blockval)
{
	mm_segment_t oldfs;
	int cc, result = 0;
	struct msghdr msg;
	struct iovec iov[2];
	struct semaphore *sem = (send) ? &lo->write_sem : &lo->read_sem;
	struct wait_queue wait = {current, NULL};
	unsigned long timeout;
	unsigned int total;

	oldfs = get_fs();
	set_fs(get_ds());

	
	/* NOTE: we must be in non-blocking mode and have 
	 * signals turned off or else we expose kernel space to user
	 * access. */
	down(sem);
	if (lo->timeout)
		timeout = jiffies + lo->timeout;
	total = hsize + size;
	while (1) {
		if (test_bit(NBD_RUNNING, &lo->flags) == 0) {
			result = -EINTR;
			break;
		}

		memset(&msg, 0, sizeof(msg));
		msg.msg_flags = MSG_NOSIGNAL | MSG_DONTWAIT;
		msg.msg_iov = iov;
		if (hsize) {
			iov[0].iov_base = header;
			iov[0].iov_len = hsize; 
			iov[1].iov_base = buf;
			iov[1].iov_len = size;
			msg.msg_iovlen = 2;
		} else {
			iov[0].iov_base = buf;
			iov[0].iov_len = size;
			msg.msg_iovlen = 1;
		}

		cc = (send) ? sock_sendmsg(lo->sock, &msg, size + hsize) : 
			sock_recvmsg(lo->sock, &msg, size + hsize, 
				     MSG_DONTWAIT);
		
		/* this is a little hokey. if we don't get/send any
		 * data, we treat it like an EAGAIN. that way, we
		 * always turn things into timeouts. */
		if ((cc == -EAGAIN) || !cc) {
			long tout;

		        if (blockval == NBD_BLOCK_NONE) {
				/* if it's not really an EAGAIN, just
				 * try again */
				if (!cc) 
					continue;
				result = -EAGAIN;
				break;
			}			

			if (!lo->timeout || (tout = timeout - jiffies) > 0) {
				add_wait_queue(&lo->sock->wait, &wait);
				current->state = TASK_INTERRUPTIBLE;
				lo->sock->flags |= SO_WAITDATA;
				if (!result) 
					up(sem);
				set_fs(oldfs);
				if (lo->timeout) 
					schedule_timeout(tout);
				else
					schedule();
				oldfs = get_fs();
				set_fs(get_ds());
				if (!result)
					down(sem);
				current->state = TASK_RUNNING;
				if (!lo->sock) {
					result = (cc) ? -EINTR : -ENODATA;
					break;
				}
				lo->sock->flags &= ~SO_WAITDATA;
				remove_wait_queue(&lo->sock->wait, &wait);
				continue;
			}
			cc = -ETIMEDOUT;
		}

		if (cc < 0) {
#ifdef PARANOIA
		       if (nbd_noisy()) {
			  printk(KERN_ERR "NBD: %s - sock=%p at header=%p, hsize=%d, buf=%p, size=%d (xferred=%d) (err=%d).\n",
			       send ? "send" : "receive", lo->sock, 
			       header, hsize, buf, size, result, cc);
			}
#endif
		       result = cc;
		       break;
		}

		result += cc;
		if (total == result)
			break; /* done */

		/* fix up our buffers for another pass */
		blockval = NBD_BLOCK_ALL;
		if (hsize > cc) {
			header += cc;
			hsize -= cc;
		} else {
			if (hsize) {
				cc -= hsize;
				hsize = 0;
			}
			buf += cc;
			size -= cc;
		}
	} 
	up(sem);
	set_fs(oldfs);
	return result;
}

/* this just clears out writes to the next read or match */
static inline void nbd_clear_write_queue(struct nbd_device *lo,
					 struct request *match)
{
        struct request *req;

	while ((req = lo->tail)) {
		if (req == match)
			break;
		if (req->cmd != WRITE) 
			break;

		nbd_end_request(req);
		if (lo->tail == lo->head)
			lo->head = NULL;
		lo->tail = lo->tail->next;
	}
}


void nbd_clear_que(struct nbd_device *lo)
{
	struct request *req;

	down(&lo->queue_sem);
	while (1) {
		req = lo->tail;
		if (!req) 
			goto clear_done;
#ifdef PARANOIA
		if (lo != &nbd_dev[MINOR(req->rq_dev)]) {
			printk(KERN_ALERT "NBD: request corrupted when clearing!\n");
			goto clear_done;
		}

		if (lo->magic != LO_MAGIC) {
			printk(KERN_ERR "NBD: nbd_dev[] corrupted: Not enough magic when clearing!\n");
			goto clear_done;
		}
#endif
		req->errors++;
		nbd_end_request(req);
		if (lo->tail == lo->head)
			lo->head = NULL;
		lo->tail = lo->tail->next;
	}

clear_done:
	up(&lo->queue_sem);
}


#define FAIL( s ) { if (nbd_noisy()) printk( KERN_ERR "NBD: " s "(result %d)\n", result ); goto error_out; }
static void nbd_send_req(struct nbd_device *lo, const u32 cmd,
			 const u64 from, void *buffer, const u32 len, 
			 struct request *req)
{
	int result;
	struct nbd_request request;

	DEBUG("NBD: sending control, ");
	request.magic = htonl(NBD_REQUEST_MAGIC);
	request.type = htonl(cmd);
	request.from = cpu_to_be64(from);
	request.len = htonl(len);
	memcpy(request.handle, &req, sizeof(req));

	if (len && !(cmd == NBD_OP_READ)) {
		DEBUG("data, ");
		result = nbd_xmit(1, lo, &request, sizeof(request),
				  buffer, len, NBD_BLOCK_ALL);
	} else 
		result = nbd_xmit(1, lo, NULL, 0, &request, sizeof(request),
				  NBD_BLOCK_ALL);
	if (result <= 0)
		FAIL("Send Control/Data failed.");
	DEBUG("done.\n");
	return;

 error_out:
	req->errors++;
}


/* we got a request. right now, it can only be a tickle. 
 * we need to prevent ourselves from getting interrupted here. */
static int nbd_handle_request(struct nbd_device *lo, void *handle)
{
	struct nbd_request *request = handle;

	if (!request)
		return 0;

	/* we currently don't handle request lengths */
	if (request->len) 
		return -1;

	/* we only deal with tickles for now */
	if (request->type == htonl(NBD_OP_TICKLE)) {
	  struct nbd_reply reply;

	  memset(&reply, 0, sizeof(reply));
	  reply.magic = htonl(NBD_REPLY_MAGIC);
	  reply.type = request->type;
	  if (nbd_xmit(1, lo, NULL, 0, &reply, lo->replysize, 
		       NBD_BLOCK_ALL) < 0)
	    DEBUG("tickle reply failed\n");
	  else 
	    DEBUG("sent tickly reply\n");
	}
	return 0;
}


#define HARDFAIL( s ) { printk( KERN_ERR "NBD: " s "(result %d)\n", result ); lo->harderror = result; result = -EINTR; goto error_out; }
static int nbd_process_stat(struct nbd_device *lo, struct nbd_reply *reply)
{
	int result;
	struct request *xreq, *req;
#if USE_GZIP
	z_streamp z;
#endif

	DEBUG("nbd_process, ");
	memcpy(&xreq, reply->handle, sizeof(xreq));

	/* skip to the needed write request, if any. */
	down (&lo->queue_sem);
	nbd_clear_write_queue(lo, xreq);

	/* paranoia. this can only happen if we over-flush our write queue. 
	 * it's only a problem in case of an error. */
	if ((req = lo->tail) == NULL) {
		up(&lo->queue_sem);
		return 0;
	}

	if (lo->tail == lo->head)
		lo->head = NULL;
	lo->tail = lo->tail->next;
	up (&lo->queue_sem);

	result = lo->info.nbd_server_version ? ntohl(reply->type) : req->cmd;
	if (xreq != req) {
		HARDFAIL("Unexpected handle received.\n");
	}

	DEBUG("ok, ");

	if (ntohl(reply->error)) {
		result = 0;
		FAIL("Other side returned error.");
	}

	if (NBD_OP_MASK(result) == NBD_OP_READ) {
#if USE_GZIP
		if ((result & NBD_FLAG_GZIP) == 0)
#endif
			result = nbd_xmit(0, lo, NULL, 0, req->buffer, 
					  req->current_nr_sectors << 9, 
					  NBD_BLOCK_ALL);
		if (result <= 0) 
			HARDFAIL("Recv data failed.");

#if USE_GZIP
		} else {
			DEBUG("dataz, ");
		if (!lo->zbuf) 
			HARDFAIL("Unable to decompress data!");

		z = lo->istreamp;
		if ((z->avail_in = ntohl(reply->len)) > lo->zlen)
			HARDFAIL("Got an impossible reply size!");

		result = nbd_xmit(0, lo, NULL, 0, lo->zbuf, z->avail_in,
				  NBD_BLOCK_ALL);
		if (result <= 0)
			HARDFAIL("Recv data failed.");

		z->next_in = lo->zbuf;
		z->next_out = req->buffer;
		z->avail_out = req->current_nr_sectors << 9;

		if ((result = inflate(z, Z_FINISH)) != Z_STREAM_END) 
			HARDFAIL("Recv uncompress failed.");
		inflateReset(z);
		break;
#endif
	}
	result = 0;
	DEBUG("done.\n");

process_done:
	nbd_end_request(req);
	return result;

/* Can we get here? Yes, if other side returns error */
error_out:
	req->errors++;
	goto process_done;
}


/* this will just return if there's nothing to do */
static int nbd_handle_input(struct nbd_device *lo)
{
	struct nbd_reply reply;
	int result, err;

	/* return if we can't handle it */
	if (!(lo && lo->file))
		return -ENXIO;

	DEBUG("+");
input_start:
	reply.magic = 0;
	err = nbd_xmit(0, lo, NULL, 0, &reply, lo->replysize,
		       NBD_BLOCK_ALL);
	
	/* bail on errors */
	if (err <= 0)  
		goto input_done;

	err = -1;
	/* check some magic */
	result = ntohl(reply.magic);
	switch (result) {
	case NBD_REPLY_MAGIC:
		break; /* always accept these */

	case NBD_REQUEST_MAGIC:
		if (lo->info.nbd_server_version) {
			DEBUG("got request\n");

			/* flush previous write requests */
			down(&lo->queue_sem);
			nbd_clear_write_queue(lo, NULL);
			up(&lo->queue_sem);
			if (nbd_handle_request(lo, &reply) < 0) {
				DEBUG("Can't handle request!\n");
				goto input_done;
			}
			/* this is just here to avoid a little
			 * function call overhead */
			goto input_start; 
		}
		/* fall through */

	default:
		DEBUG("Not enough magic.\n");
		goto input_done;
	}

	err = nbd_process_stat(lo, &reply);

input_done:
	return err;
}


/* because we fake up our own request structure, 
 * we have to handle both the send and reply ourselves. */
static void nbd_send_config(struct nbd_device *lo)
{
	struct request req;
	u16 value;
	u8 buf[4]; /* config info */

	memset(buf, 0, sizeof(buf));

	/* fill up version and flags fields */
	value = htons(NBD_VERSION);
	memcpy(buf, &value, sizeof(value));

	/* tell the client that we support read compression. */
#if USE_GZIP
	if (lo->info.nbd_client_opts & NBD_FLAG_GZIP) {
	  value = htons(NBD_FLAG_GZIP);
	  memcpy(buf + sizeof(value), &value, sizeof(value));
	}
#endif

	DEBUG("NBD: sending config, ");
	req.cmd = NBD_OP_CONFIG;
	return nbd_send_req(lo, NBD_OP_CONFIG, 0, buf, sizeof(buf), &req);
}


/* initialize a few things */
static void nbd_initialize(struct nbd_device *lo, const int minor)
{
	/* if it's version 0, don't do anything */
	DEBUG("NBD: initialization\n");
	if (!lo->info.nbd_server_version) 
		return;

	/* first, set up our scratch buffer */
#if USE_GZIP
	if (lo->zbuf)
		kfree(lo->zbuf);
	lo->zlen = max_readahead[MAJOR_NR] ? max_readahead[MAJOR_NR][minor] :
	  MAX_READAHEAD;
	lo->zlen = MIN(lo->zlen, NBD_DATA_MAX);
	lo->zbuf = kmalloc(lo->zlen, GFP_KERNEL);
	if (!lo->zbuf) {
		printk(KERN_ALERT "NBD: Unable to allocate a scratch buffer for compression\n");
		goto pre_doit_err;
	}

	/* now, set up our input buffer. */
	lo->istreamp = kmalloc(sizeof(z_stream), GFP_KERNEL);
	if (lo->istreamp) {
		z_streamp z = lo->istreamp;

		/* set up memory allocation routines */
		memset(z, 0, sizeof(z_stream));
	 	z->zalloc = zalloc;
		z->zfree = zfree;
	
		if (inflateInit(z) != Z_OK) {
			kfree(lo->istreamp);
			lo->istreamp = NULL;
			printk(KERN_ALERT "NBD: Unable to initialize inflate buffer.\n");
			goto inflate_err;
		}
	} else {
		printk(KERN_ALERT "NBD: Unable to allocate space for compressed input.\n");
		goto inflate_err;
	}
	
	/* finally, set up our output buffer. if this fails, just turn
	 * off the compress flag and continue. */
	if (lo->info.nbd_server_opts & NBD_FLAG_GZIP) {
		lo->ostreamp = kmalloc(sizeof(z_stream), GFP_KERNEL);
		if (lo->ostreamp) {
			z_streamp z = lo->ostreamp;

			/* set up the memory allocation routines. */
			memset(z, 0, sizeof(z_stream));
			z->zalloc = zalloc;
			z->zfree = zfree;	
			if (deflateInit(z, Z_BEST_SPEED) == Z_OK) {
				return; /* okay */
			}
			kfree(lo->ostreamp);
			lo->ostreamp = NULL;
		}
		printk(KERN_ALERT "NBD: Unable to allocate space for compressed output\n");
		goto pre_doit_err;
	}
	goto done;

inflate_err:
	kfree(lo->zbuf);
	lo->zlen = 0;
	lo->zbuf = NULL;

pre_doit_err:
#endif
	lo->info.nbd_server_opts &= ~NBD_FLAG_GZIP;

#if USE_GZIP
done:
#endif
	nbd_send_config(lo);
}


/* de-initialize stuff. we run with interrupts off so that
 * we don't get trounced on multiple times. */
static void nbd_cleanup(struct nbd_device *lo)
{
	unsigned long flags;
	struct file *file;

	/* we should try to send a done packet. */

	spin_lock_irqsave(&nbd_lock, flags);
	if ((file = lo->file) == NULL) 
		goto cleanup_done;
	
	nbd_clear_que(lo); /* flush any requests */

#if USE_GZIP
	/* clean up compression bits */
	if (lo->istreamp) {
		inflateEnd(lo->istreamp);
		kfree(lo->istreamp);
		lo->istreamp = NULL;
	}
	
	if (lo->ostreamp) {
		deflateEnd(lo->ostreamp);
		kfree(lo->ostreamp);
		lo->ostreamp = NULL;
	}
	
	if (lo->zbuf) {
		kfree(lo->zbuf);
		lo->zlen = 0;
		lo->zbuf = NULL;
	}
#endif		

	/* now, reset everything */
	lo->sock = NULL;
	lo->file = NULL;
	lo->info.nbd_server_version = lo->info.nbd_server_opts = 0;
        lo->replysize = NBD_REPLY_SIZE_V0;
	set_bit(NBD_INITIALISED, &lo->flags);
	fput(file);

cleanup_done:
	spin_unlock_irqrestore(&nbd_lock, flags);
}

/*
 * now, we do the following: 
 * 1) queue the read and write requests separately.
 * 2) when a new request appears, mark anything up to that point as okay.
 * 3) request write acks periodically to prevent a stalled situation
 *    from occurring.
 */

#undef FAIL
#define FAIL( s ) { if (nbd_noisy()) printk( KERN_ERR "NBD, minor %d: " s "\n", dev ); goto error_out; }
static void do_nbd_request(void)
{
	struct nbd_device *lo;
	struct request *req;
	int dev, ack = 0;

	while (CURRENT) {
		req = CURRENT;
		dev = MINOR(req->rq_dev);
#ifdef PARANOIA
		if (dev >= MAX_NBD)
			FAIL("Minor too big.");		/* Probably can not happen */
#endif
		lo = &nbd_dev[dev];
		if (!lo->file)
			FAIL("Request when not-ready.");
		if ((req->cmd == WRITE) && 
		    (test_bit(NBD_READ_ONLY, &lo->flags)))
			FAIL("Write on read-only");
#ifdef PARANOIA
		if (lo->magic != LO_MAGIC)
			FAIL("nbd[] is not magical!");
		requests_in++;
#endif
		req->errors = 0;
		CURRENT = CURRENT->next;
		req->next = NULL;

		spin_unlock_irq(&io_reqeust_lock);
		down (&lo->queue_sem);

		/* set up an ack after every X write requests */
		if (lo->info.nbd_server_version && (req->cmd == WRITE) && 
		    (++lo->req_pending > NBD_MAX_PEND)) { 
			lo->req_pending = 0;
			ack = NBD_FLAG_ACK;
		}

		if (lo->head == NULL) {
			lo->head = req;
			lo->tail = req;
		} else {
			lo->head->next = req;
			lo->head = req;
		}
	
		/* if the other side supports compression,
		   try to compress the buffer on writes */
#if USE_GZIP
		if ((req->cmd == WRITE) &&
		    (lo->info.nbd_server_opts & NBD_OPTION_GZIP)) {
			z_streamp z = lo->ostreamp;
			int len = req->current_nr_sectors << 9;
			
			z->next_in = req->buffer;
			z->avail_in = len;
			z->next_out = lo->zbuf;
			z->avail_out = MIN(len, lo->zlen) - 1;

			/* we use z->avail_out to guarantee that a 
			 * successful deflation will be smaller than
			 * just sending the block. */
			if (deflate(z, Z_FINISH) == Z_STREAM_END) {
				nbd_send_req(lo, NBD_OP_WRITE | 
					     NBD_FLAG_GZIP | ack,
					     (u64) req->sector << 9,
					     lo->zbuf, z->total_out, req);
				deflateReset(z);
				goto writez_ok;
			}
			deflateReset(z);
		}
#endif

		/* this will schedule itself if necessary */
		nbd_send_req(lo, req->cmd | ack, (u64) req->sector << 9,
			     req->buffer, req->current_nr_sectors << 9,
			     req);

#if USE_GZIP
writez_ok:
#endif
		up(&lo->queue_sem);
		spin_lock_irq(&io_request_lock);
		continue;

	error_out: 
		req->errors++;
		spin_unlock_irq(&io_request_lock);
		nbd_end_request(req);
		spin_lock_irq(&io_request_lock);
		CURRENT = CURRENT->next;
	}
}

/* the client thread */
static int nbd_client_thread(void *data)
{
	struct nbd_device *lo = data;
	int err;

	lock_kernel();
	exit_mm(current); /* we don't need mmaps */
	current->session = 1;
	current->pgrp = 1;
	strcpy(current->comm, "nbd-clientd");
	sigfillset(&current->blocked);

	/* up our priority */
	current->priority = 40;
	unlock_kernel();

	/* loop here until it's over */
	while (1) {
		err = nbd_handle_input(lo);
		if (err < 0)
			break;
	}
	nbd_cleanup(lo);
	return 0;
}


static int nbd_thread_start(struct nbd_device *lo, int dev)
{
	int err;

	/* initialize thread here */
	set_bit(NBD_RUNNING, &lo->flags);
	nbd_initialize(lo, dev);
	err = kernel_thread(nbd_client_thread, lo, 0);
	if (err < 0) {
	        clear_bit(NBD_RUNNING, &lo->flags);
		return err;
	}		
	return 0;
}

/* tell the thread to stop running. this actually just
 * makes nbd_xmit return an EINTR to the thread. */
static int nbd_thread_delete(struct nbd_device *lo)
{
	struct request req;
	nbd_send_req(lo, NBD_OP_DONE, 0, NULL, 0, &req);
	clear_bit(NBD_RUNNING, &lo->flags);
	return 0;
}

static int nbd_set_info(struct nbd_device *lo, int dev, struct nbd_info *arg)
{
	struct nbd_info info;
	int temp;

	if (!arg)
		return -EINVAL;

	if (copy_from_user(&info, arg, sizeof(struct nbd_info)))
		return -EFAULT;

	/* set the block size */
	if (info.nbd_blksize) {
		if ((info.nbd_blksize & (info.nbd_blksize - 1)) ||
		    (info.nbd_blksize < 512) || (info.nbd_blksize > PAGE_SIZE))
			return -EINVAL;
		
		nbd_blksizes[dev] = info.nbd_blksize;
		temp = info.nbd_blksize >> 9;
		nbd_blksize_bits[dev] = 9;
		while (temp > 1) {
			nbd_blksize_bits[dev]++;
			temp >>= 1;
		}
		nbd_sizes[dev] = nbd_bytesizes[dev] >> nbd_blksize_bits[dev];
		nbd_bytesizes[dev] = nbd_sizes[dev] << nbd_blksize_bits[dev];
	}

	/* set the size in blocks */
	if (info.nbd_size_blocks) {
		nbd_sizes[dev] = info.nbd_size_blocks;
		nbd_bytesizes[dev] = 
			((u64) info.nbd_size_blocks) << nbd_blksize_bits[dev];
	}

	/* set reply size */
	lo->replysize = info.nbd_server_version ? NBD_REPLY_SIZE : 
		NBD_REPLY_SIZE_V0;

	/* set tickle */
	if (info.nbd_tickleval) 
		lo->timeout = (info.nbd_tickleval*HZ/1000)*2 + NBD_TIMEOUT;

	/* copy everything to info field */
	memcpy(&lo->info, &info, sizeof(info));

	/* client version */
	lo->info.nbd_client_version = NBD_VERSION;
	return 0;
}


static int nbd_get_info(struct nbd_device *lo, int dev, struct nbd_info *arg)
{
	if (!arg)
		return -EINVAL;

	return copy_to_user(arg, &lo->info, sizeof(lo->info)) ? -EFAULT : 0;
}

static int nbd_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg)
{
	struct nbd_device *lo;
	int dev, error;

	/* Anyone capable of this syscall can do *real bad* things */
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;
	if (!inode)
		return -EINVAL;
	dev = MINOR(inode->i_rdev);
	if (dev >= MAX_NBD)
		return -ENODEV;

	lo = &nbd_dev[dev];
	switch (cmd) {
	case NBD_CLEAR_SOCK:
	        if (!lo->file)
		        return -ENXIO;
		return nbd_thread_delete(lo);
	case NBD_SET_SOCK:
		if (lo->file)
			return -EBUSY;
		error = -EINVAL;
		file = fget(arg);
		if (file) {
			inode = file->f_dentry->d_inode;
			/* N.B. Should verify that it's a socket */
			lo->file = file;
			lo->sock = &inode->u.socket_i;
			error = nbd_thread_start(lo, dev);
		}
		return error;
	case NBD_SET_INFO:
		if (lo->file)
			return -EBUSY;
		return nbd_set_info(lo, dev, (struct nbd_info *) arg);
	case NBD_GET_INFO:
		if (!lo->file)
			return -ENXIO;
		return nbd_get_info(lo, dev, (struct nbd_info *) arg);
		return lo->harderror;
	case NBD_CLEAR_QUE:
		nbd_clear_que(lo);
		return 0;
#ifdef PARANOIA
	case NBD_PRINT_DEBUG:
		printk(KERN_INFO "NBD device %d: head = %lx, tail = %lx. Global: in %d, out %d\n",
		       dev, (long) lo->head, (long) lo->tail, requests_in, requests_out);
		return 0;
#endif
	case BLKGETSIZE:
		return put_user(nbd_bytesizes[dev] >> 9, (long *) arg);
	}
	return -EINVAL;
}

static int nbd_release(struct inode *inode, struct file *file)
{
	struct nbd_device *lo;
	int dev;

	if (!inode)
		return -ENODEV;
	dev = MINOR(inode->i_rdev);
	if (dev >= MAX_NBD)
		return -ENODEV;
	fsync_dev(inode->i_rdev);
	invalidate_buffers(inode->i_rdev);
	lo = &nbd_dev[dev];
	if (lo->refcnt <= 0)
		printk(KERN_ALERT "nbd_release: refcount(%d) <= 0\n", lo->refcnt);
	lo->refcnt--;
	/* N.B. Doesn't lo->file need an fput?? */
	MOD_DEC_USE_COUNT;
	return 0;
}

static struct file_operations nbd_fops =
{
	NULL,			/* lseek - default */
	block_read,		/* read - general block-dev read */
	block_write,		/* write - general block-dev write */
	NULL,			/* readdir - bad */
	NULL,			/* select */
	nbd_ioctl,		/* ioctl */
	NULL,			/* mmap */
	nbd_open,		/* open */
	NULL,			/* flush */
	nbd_release		/* release */
};

/*
 * And here should be modules and kernel interface 
 *  (Just smiley confuses emacs :-)
 */

#ifdef MODULE
#define nbd_init init_module
#endif

#ifdef NBD_REBOOT_TEARDOWN
static int nbd_notify_teardown(struct notifier_block *this, 
			       unsigned long code, void *x)
{
	struct nbd_device *lo;
	int i;

	for (i = 0; i < MAX_NBD; i++) {
		lo = &nbd_dev[i];
		if (lo->file)
			nbd_thread_delete(lo);
	}
	return NOTIFY_DONE;
}


/* tear down nbd before quiting */
static struct notifier_block teardown_notifier = {
	nbd_notify_teardown,
	NULL,
	2
};
#endif

int nbd_init(void)
{
	int i;

	if (sizeof(struct nbd_request) != 28) {
		printk(KERN_CRIT "Sizeof nbd_request needs to be 28 in order to work!\n" );
		return -EIO;
	}

	if (register_blkdev(MAJOR_NR, "nbd", &nbd_fops)) {
		printk("Unable to get major number %d for NBD\n",
		       MAJOR_NR);
		return -EIO;
	}
#ifdef MODULE
	printk("nbd: registered device at major %d\n", MAJOR_NR);
#endif
	blksize_size[MAJOR_NR] = nbd_blksizes;
	blk_size[MAJOR_NR] = nbd_sizes;
	blk_dev[MAJOR_NR].request_fn = do_nbd_request;
	for (i = 0; i < MAX_NBD; i++) {
		memset(nbd_dev + i, 0, sizeof(struct nbd_device));
		nbd_dev[i].magic = LO_MAGIC;
		nbd_dev[i].replysize = NBD_REPLY_SIZE_V0;
		nbd_blksizes[i] = 1024;
		nbd_blksize_bits[i] = 10;
		nbd_bytesizes[i] = 0x7ffffc00; /* 2GB */
		nbd_sizes[i] = nbd_bytesizes[i] >> nbd_blksize_bits[i];
	}

#ifdef NBD_REBOOT_TEARDOWN
	register_reboot_notifier(&teardown_notifier); 
	notifier_chain_register(&panic_notifier_list, &teardown_notifier); 
#endif	
	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{
	if (unregister_blkdev(MAJOR_NR, "nbd") != 0)
		printk("nbd: cleanup_module failed\n");
	else
		printk("nbd: module cleaned up.\n");

#ifdef NBD_REBOOT_TEARDOWN
	unregister_reboot_notifier(&teardown_notifier); 
	notifier_chain_unregister(&panic_notifier_list, &teardown_notifier); 
#endif
}
#endif
