/*
 *	LCD, LED, and Button interface for Cobalt	
 *      Andrew Bose
 *	Timothy Stonis (x86 version)
 * 	Tim Hockin
 * 	Adrian Sun
 *	Duncan Laurie (RaQ5)
 */
#include <linux/config.h>
#include <linux/module.h>

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/mc146818rtc.h>
#include <linux/netdevice.h>
#include <linux/proc_fs.h>
#include <linux/in6.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/timer.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/checksum.h>
#include <linux/delay.h>
#include <linux/lcd.h>
#include <linux/cobalt_i2c.h>
#include <linux/pc87317.h>

#ifndef min
#define min(a, b)  ((a) < (b) ? (a) : (b))
#endif

#define LCD_DRIVER	"lcd: Cobalt LCD Driver v3.12"

#define USE_I2C_LCD		1


/* io registers */
#define LPT			0x0378
#define LCD_DATA_ADDRESS	LPT+0
#define LCD_CONTROL_ADDRESS	LPT+2

/* LCD device info */
#define LCD_Addr		0x80
#define DD_R00			0x00
#define DD_R01			0x27
#define DD_R10			0x40
#define DD_R11			0x67

enum {
	BOARD_UNKNOWN,
	BOARD_RAQ3,
	BOARD_QUBE3,
	BOARD_RAQ5
} board_type = BOARD_UNKNOWN;

/* driver functions */
static int lcd_ioctl(struct inode *inode, struct file *file, unsigned int cmd, 
	unsigned long arg);
static ssize_t lcd_read(struct file *file, char *buf, size_t count, 
	loff_t *ppos);
static int lcd_open(struct inode *inode, struct file *file);
static int lcd_read_proc(char *buf, char **start, off_t offset, int len, 
	int *eof, void *private);

/* internal functions */
static int lcddev_write_line(int lineno, char *text);
static char *lcddev_read_line(int lineno);

static void lcd_setbits(unsigned int state, unsigned int test, unsigned int value);

/* globals used throughout */
static struct pci_dev *led_dev = NULL;
static unsigned int led_state = 0;
static unsigned int led_bitstate = 0;
static int lcd_present = 0;
static struct proc_dir_entry *proc_lcd = NULL;
static unsigned int use_i2c=0; 
static u16 superio_port = 0;

/* for monterey leds */
static unsigned int fp_leds = 0;
static struct timer_list timer;
static struct lcd_handler lcd_handlers = { NULL, &lcd_handlers, &lcd_handlers};

/* various file operations we support for this driver */
static struct file_operations lcd_fops = {
	NULL,		/* No seek */
	lcd_read,
	NULL,		/* No write */ 
	NULL,		/* No readdir */
	NULL,    	/* No poll */
	lcd_ioctl,
	NULL,		/* No mmap */
	lcd_open,
	NULL		/* No release */	
};

/* device structure */
static struct miscdevice lcd_dev=
{
	COBALT_LCD_MINOR,
	"lcd",
	&lcd_fops
};

#define COBALT_LCD_NEW_MINOR	156
/* for forward compatibility with 2.4 kernels */
static struct miscdevice lcd_new_dev=
{
	COBALT_LCD_NEW_MINOR,
	"lcd",
	&lcd_fops
};

/* 
 * led bits (used in _led_set)
 */

/* RaQ3/4 */
#define RAQ3_SHUTLOGO_ADDR	0x7e
#define RAQ3_SHUTDOWN_OFF	0x40 /* reverse polarity on RaQ3/4 */
#define RAQ3_COBALTLOGO_ON	0x80
#define QUBE3_LIGHTBAR_ON	0xc0 /* direct polarity on Qube3 */

#define RAQ3_WEBLIGHT_ADDR	0xb8
#define RAQ3_WEBLIGHT_ON	0x80

/* RaQ5 */

/* Front panel LEDs on XTR */
#define RAQ5_FP_LED_ETH0_TXRX	0x8000
#define RAQ5_FP_LED_ETH0_LINK	0x1000
#define RAQ5_FP_LED_ETH1_TXRX	0x4000
#define RAQ5_FP_LED_ETH1_LINK	0x0800
#define RAQ5_FP_LED_DISK0	0x2000
#define RAQ5_FP_LED_DISK1	0x0200
#define RAQ5_FP_LED_DISK2	0x0080
#define RAQ5_FP_LED_DISK3	0x0040
#define RAQ5_FP_LED_WEB 	0x0400
#define RAQ5_FP_LED_HEART	0x0100
#define RAQ5_FP_LED_SPARE	0x0020
#define RAQ5_FP_LED_SPARE2	0x0010
/*#define RAQ5_FP_LED_DISK_FAULT	0x0010*/
#define RAQ5_FP_LED_MASK	( RAQ5_FP_LED_ETH0_TXRX| RAQ5_FP_LED_ETH0_LINK| RAQ5_FP_LED_ETH1_TXRX| RAQ5_FP_LED_ETH1_LINK| RAQ5_FP_LED_DISK0| RAQ5_FP_LED_DISK1| RAQ5_FP_LED_DISK2| RAQ5_FP_LED_DISK3)	

#define RAQ5_LED_SLED0          (1 << 3)
#define RAQ5_LED_SLED1          (1 << 2)
#define RAQ5_LED_SLED2          (1 << 1)
#define RAQ5_LED_SLED3          (1 << 0)

static int cobalt_lcd_en = 1; /* zero if LCD should not be loaded */
static int cobalt_i2c_lcd_en = 1; /* zero if I2C LCD should not be loaded */

__initfunc(void cobalt_lcd_disable(char *str, int *ints))
{
    cobalt_lcd_en = 0;
}

__initfunc(void cobalt_i2clcd_disable(char *str, int *ints))
{
    cobalt_i2c_lcd_en = 0;
}

#define CMOS_ADDR_PORT     0x70
#define CMOS_DATA_PORT     0x71
#define CMOS_INFO_MAX      0xff
#define CMOS_FLAG_BYTE_0   0x10
#define CMOS_FLAG_BYTE_1   0x11

static inline u8 cmos_read_byte(int index)
{
	if (index > CMOS_INFO_MAX)
		return 0;

	outb(index, CMOS_ADDR_PORT);
	return inb(CMOS_DATA_PORT);
}

int cmos_read_flag(const unsigned int flag)
{
	u16 flags;

	flags = cmos_read_byte(CMOS_FLAG_BYTE_0) << 8 |
		cmos_read_byte(CMOS_FLAG_BYTE_1);
	
	return (flags & flag) ? 1 : 0;
}

/* deal with COBALTLOGO */
static void lcd_setlogo(const unsigned int ledstate)
{
	u16 addr;
	u8 val;

	outb(SUPERIO_LOGICAL_DEV, SUPERIO_INDEX_PORT); /* select apc */
	outb(SUPERIO_DEV_RTC, SUPERIO_DATA_PORT);

	outb(SUPERIO_ADDR_HIGH, SUPERIO_INDEX_PORT); /* figure out base addr */
	addr = inb(SUPERIO_DATA_PORT) << 8;
	outb(SUPERIO_ADDR_LOW, SUPERIO_INDEX_PORT);
	addr |= inb(SUPERIO_DATA_PORT);
	if (addr) {
		/* set up bank 2 */
		outb(SUPERIO_RTC_CRA, addr);
		val = inb(addr + 1) & 0x8f;
		outb(val | SUPERIO_RTC_BANK_2, addr + 1);

		/* get the current state of APCR4 */
		outb(SUPERIO_APCR4, addr);
		val = inb(addr + 1) & 0x3f;

		if (ledstate & LED_COBALTLOGO) 
			val |= (ledstate & LED_SYSFAULT) ? 0x80 : 0x0;
		else 
			val |= 0x40; /* logo is off */
		outb(val, addr + 1);

		/* go back to bank 0 */
		outb(SUPERIO_RTC_CRA, addr);
		val = inb(addr + 1) & 0x8f;
		outb(val | SUPERIO_RTC_BANK_0, addr + 1);
	}
}

/* deal with sled leds -- all the reset lines are kept high. 
   NOTE: the reset lines are in the reverse order of the switches. */
static void lcd_setsled(const unsigned int ledstate)
{
	u8 leds = 0xf0;

	if (ledstate & LED_SLED0)
		leds |= RAQ5_LED_SLED0;
	if (ledstate & LED_SLED1)
		leds |= RAQ5_LED_SLED1;
	if (ledstate & LED_SLED2)
		leds |= RAQ5_LED_SLED2;
	if (ledstate & LED_SLED3)
		leds |= RAQ5_LED_SLED3;

	cobalt_i2c_write_byte(COBALT_I2C_DEV_RULER, 0, leds);
}

/* 
 * _led_set: actually set the leds (icky details hidden within)
 */
static void _led_set(const unsigned int ledstate)
{
	unsigned char tmp;

	if (!led_dev && ! (BOARD_RAQ5 == board_type)) {
	   return;
	}

	/* RaQ3/4 and Qube3 */
	if (board_type == BOARD_QUBE3 || board_type == BOARD_RAQ3) {
		/* read the current state of shutdown/logo lights */
		pci_read_config_byte(led_dev, RAQ3_SHUTLOGO_ADDR, &tmp);

		/* shutdown light */
		if (board_type == BOARD_RAQ3) {
			/* reverse polarity for shutdown light on RaQ3/4 */
			if (ledstate & LED_SHUTDOWN) {
				tmp &= ~RAQ3_SHUTDOWN_OFF;
			} else {
				tmp |= RAQ3_SHUTDOWN_OFF;
			}
		} else {
			/* Qube3 ignores LED_SHUTDOWN */
		}

		/* logo light */
		if (board_type == BOARD_RAQ3) {
			if (ledstate & LED_COBALTLOGO) {
				/* RaQ3/4 are straight forward */
				tmp |= RAQ3_COBALTLOGO_ON;
			} else {
				tmp &= ~RAQ3_COBALTLOGO_ON;
			}
		} else {
			if (ledstate & LED_COBALTLOGO) {
				/* Qube 3 has a lightbar, not a LOGO */
				tmp |= QUBE3_LIGHTBAR_ON;
			} else {
				tmp &= ~QUBE3_LIGHTBAR_ON;
			}
		}
			
		/* write new values */
		pci_write_config_byte(led_dev, RAQ3_SHUTLOGO_ADDR, tmp);

		/* web light */
		if (board_type == BOARD_RAQ3) {
			pci_read_config_byte(led_dev, RAQ3_WEBLIGHT_ADDR, &tmp);
			if (ledstate & LED_WEBLIGHT) {
				tmp |= RAQ3_WEBLIGHT_ON;
			} else {
				tmp &= ~RAQ3_WEBLIGHT_ON;
			}
			pci_write_config_byte(led_dev, RAQ3_WEBLIGHT_ADDR, tmp);
		} else {
			/*
			 * If we are on a Qube3, dont use the web led.
			 * It's wired to the reset line on the 2nd ide bus
			 * to achieve staggered startup.
			 */
		}
	} else if ( BOARD_RAQ5 == board_type ) {
		lcd_setbits(ledstate, LED_WEBLIGHT, RAQ5_FP_LED_WEB); 	
		lcd_setbits(ledstate, LED_ETH0_TXRX, RAQ5_FP_LED_ETH0_TXRX);
		lcd_setbits(ledstate, LED_ETH0_LINK, RAQ5_FP_LED_ETH0_LINK);
		lcd_setbits(ledstate, LED_ETH1_TXRX, RAQ5_FP_LED_ETH1_TXRX);
		lcd_setbits(ledstate, LED_ETH1_LINK, RAQ5_FP_LED_ETH1_LINK);
		lcd_setbits(ledstate, LED_DISK0, RAQ5_FP_LED_DISK0);
		lcd_setbits(ledstate, LED_DISK1, RAQ5_FP_LED_DISK1);
		lcd_setbits(ledstate, LED_DISK2, RAQ5_FP_LED_DISK2);
		lcd_setbits(ledstate, LED_DISK3, RAQ5_FP_LED_DISK3);
		lcd_setbits(ledstate, LED_HEART, RAQ5_FP_LED_HEART);
		lcd_setbits(ledstate, LED_SPARE, RAQ5_FP_LED_SPARE);
		lcd_setbits(ledstate, LED_SPARE2, RAQ5_FP_LED_SPARE2);
	  	cobalt_i2c_write_byte( COBALT_I2C_DEV_LED_I, 0, 0x00ff & led_bitstate);
	  	cobalt_i2c_write_byte( COBALT_I2C_DEV_LED_II, 0, led_bitstate >> 8 );
		lcd_setlogo(ledstate);
		lcd_setsled(ledstate);
	} 

}

static void lcd_setbits(unsigned int state, unsigned int test, unsigned int value) {
	if ( state & test ) {
		led_bitstate |= value;	
	} else {
		led_bitstate &= ~value;
	}
}

/*
** lcddev_read_inst: Read a control instruction from the LCD
*/
static inline unsigned char 
lcddev_read_inst(void)
{
        unsigned char a;

	if (use_i2c) {
	  a = cobalt_i2c_read_byte( COBALT_I2C_DEV_LCD_INST | COBALT_I2C_READ, 0 );
	} else {
	  outb(0x21,LCD_CONTROL_ADDRESS);      // RS=0, R/W=1, E=0
	  outb(0x20,LCD_CONTROL_ADDRESS);      // RS=0, R/W=1, E=1
	  a=inb(LCD_DATA_ADDRESS);
	  outb(0x21,LCD_CONTROL_ADDRESS);      // RS=0, R/W=1, E=0
	  outb(0x01,LCD_CONTROL_ADDRESS);      // RS=0, R/W=1, E=0
	}

        // This is bad... use jiffies.
        udelay(100);                       // 10ms delay

        return a;
}

#define LCD_MAX_POLL 10000
static inline void 
lcddev_poll_wait(void) 
{
        int i = 0;  
	while (i++ < LCD_MAX_POLL) {
	  if ((lcddev_read_inst() & 0x80) == 0x80)
	      continue;
	  break;
	}
}

/*
** lcddev_write_inst: Write a control instruction to the LCD
*/
static inline void 
lcddev_write_inst(unsigned char data)
{
	lcddev_poll_wait();

	if (use_i2c) {
	  cobalt_i2c_write_byte( COBALT_I2C_DEV_LCD_INST | COBALT_I2C_WRITE, 0, data);
	} else {
	  outb(0x03,LCD_CONTROL_ADDRESS);      // RS=0, R/W=0, E=0
	  outb(data,LCD_DATA_ADDRESS);
	  outb(0x02,LCD_CONTROL_ADDRESS);      // RS=0, R/W=0, E=1
	  outb(0x03,LCD_CONTROL_ADDRESS);      // RS=0, R/W=0, E=0
	  outb(0x01,LCD_CONTROL_ADDRESS);      // RS=0, R/W=1, E=0
	}

        // This is bad... use jiffies.
        udelay(100);                       // 10ms delay
}

/*
** lcddev_write_data: Write one byte of data to the LCD
*/
static inline void 
lcddev_write_data(unsigned char data)
{
	lcddev_poll_wait();
	if (use_i2c) {
	  cobalt_i2c_write_byte( COBALT_I2C_DEV_LCD_DATA | COBALT_I2C_WRITE, 0, data);
	} else {
	  outb(0x07,LCD_CONTROL_ADDRESS);      // RS=1, R/W=0, E=0
	  outb(data,LCD_DATA_ADDRESS);
	  outb(0x06,LCD_CONTROL_ADDRESS);      // RS=1, R/W=0, E=1
	  outb(0x07,LCD_CONTROL_ADDRESS);      // RS=1, R/W=0, E=0
	  outb(0x05,LCD_CONTROL_ADDRESS);      // RS=1, R/W=1, E=0
	}

        // This is bad... use jiffies.
        udelay(100);                         // 1ms delay
}

/*
** lcddev_read_data: Read one byte of data from the LCD
*/
static inline unsigned char 
lcddev_read_data(void)
{
        unsigned char a;

	lcddev_poll_wait();
	if (use_i2c) {
	  a = cobalt_i2c_read_byte( COBALT_I2C_DEV_LCD_DATA | COBALT_I2C_READ, 0 );
	} else {
	  outb(0x25,LCD_CONTROL_ADDRESS);      // RS=1, R/W=1, E=0
	  outb(0x24,LCD_CONTROL_ADDRESS);      // RS=1, R/W=1, E=1
	  a = inb(LCD_DATA_ADDRESS);
	  outb(0x25,LCD_CONTROL_ADDRESS);      // RS=1, R/W=1, E=0
	  outb(0x01,LCD_CONTROL_ADDRESS);      // RS=1, R/W=1, E=0
	}

        // This is bad... use jiffies.
        udelay(100);                        // 1ms delay

        return a;
}

static inline void 
lcddev_init(void)
{
	lcddev_write_inst(0x38);
	lcddev_write_inst(0x38);
	lcddev_write_inst(0x38);
	lcddev_write_inst(0x06);
	lcddev_write_inst(0x0c);
}

static inline char 
ReadButtons ( void )
{
  /* first, check to see if the power button has been pushed */
  if (superio_port &&
      (inb(superio_port + SUPERIO_PM1_STATUS) & SUPERIO_PWRBUTTON)) {
	  outb(SUPERIO_PWRBUTTON, superio_port + SUPERIO_PM1_STATUS);
	  return BUTTON_Power;
  }

  if (use_i2c) {
    unsigned char inst = cobalt_i2c_read_byte( 0x41, 0 );
    if (inst == 0x3e)
      return BUTTON_Next_B;
    if (inst == 0x3d)
      return BUTTON_Enter_B;
    if (inst == 0x1f)
      return BUTTON_Left_B;
    if (inst == 0x3b)
      return BUTTON_Right_B;
    if (inst == 0x2f)
      return BUTTON_Up_B;
    if (inst == 0x37)
      return BUTTON_Down_B;
    if (inst == 0x3f)
      return BUTTON_NONE_B;
    return BUTTON_NONE_B;
  } else {
    outb(0x29,LCD_CONTROL_ADDRESS);  // Sel=0, Bi=1
    return( inb(LCD_DATA_ADDRESS) );
  }
}

static inline int 
button_pressed(void)
{
        unsigned char buttons = ReadButtons();

        if ( (buttons == BUTTON_Next) || (buttons == BUTTON_Next_B) ||
             (buttons == BUTTON_Reset_B) || (buttons == BUTTON_Power) )
                return buttons;
        return 0;
}


/* our driver's ioctl() */
static int
lcd_ioctl(struct inode *inode, struct file *file, unsigned int cmd, 
	unsigned long arg)
{
	struct lcd_display button_display;
	unsigned long address, a;
	int index;

	switch (cmd) {
	/* Turn the LCD on */
	case LCD_On:
		lcddev_write_inst(0x0F);
		break;		
	/* Turn the LCD off */
	case LCD_Off:
		lcddev_write_inst(0x08);
		break;
	/* Reset the LCD */
	case LCD_Reset:
		lcddev_write_inst(0x3F);
		lcddev_write_inst(0x3F);
		lcddev_write_inst(0x3F);
		lcddev_write_inst(0x3F);
		lcddev_write_inst(0x01);
		lcddev_write_inst(0x06);
		break;

	/* Clear the LCD */
	case LCD_Clear:
       		lcddev_write_inst(0x01);     
		break;

	/* Move the cursor one position to the left */
	case LCD_Cursor_Left:
		lcddev_write_inst(0x10);
		break;

	/* Move the cursor one position to the right */
	case LCD_Cursor_Right:
		lcddev_write_inst(0x14);
		break;	

	/* Turn the cursor off */
	case LCD_Cursor_Off:
                lcddev_write_inst(0x0C);
	        break;

	/* Turn the cursor on */
        case LCD_Cursor_On:
                lcddev_write_inst(0x0F);
                break;

	/* Turn blinking off? I don't know what this does - TJS */
        case LCD_Blink_Off:
                lcddev_write_inst(0x0E);
                break;

	/* Get the current cursor position */
        case LCD_Get_Cursor_Pos:{
                struct lcd_display display;

                display.cursor_address = ( lcddev_read_inst() );
                display.cursor_address = ( display.cursor_address & 0x07F );
                copy_to_user((struct lcd_display*)arg, &display, 
		             sizeof(struct lcd_display));

                break;
                }

	/* Set the cursor position */
        case LCD_Set_Cursor_Pos: {
                struct lcd_display display;
                copy_from_user(&display, (struct lcd_display*)arg, 
                              sizeof(struct lcd_display));
                a = (display.cursor_address | LCD_Addr );
                lcddev_write_inst( a );
                break;
                }
	/* Get the value at the current cursor position? - TJS */
        case LCD_Get_Cursor: {
                struct lcd_display display;
                display.character = lcddev_read_data();
                copy_to_user((struct lcd_display*)arg, &display, 
		             sizeof(struct lcd_display));
                lcddev_write_inst(0x10);
                break;
                }

	/* Set the character at the cursor position? - TJS */
	case LCD_Set_Cursor:{
                struct lcd_display display;
                copy_from_user(&display, (struct lcd_display*)arg, 
                              sizeof(struct lcd_display));
                lcddev_write_data( display.character );
                lcddev_write_inst(0x10);
                break;
                }

	/* Dunno what this does - TJS */ 
	case LCD_Disp_Left:
		lcddev_write_inst(0x18);
		break;

	/* Dunno what this does - TJS */ 
	case LCD_Disp_Right:
		lcddev_write_inst(0x1C);
		break;

	/* Dunno what this does - TJS */ 
	case LCD_Home:
		lcddev_write_inst(0x02);
		break;

	/* Write a string to the LCD */
	case LCD_Write: {
		struct lcd_display display;
                copy_from_user(&display, (struct lcd_display*)arg, 
                              sizeof(struct lcd_display));

		display.size1 = display.size1 > 0 ? 
		  min(display.size1, (int) sizeof(display.line1)) : 0;
		display.size2 = display.size2 > 0 ? 
		  min(display.size2, (int) sizeof(display.line2)) : 0;

		/* First line */	
                lcddev_write_inst(0x80);
		for (index = 0; index < display.size1; index++) {
			lcddev_write_data( display.line1[index]);
		}		

		/* Second line */
		lcddev_write_inst(0xC0);	
                for (index = 0; index < display.size2; index++) {
                        lcddev_write_data( display.line2[index]);
		}
 
		break;	
	}

	/* Read what's on the LCD */
        case LCD_Read: {
                struct lcd_display display;

                for (address = DD_R00; address <= DD_R01; address++) {
                        a = (address | LCD_Addr );

                        lcddev_write_inst( a );
                        display.line1[address] = lcddev_read_data();
                }
                display.line1[ 0x27 ] = '\0';
                for (address = DD_R10; address <= DD_R11; address++) {
                        a = (address | LCD_Addr );
                        lcddev_write_inst( a );
                        display.line2[address - 0x40 ] = lcddev_read_data();
                 }
                display.line2[ 0x27 ] = '\0';
                copy_to_user((struct lcd_display*)arg, &display,
                            sizeof(struct lcd_display));
                break;
        }

	case LCD_Raw_Inst: {
                struct lcd_display display;
                copy_from_user(&display, (struct lcd_display*)arg,
                              sizeof(struct lcd_display));
                lcddev_write_inst( display.character );
                break;
                }

        case LCD_Raw_Data: {
                struct lcd_display display;
                copy_from_user(&display, (struct lcd_display*)arg,
                              sizeof(struct lcd_display));
                lcddev_write_data( display.character );
                break;
                }

	case LCD_Type:
	        put_user(use_i2c ? LCD_TYPE_I2C : LCD_TYPE_PARALLEL_B, 
			 (int *) arg);
		break;

	/* Read the buttons */
        case BUTTON_Read: {
                button_display.buttons = ReadButtons();;
                copy_to_user((struct lcd_display*)arg, &button_display, 
		             sizeof(struct lcd_display));
                break;
        }
	
	/* a slightly different api that allows you to set 32 leds */
	case LED32_Set:
		fp_leds |= arg;
		led_state = arg;
		break;

	case LED32_Bit_Set:
		fp_leds |= arg;
		led_state |= arg;
		break;

	case LED32_Bit_Clear: 
		led_state &= ~arg;
		break;

	/* set all the leds */
	case LED_Set:
	  {
                struct lcd_display display;
                copy_from_user(&display, (struct lcd_display*)arg,
                              sizeof(struct lcd_display));
		fp_leds |= display.leds;
		led_state = display.leds;
		break;
	  }

	/* set a single led */
	case LED_Bit_Set:
	  { 
                struct lcd_display display;
                copy_from_user(&display, (struct lcd_display*)arg,
                              sizeof(struct lcd_display));
		fp_leds |= display.leds;
		led_state |= display.leds;
	  }
	  break;

	/* clear an led */
	case LED_Bit_Clear:
	  { 
                struct lcd_display display;
                copy_from_user(&display, (struct lcd_display*)arg,
                              sizeof(struct lcd_display));
		led_state &= ~display.leds;
	  }
	  break;

	default:
		return 0;
		break;

	}

	return 0;
}

/* our driver's open() */
static int 
lcd_open(struct inode *inode, struct file *file)
{
	if (!lcd_present)
		return -ENXIO;
	else
		return 0;
}

/* LED daemon sits on this, we wake it up once a key is pressed */
static int lcd_waiters = 0;

/* our driver's read() */
static ssize_t 
lcd_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
        int buttons_now;

        if(lcd_waiters > 0)
                return -EINVAL;

        lcd_waiters++;
        while(((buttons_now = (int)button_pressed()) == 0) &&
              !(signal_pending(current))) {
                current->state = TASK_INTERRUPTIBLE;
                schedule_timeout((2 * HZ));
        }
        lcd_waiters--;

        if(signal_pending(current))
                return -ERESTARTSYS;
        return buttons_now;
}

/* read a single line from the LCD into a string */
static char *
lcddev_read_line(int lineno)
{
	char *line;
	unsigned long address;
	unsigned long min, max;

	line = kmalloc(LCD_CHARS_PER_LINE, GFP_KERNEL);
	if (!line)
		return NULL;

	if (lineno == 0) {
		min = DD_R00;
		max = DD_R01;
	} else if (lineno == 1) {
		min = DD_R10;
		max = DD_R11;
	} else {
		min = 1;
		max = 0;
	}

	for (address = min; address <= max; address++) {
                unsigned long a = (address | LCD_Addr );

                udelay(150);
                lcddev_write_inst(a);

                udelay(150);
                line[address-min] = lcddev_read_data();
	}
	line[address-min] = '\0';

	return line;
}

/* write a string to a given line of the LCD */
static int 
lcddev_write_line(int lineno, char *text)
{
	int i;
	char *blank = "                ";

	udelay(150);
	if (lineno == 0) {
		lcddev_write_inst(0x80);
	} else if (lineno == 1) {
		lcddev_write_inst(0xC0);
	}
	udelay(150);

	if (!text)
		text = blank;

	for (i = 0; i < 16; i++) {
		udelay(150);
		if (i < strlen(text)) {
			lcddev_write_data(text[i]);
		} else {
			lcddev_write_data(' ');
		}
	}

	return strlen(text);
}

static int 
lcd_read_proc(char *buf, char **start, off_t offset, int len, int *eof, 
	void *private)
{
	int plen = 0;
	char *tmp;

	/* first line */
	tmp = lcddev_read_line(0);
	if (tmp){
		plen += sprintf(buf+plen, "%s\n", tmp);
		kfree(tmp);
	} else {
		plen += sprintf(buf+plen, "\n");
	}

	/* second line */
	tmp = lcddev_read_line(1);
	if (tmp) {
		plen += sprintf(buf+plen, "%s\n", tmp);
		kfree(tmp);
	} else {
		plen += sprintf(buf+plen, "\n");
	}

        return plen;
}

/* toggle the front panel leds. fp_leds is there so that
 * leds will at least show once. */
static void lcd_timer(unsigned long data)
{
	unsigned int *leds = (unsigned int *) data;
	struct lcd_handler *start = lcd_handlers.next;
	
	/* run lcd handlers */
	while (start != &lcd_handlers) {
		if (start->fcn)
			start->fcn();
		start = start->next;
	}

	/* set the leds and then clear out special leds */
	_led_set(*leds | led_state);
	*leds = 0;

	timer.expires = jiffies + HZ/20;
	add_timer(&timer);
}

/* exported functions -- always use the timer to set stuff. */
void 
lcd_led_set(const unsigned int leds)
{
	led_state = leds;
}

unsigned int 
lcd_led_get(void)
{
	return led_state;
}

void lcd_fp_led_state(const unsigned char set, const unsigned int val)
{
	if (set)
		fp_leds |= val;
	else 
		fp_leds &= ~val;
}

int lcd_handler_add(struct lcd_handler *handle)
{
	if (!handle)
		return -1;
	handle->next = lcd_handlers.next;
	handle->prev = &lcd_handlers;
	lcd_handlers.next->prev = handle;
	lcd_handlers.next = handle;
	return 0;
}

void lcd_handler_del(struct lcd_handler *handle)
{
	if (!handle)
		return;
	handle->next->prev = handle->prev;
	handle->prev->next = handle->next;
}
EXPORT_SYMBOL(lcd_led_set);
EXPORT_SYMBOL(lcd_led_get);
EXPORT_SYMBOL(lcd_fp_led_state);
EXPORT_SYMBOL(lcd_handler_add);
EXPORT_SYMBOL(lcd_handler_del);

/* called at init time only */
/* FIXME: __INITFUNC() ? */
int __init lcd_init(void)
{	
    unsigned int leds = LED_SHUTDOWN | LED_COBALTLOGO;

    if (!cobalt_lcd_en) {
	printk("%s [DISABLED]\r\n", LCD_DRIVER);
	return -1;
    }

    printk("%s\r\n", LCD_DRIVER);
    misc_register(&lcd_dev);
    misc_register(&lcd_new_dev);
    
    /* scan the pci bus for led ports */
    if (!pci_present())
	return -1;
    
    /* there is a board identifier for RaQ3/4 vs Qube3 on the PMU @ 0x7f */
    led_dev = pci_find_device(PCI_VENDOR_ID_AL,
			      PCI_DEVICE_ID_AL_M7101, led_dev);

    if (led_dev) {
	/* 
	 * check to see what board we are on
	 * (RaQ[34]/Qube3 ONLY)
	 */
	unsigned char i, val;

	/* momentarily set DOGB# to input */
	pci_read_config_byte(led_dev, 0x7d, &i);
	i &= ~0x20;
	pci_write_config_byte(led_dev, 0x7d, i);

	/* read the GPIO register */
	pci_read_config_byte(led_dev, 0x7f, &val);

	/* change DOGB back to output */
	pci_read_config_byte(led_dev, 0x7d, &i);
	i |= 0x20;
	pci_write_config_byte(led_dev, 0x7d, i);

	/* RaQ3/4 boards have DOGB (0x20) pulled high, Qube3 has DOGB low */ 
	board_type = (val & 0x20) ? BOARD_RAQ3 : BOARD_QUBE3;
    }

#if USE_I2C_LCD
    else {
	/*
	 * check for RaQ5 board -- set initial sled led states.
	 */
#ifdef CONFIG_COBALT_GEN_V
	    leds |= (LED_SLED0 | LED_SLED1 | LED_SLED2 | LED_SLED3);
#endif
	    if (cobalt_i2c_lcd_en) {
		    int type = cobalt_i2c_type();
		    if ((type == COBALT_I2C_TYPE_GEN_V) &&
			(cobalt_i2c_read_byte(COBALT_I2C_DEV_BUTTON, 0) != 0xff)) {
			    u16 addr;

			    /* superio power button -- get ioport */
			    outb(SUPERIO_LOGICAL_DEV, SUPERIO_INDEX_PORT);
			    outb(SUPERIO_DEV_PM, SUPERIO_DATA_PORT);
			    outb(SUPERIO_ADDR_HIGH, SUPERIO_INDEX_PORT);
			    addr = inb(SUPERIO_DATA_PORT) << 8;
			    outb(SUPERIO_ADDR_LOW, SUPERIO_INDEX_PORT);
			    addr |= inb(SUPERIO_DATA_PORT);
			    if (addr) {
				    outb(0x08, addr);
				    superio_port = inb(addr + 1);
				    outb(0x09, addr);
				    superio_port |= inb(addr + 1) << 8;
			    }

			    use_i2c=1;
			    board_type = BOARD_RAQ5;
		    }
	    }
    }
#endif

    /* setup up timer for leds */
    init_timer(&timer);
    timer.expires = jiffies + HZ/20;
    timer.data = (unsigned long) &fp_leds;
    timer.function = &lcd_timer;
    add_timer(&timer);

    /* set the initial state */
    lcd_led_set(leds | (cmos_read_flag(CMOS_SYSFAULT_FLAG) ? 
		       LED_SYSFAULT : 0));

    /* flag ourselves as present */
    lcd_present = 1;

    /* initialize the device */
    lcddev_init();

    /* create /proc/lcd */
    proc_lcd = create_proc_entry("lcd", 0, NULL);
    if (proc_lcd) {
	proc_lcd->read_proc = lcd_read_proc;
    }

    return 0;
}
