/* 
 * cobalt acpi driver 
 * Copyright (c) 2000, Cobalt Networks, Inc.
 *
 * author: asun@cobalt.com
 *
 * right now, this driver just handles the drive bay trigger detection
 *
 * if acpi support really existed in the kernel, we would read
 * data from the ACPI tables. however, it doesn't. as a result,
 * we use some hardcoded values. 
 */

#include <stdarg.h>
#include <stddef.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/config.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <linux/ide.h>
#include <linux/hdreg.h>
#include <linux/notifier.h>
#include <linux/sysctl.h>
#include <linux/reboot.h>
#include <linux/cobalt_i2c.h>
#include <linux/pc87317.h>
#include <linux/delay.h>
#include <linux/lcd.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/spinlock.h>

#if defined(CONFIG_COBALT_GEN_V)

#define POWER_BUTTON_SHUTDOWN     0

#define ACPI_IRQ         10       /* XXX: hardcoded interrupt */
#define ACPI_NAME        "sci"

#define OSB4_EVENT       0x40
#define OSB4_INDEX_PORT  0x0cd6
#define OSB4_DATA_PORT   0x0cd7

#define ACPI_TIMEOUT     (HZ >> 1)  /* .5s */
#define MAX_SWITCHES     4

#define SUPERIO_EVENT	0xff

extern void ide_revalidate_disk(kdev_t);

char cobalt_acpi_cmd[256] = "/etc/rc.d/rc.raid";

static spinlock_t acpi_lock = SPIN_LOCK_UNLOCKED;
static struct timer_list cobalt_acpi_timer;
static ide_hwif_t *channels[MAX_SWITCHES];
static u16 osb4_port, superio_port = 0;
static u8 switches, ruler_detect;

/* this is just here in case we want to make this a module */
static struct ctl_table_header *table_header;

static ctl_table hotadd_table[] = {
	{DEV_COBALT_ACPI_HOTADD, "hotaddcmd", &cobalt_acpi_cmd, 
	 sizeof(cobalt_acpi_cmd), 0644, NULL, &proc_dostring},
	{0}
};

static ctl_table dir_table[] = {
	{DEV_COBALT_ACPI, "cobalt_acpi", NULL, 0, 0555, 
	 hotadd_table},
	{0}
};

static ctl_table root_table[] = {
	{CTL_DEV, "dev", NULL, 0, 0555, dir_table},
	{0}
};

static u16 get_reg(u16 index, u16 data, u8 port)
{
	u16 reg;

	outb(port, index);
	reg = inb(data);
	outb(port + 1, index);
	reg |= inb(data) << 8;
	return reg;
}

static void acpi_timer(unsigned long data)
{
	unsigned long flags;
	u8 state;
	int i, j, k, bit;

	spin_lock_irqsave(&acpi_lock, flags);
	state = cobalt_i2c_read_byte(COBALT_I2C_DEV_SWITCH, 0);
	if ((state & 0xf0) == 0xf0)
		state = cobalt_i2c_read_byte(COBALT_I2C_DEV_SWITCH, 0);
	if ((state & 0xf0) == 0xf0)
		goto unlock_restore;

	for (state &= 0xf, i = 0; i < MAX_SWITCHES; i++) {
	  	bit = 1 << i;
		j = state & bit;
		k = switches & bit;
		if (j == k)
			continue;

	  	if (j) {
			switches |= bit;
	    		if (channels[i]) {
				printk("disabling ide ruler channel %d\n", i);
	      			channels[i]->busproc(&channels[i]->drives[0], 
						     BUSSTATE_TRISTATE);
	    		}
	  	} else {
#if 0
			/* poke the acpi_thread */
			switches &= ~bit;
			if (channels[i]) {
				cobalt_acpi_wakeup();
			}
#endif
		}
	}

unlock_restore:
	spin_unlock_irqrestore(&acpi_lock, flags);
}


/* this does the following:
   1) reads the power button. if it has been pushed,
      a system shutdown gets called.
   2) reads the switch to reset itself if necessary.
      set up a timer to check the state in a little bit. 
   3) clears the ACPI generated interrupts. 
*/
static void acpi_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned long flags;
	spin_lock_irqsave(&acpi_lock, flags);
	if (ruler_detect) {
		u8 state;

		outb(OSB4_EVENT, osb4_port); /* turn off SCI generation */
		state = cobalt_i2c_read_byte(COBALT_I2C_DEV_SWITCH, 0);
		if ((state & 0xf0) == 0xf0)
			state = cobalt_i2c_read_byte(COBALT_I2C_DEV_SWITCH, 0);
		if ((state & 0xf0) != 0xf0)
			mod_timer(&cobalt_acpi_timer, jiffies + ACPI_TIMEOUT);
		udelay(10);
	}

	if (superio_port) {
#if POWER_BUTTON_SHUTDOWN
		u8 power = inb(superio_port + SUPERIO_PM1_STATUS);
		if (power & SUPERIO_PWRBUTTON) {
			outb(SUPERIO_PWRBUTTON, 
			     superio_port + SUPERIO_PM1_STATUS);
			notifier_call_chain(&reboot_notifier_list, 
					    SYS_POWER_OFF, NULL);
			printk(KERN_EMERG "Power down.\n");
			machine_power_off();
		}
#endif
		outb(SUPERIO_EVENT, superio_port); 
		outb(SUPERIO_EVENT, superio_port + 1);
	}
	spin_unlock_irqrestore(&acpi_lock, flags);
}


/* XXX: hacky hacky hacky. this will change if someone sticks
 * another ide card in */
void cobalt_acpi_switch(ide_hwif_t *hwif)
{
        int i;

        if (!hwif)
	       return;

	i = hwif->name[3] - '0' - 2;
	if (i >= MAX_SWITCHES)
	       return;

	channels[i] = hwif;
}

void __init cobalt_acpi_init(void)
{
	int type = cobalt_i2c_type();
	int err, reg;
	u16 addr;


	if (type != COBALT_I2C_TYPE_GEN_V)
		return;
	
	/* initialize switches. */
	switches = cobalt_i2c_read_byte(COBALT_I2C_DEV_SWITCH, 0);
	if ((switches & 0xf0) == 0xf0)
		switches = cobalt_i2c_read_byte(COBALT_I2C_DEV_SWITCH, 0);
	ruler_detect = ((switches & 0xf0) == 0xf0) ? 0 : 1;
	switches &= 0xf;

	/* setup osb4 i/o regions */
	if ((reg = get_reg(OSB4_INDEX_PORT, OSB4_DATA_PORT, 0x20)))
		request_region(reg, 4, "OSB4 (pm1a_evt_blk)");
	if ((reg = get_reg(OSB4_INDEX_PORT, OSB4_DATA_PORT, 0x22)))
		request_region(reg, 2, "OSB4 (pm1a_cnt_blk)");
	if ((reg = get_reg(OSB4_INDEX_PORT, OSB4_DATA_PORT, 0x24)))
		request_region(reg, 4, "OSB4 (pm_tmr_blk)");
	if ((reg = get_reg(OSB4_INDEX_PORT, OSB4_DATA_PORT, 0x26)))
		request_region(reg, 6, "OSB4 (p_cnt_blk)");
	osb4_port = reg = get_reg(OSB4_INDEX_PORT, OSB4_DATA_PORT, 0x28);
	if (reg) 
		request_region(reg, 8, "OSB4 (gpe0_blk)");
	else
		ruler_detect = 0;

	/* superi/o -- select pm logical device and get base address */
	outb(SUPERIO_LOGICAL_DEV, SUPERIO_INDEX_PORT);
	outb(SUPERIO_DEV_PM, SUPERIO_DATA_PORT);
	outb(SUPERIO_ADDR_HIGH, SUPERIO_INDEX_PORT);
	addr = inb(SUPERIO_DATA_PORT) << 8;
	outb(SUPERIO_ADDR_LOW, SUPERIO_INDEX_PORT);
	addr |= inb(SUPERIO_DATA_PORT); 
	if (addr) {
		/* get registers */
		if ((reg = get_reg(addr, addr + 1, 0x08))) {
			request_region(reg, 4, "SUPERIO (pm1_evt_blk)");
			superio_port = reg;
		}
		if ((reg = get_reg(addr, addr + 1, 0x0c))) 
			request_region(reg, 2, "SUPERIO (pm1_cnt_blk)");
		if ((reg = get_reg(addr, addr + 1, 0x0e)))
			request_region(reg, 16, "SUPERIO (gpe_blk)");
	}

	/* initialize hwif variable */
	memset(channels, 0, sizeof(channels));

	/* initialize sysctl entry */
	table_header = register_sysctl_table(root_table, 1);
	
	/* initialize our sci timer */
	init_timer(&cobalt_acpi_timer);
	cobalt_acpi_timer.function = acpi_timer;
	
	/* setup an interrupt handler for an ACPI SCI */
	err = request_irq(ACPI_IRQ, acpi_interrupt, 
			  SA_SHIRQ, ACPI_NAME, NULL);
	if (err)
		printk("couldn't assign ACPI IRQ\n");
}


#endif
