/*
 * $Id: cobalt_i2c.c,v 1.10 2001/10/20 07:22:49 asun Exp $
 * cobalt i2c driver support.
 *
 * Copyright (C) 2000 Cobalt Networks, Inc. 
 *
 * borrowed from the rom code. 
 *
 */
#include <stddef.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/config.h>
#include <linux/pci.h>
#include <linux/cobalt_i2c.h>
#include <linux/delay.h>
#include <asm/io.h>

#if defined(CONFIG_COBALT_GEN_III) || defined(CONFIG_COBALT_GEN_V)

#define I2C_3K_STATUS                   0x00
#define I2C_3K_CMD                      0x01
#define I2C_3K_START                    0x02
#define I2C_3K_ADDR                     0x03
#define I2C_3K_LOW_DATA                 0x04
#define I2C_3K_HIGH_DATA                0x05
#define I2C_3K_BLOCK_DATA               0x06
#define I2C_3K_INDEX                    0x07

#define I2C_3K_STATUS_IDLE              0x04

#define I2C_3K_CMD_RW_BYTE              0x20
#define I2C_3K_CMD_RW_WORD              0x30
#define I2C_3K_CMD_RW_BLOCK             0xC0
#define I2C_3K_CMD_RESET_PTR            0x80

#define I2C_5K_HOST_STATUS              0x00
#define I2C_5K_SLAVE_STATUS             0x01
#define I2C_5K_HOST_CONTROL             0x02
#define I2C_5K_HOST_COMMAND             0x03
#define I2C_5K_HOST_ADDR                0x04
#define I2C_5K_DATA_0                   0x05
#define I2C_5K_DATA_1                   0x06
#define I2C_5K_BLOCK_DATA               0x07
#define I2C_5K_SLAVE_CONTROL            0x08
#define I2C_5K_SHADOW_COMMAND           0x09
#define I2C_5K_SLAVE_EVENT              0x0a
#define I2C_5K_SLAVE_DATA               0x0c

#define I2C_5K_STATUS_BUSY              0x01

#define I2C_5K_HOST_CMD_START           0x40
#define I2C_5K_HOST_CMD_QUICK_RW        (0x00 << 2)
#define I2C_5K_HOST_CMD_BYTE_RW         (0x01 << 2)
#define I2C_5K_HOST_CMD_BYTE_DATA_RW    (0x02 << 2)
#define I2C_5K_HOST_CMD_WORD_DATA_RW    (0x03 << 2)
#define I2C_5K_HOST_CMD_BLOCK_DATA_RW   (0x05 << 2)

#define I2C_WRITE                       0x0
#define I2C_READ                        0x1

#define I2C_SYSTYPE                     0
#define I2C_STATUS                      1
#define I2C_ADDR                        2
#define I2C_INDEX                       3
#define I2C_START                       4
#define I2C_DATA_0                      5
#define I2C_DATA_1                      6
#define I2C_DATA_BLOCK                  7
#define I2C_RW_BYTE                     8
#define I2C_RW_WORD                     9
#define I2C_RW_BLOCK                    10
#define I2C_MAX                         11

/* status bits for i2c bus locking */
#define I2C_LOCK                        0
#define I2C_DEAD                        1

/* internal functions */
static inline int cobalt_i2c_lock(void); 
static inline void cobalt_i2c_unlock(void);

static unsigned int i2c_io_port = 0;
static unsigned char map[I2C_MAX];
static unsigned char i2c_state = 0;
static spinlock_t i2c_lock = SPIN_LOCK_UNLOCKED;
static unsigned long i2c_flags;

static inline int cobalt_i2c_lock(void) 
{
	if (test_bit(I2C_DEAD, &i2c_state))
		return -1;
	spin_lock_irqsave(&i2c_lock, i2c_flags);
	udelay(1);
	return 0;
}

static inline void cobalt_i2c_unlock(void)
{
	spin_unlock_irqrestore(&i2c_lock, i2c_flags);
}

/* do a little squelching */
#define NOISE_RATE (5*HZ)
static int i2c_noisy(void)
{
	static unsigned long last_time = 0;
	static unsigned int messages = 0;

	if ((long) (jiffies - last_time) > NOISE_RATE) {
		last_time = jiffies;
		if (messages) {
			printk("cobalt_i2c: skipped %u kernel messages\n", 
			       messages);
			messages = 0;
		}
		return 1;
	}
	++messages;
	return 0;
}

static int i2c_wait_for_smi(void)
{
	static unsigned int shutup = 0;
	int timeout=10;
	int status;
	
	while (timeout-- > 0) {
		udelay(500);
		status = inb_p(i2c_io_port + map[I2C_STATUS]);
		if (map[I2C_SYSTYPE] == COBALT_I2C_TYPE_GEN_V) {
		  if ((status & I2C_5K_STATUS_BUSY) == 0)
				  return 0;
		} else if (status & I2C_3K_STATUS_IDLE)
			return 0;
		outb_p(status, i2c_io_port + map[I2C_STATUS]);
	}

	/* still busy - punch the abort bit */
	if (i2c_noisy()) {
		printk("i2c_wait_for_smi: status busy (0x%x) : resetting controller\n", status);
		if (++shutup > 2)
			set_bit(I2C_DEAD, &i2c_state);
	}

	if (map[I2C_SYSTYPE] == COBALT_I2C_TYPE_GEN_V) {
		outb_p(0x2, i2c_io_port + I2C_5K_HOST_CONTROL);
		outb_p(0x1, i2c_io_port + I2C_5K_HOST_CONTROL); 
		/* now, reset the status registers */
		outb_p(0xff, i2c_io_port + I2C_5K_HOST_STATUS);
	} else 
		outb_p(0x4, i2c_io_port + I2C_3K_CMD);
	return -1;
}

static inline int i2c_setup(const int dev, const int index, const int r)
{
	if ( i2c_wait_for_smi() < 0 )
		return -1;
	outb_p( 0xff, i2c_io_port + map[I2C_STATUS]); /* clear status */
	outb_p((dev | r) & 0xff, i2c_io_port + map[I2C_ADDR]); /*device address*/
	outb_p(index&0xff, i2c_io_port + map[I2C_INDEX]); /* I2C index */
	return 0;
}
	
static inline int i2c_cmd(const unsigned char command)
{
	if (map[I2C_SYSTYPE] == COBALT_I2C_TYPE_GEN_V) 
		outb_p( I2C_5K_HOST_CMD_START | command,
		      i2c_io_port + I2C_5K_HOST_CONTROL);
	else {
		outb_p( command, i2c_io_port + I2C_3K_CMD ); 
		outb_p( 0xff, i2c_io_port + I2C_3K_START );
	}

	if ( i2c_wait_for_smi() < 0 )
		return -1;

	return 0;
}

void __init cobalt_i2c_init(void)
{
	struct pci_dev *pmu = pci_find_device(PCI_VENDOR_ID_AL, 
					      PCI_DEVICE_ID_AL_M7101, NULL);
	memset(map, 0, sizeof(map));
        if (pmu) {
              pci_read_config_dword(pmu, 0x14, &i2c_io_port);
	      map[I2C_SYSTYPE] = COBALT_I2C_TYPE_GEN_III;

	      map[I2C_STATUS] = I2C_3K_STATUS;
	      map[I2C_ADDR] = I2C_3K_ADDR;
	      map[I2C_INDEX] = I2C_3K_INDEX;
	      map[I2C_DATA_0] = I2C_3K_LOW_DATA;
	      map[I2C_DATA_1] = I2C_3K_HIGH_DATA;
	      map[I2C_DATA_BLOCK] = I2C_3K_BLOCK_DATA;
	      map[I2C_RW_BYTE] = I2C_3K_CMD_RW_BYTE;
	      map[I2C_RW_WORD] = I2C_3K_CMD_RW_WORD;
	      map[I2C_RW_BLOCK] = I2C_3K_CMD_RW_BLOCK;

        } else if ((pmu = pci_find_device(PCI_VENDOR_ID_SERVERWORKS,
					  PCI_DEVICE_ID_SERVERWORKS_OSB4, 
					  NULL))) {
		pci_read_config_dword(pmu, 0x90, &i2c_io_port);
		map[I2C_SYSTYPE] = COBALT_I2C_TYPE_GEN_V;

		map[I2C_STATUS] = I2C_5K_HOST_STATUS;
		map[I2C_ADDR] = I2C_5K_HOST_ADDR;
		map[I2C_INDEX] = I2C_5K_HOST_COMMAND;
		map[I2C_DATA_0] = I2C_5K_DATA_0;
		map[I2C_DATA_1] = I2C_5K_DATA_1;
		map[I2C_DATA_BLOCK] = I2C_5K_BLOCK_DATA;
		map[I2C_RW_BYTE] = I2C_5K_HOST_CMD_BYTE_DATA_RW;
		map[I2C_RW_WORD] = I2C_5K_HOST_CMD_WORD_DATA_RW;
		map[I2C_RW_BLOCK] = I2C_5K_HOST_CMD_BLOCK_DATA_RW;
	}
          
        i2c_io_port &= 0xfff0;
        if (pmu && i2c_io_port) 
		printk("Cobalt I2C bus initialized\n");
}


int cobalt_i2c_type(void)
{
	return map[I2C_SYSTYPE];
}


int cobalt_i2c_reset(void)
{
	int ret;

	if ( cobalt_i2c_lock() < 0 )
		return -1;

	if (map[I2C_SYSTYPE] == COBALT_I2C_TYPE_GEN_V) {
		/* i'm not sure if this works */
		outb_p(0x2, i2c_io_port + I2C_5K_HOST_CONTROL);
		outb_p(0x1, i2c_io_port + I2C_5K_HOST_CONTROL);
		outb_p(0xff, i2c_io_port + I2C_5K_HOST_STATUS);  /* clear status */
		outb_p( I2C_5K_HOST_CMD_START | 0x08,
		      i2c_io_port + I2C_5K_HOST_CONTROL);
	} else {
		outb_p(0xff, i2c_io_port + I2C_3K_STATUS);  /* clear status */
		outb_p(0x08, i2c_io_port + I2C_3K_CMD);     /* reset SMB devices */
		outb_p(0xff, i2c_io_port + I2C_3K_START);   /* start command */
	}
	ret = i2c_wait_for_smi();
	cobalt_i2c_unlock();
	return ret;
}

int cobalt_i2c_read_byte(const int dev, const int index )
{
	int val;

	if ( cobalt_i2c_lock() < 0 )
		return 0; /* -1 will make the lcd driver spin in a loop */

	val = -1;
	if ( i2c_setup(dev, index, I2C_READ) < 0 || 
	     i2c_cmd(map[I2C_RW_BYTE]) < 0 ) 
		goto read_byte_done;

	val = inb_p(i2c_io_port + map[I2C_DATA_0]);

read_byte_done:
	cobalt_i2c_unlock();
	return val;
}

int cobalt_i2c_read_word(const int dev, const int index)
{
	int val;

	if ( cobalt_i2c_lock() < 0 )
		return -1;
	
	val = -1;
	if ( i2c_setup(dev, index, I2C_READ) < 0 ||
	     i2c_cmd(map[I2C_RW_WORD]) < 0) 
		goto read_word_done;

	val = inb_p(i2c_io_port + map[I2C_DATA_0]) +
		(inb_p(i2c_io_port + map[I2C_DATA_1]) << 8);

read_word_done:
	cobalt_i2c_unlock();
	return val;
}

int cobalt_i2c_read_block(const int dev, const int index, unsigned char *data,
		   int count )
{
	int ret;

	if ( cobalt_i2c_lock() < 0 )
		return -1;
	
	ret = -1;
	if ( i2c_setup(dev, index, I2C_READ) < 0 ) 
		goto read_block_done;

	outb_p( count & 0xff, i2c_io_port + map[I2C_DATA_0]);
	outb_p( count & 0xff, i2c_io_port + map[I2C_DATA_1]);
	if (i2c_cmd(map[I2C_RW_BLOCK]) < 0) 
		goto read_block_done;

	while (count) {
		/* read a byte of block data */
		*data = inb_p(i2c_io_port + map[I2C_DATA_BLOCK]);
		data++;
		count--;
	}
	ret = 0;

read_block_done:
	cobalt_i2c_unlock();
	return ret;	
}

int cobalt_i2c_write_byte(const int dev, const int index, const u8 val)
{
	int ret;

	if ( cobalt_i2c_lock() < 0 ) 
		return -1;

	ret = -1;
	if ( i2c_setup(dev, index, I2C_WRITE) < 0 ) 
		goto write_byte_done;

	outb_p(val & 0xff, i2c_io_port + map[I2C_DATA_0]);
	if (i2c_cmd(map[I2C_RW_BYTE]) < 0) 
		goto write_byte_done;

	ret = 0;
write_byte_done:
	cobalt_i2c_unlock();
	return ret;	
}

int cobalt_i2c_write_word(const int dev, const int index, const u16 val )
{
	int ret;
	if ( cobalt_i2c_lock() < 0 ) 
		return -1;

	ret = -1;
	if ( i2c_setup(dev, index, I2C_WRITE) < 0 ) 
		goto write_word_done;

	outb_p(val & 0xff, i2c_io_port + map[I2C_DATA_0]);
	outb_p((val >> 8) & 0xff, i2c_io_port + map[I2C_DATA_1]);
	if (i2c_cmd(map[I2C_RW_WORD]) < 0) 
		goto write_word_done;
	ret = 0;

write_word_done:
	cobalt_i2c_unlock();
	return ret;	
}

int cobalt_i2c_write_block( int dev, int index, unsigned char *data, int count)
{
	int ret;

	if ( cobalt_i2c_lock() < 0 ) 
		return -1;

	ret = -1;
	if ( i2c_setup(dev, index, I2C_WRITE) < 0 ) 
		goto write_block_done;

	outb_p( count & 0xff, i2c_io_port + map[I2C_DATA_0]);
	outb_p( count & 0xff, i2c_io_port + map[I2C_DATA_1]);
	if (i2c_cmd(map[I2C_RW_BLOCK]) < 0) 
		goto write_block_done;

	while( count )	{
		/* write a byte of block data */
		outb_p( *data, i2c_io_port + map[I2C_DATA_BLOCK]); 
		data++;
		count--;
	}
	ret = 0;

write_block_done:
	cobalt_i2c_unlock();
	return ret;	
}


#endif
