/*
 *
 *   Author: Philip Gladstone, Axent Technologies
 *           modified for Nat Semi DP83717 by asun@cobalt.com
 *   Copyright (c) 2000  Axent Technologies, Cobalt Networks
 *
 *   This module interrogates the DS2401 Silicon Serial Number chip
 *   that is attached to the GPIO3 pin of the ALI 1543
 *
 *   It exports /proc/hostid which is the four significant bytes of the
 *   id. It can be linked to /var/adm/hostid or /etc/hostid for the 
 *   hostid command to use.
 *
 *   It exports /proc/serialnumber which is the entire 64 bit value read
 *   back (in ascii).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <linux/module.h>

#include <linux/stddef.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#if LINUX_VERSION_CODE < 131584
#else
#include <asm/uaccess.h>
#define proc_register_dynamic proc_register
#endif

#include <linux/pci.h>
#include <linux/pc87317.h>
#include <asm/io.h>


static char rcsid[] = "@(#) $Id: cobalt_ssn.c,v 1.9 2001/05/17 16:53:11 thockin Exp $";

#ifndef __initfunc
#define __initfunc(x)   x
#endif

#define SSN_SIZE     8

static unsigned char ssn[SSN_SIZE];

static char ssn_string[SSN_SIZE*2 + 2];

static unsigned char hostid[4];

static int debug;

#ifdef MODULE
MODULE_PARM(debug,"i");
#endif

#define IO_TYPE_PCI    0
#define IO_TYPE_IO     1

static struct _chip_type {
    char *name;
    u16 vendor, device;
    u8 mask;
    u16 direction_addr, output_addr, input_addr;
    struct pci_dev *dev;
    u8 io_type;	
} *chip_type;

#define ALI_CHIP(x)   { #x, PCI_VENDOR_ID_AL, PCI_DEVICE_ID_AL_ ## x, 0x08, 0x7d, 0x7e, 0x7f, NULL, IO_TYPE_PCI }
#define SERVERWORKS_CHIP(x)   { #x, PCI_VENDOR_ID_SERVERWORKS, PCI_DEVICE_ID_SERVERWORKS_ ## x, 0x01, 0x01, 0x0, 0x0, NULL, IO_TYPE_IO }
static struct _chip_type chip_types[] = {
    ALI_CHIP(M7101),
    SERVERWORKS_CHIP(OSB4)
};


#if LINUX_VERSION_CODE < 131584
static int hostid_lseek(struct inode *inode, struct file *file, off_t offset, int orig)
{
#else
static loff_t hostid_lseek(struct file *file, loff_t offset, int orig)
{
    struct inode * inode = file->f_dentry->d_inode;
#endif
    struct proc_dir_entry *pde = (struct proc_dir_entry *) inode->u.generic_ip;

    switch (orig) {
    case 0:
	file->f_pos = offset;
	return file->f_pos;
    case 1:
	file->f_pos += offset;
	return file->f_pos;
    case 2:
	file->f_pos = offset + pde->size;
	return file->f_pos;
    default:
	return -EINVAL;
    }
}

#if LINUX_VERSION_CODE < 131584
static int hostid_read(struct inode *inode, struct file *file, char *buf, int count)
{
#else
static ssize_t hostid_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
    struct inode * inode = file->f_dentry->d_inode;
#endif
    struct proc_dir_entry *pde = (struct proc_dir_entry *) inode->u.generic_ip;
    int orig_count;

    for (orig_count = count; count > 0; count--, buf++) {
#if LINUX_VERSION_CODE < 131584
	int loc = file->f_pos;
#else
	int loc = *ppos;
#endif

	if (loc < 0)
	    break;

	if (loc >= pde->size)
	    break;

	put_user(((unsigned char *)pde->data)[loc], buf);
#if LINUX_VERSION_CODE < 131584
	file->f_pos++;
#else
	(*ppos)++;
#endif
    }

    return orig_count - count;
}

static int hostid_open(struct inode *inode, struct file *file)
{
    MOD_INC_USE_COUNT;
    return 0;
}

#if LINUX_VERSION_CODE < 131584
static void hostid_release(struct inode *inode, struct file *file)
#else
static int hostid_release(struct inode *inode, struct file *file)
#endif
{
    MOD_DEC_USE_COUNT;

#if LINUX_VERSION_CODE < 131584
#else
    return 0;
#endif
}

static struct file_operations hostid_operations = {
        hostid_lseek,          /* lseek */
        hostid_read,	        /* read */
        NULL,             	/* write */
        NULL,                   /* readdir */
        NULL,          		/* select */
        NULL,                   /* ioctl */
        NULL,                   /* mmap */
        hostid_open,           /* open */
#if LINUX_VERSION_CODE >= 131584
        NULL,                   /* flush */
#endif
        hostid_release,        /* release */
        NULL                    /* can't fsync */
};

static struct inode_operations hostid_inode_operations = {
	&hostid_operations
};

static struct proc_dir_entry hostid_entry = {
	0, 6, "hostid", 
	S_IFREG | S_IRUSR, 1, 0, 0,
	sizeof(hostid), &hostid_inode_operations,
        0, 0, 0, 0, 0,
        hostid,
};

static struct proc_dir_entry ssn_entry = {
	0, 12, "serialnumber", 
	S_IFREG | S_IRUSR, 1, 0, 0,
	SSN_SIZE * 2 + 1, &hostid_inode_operations,
        0, 0, 0, 0, 0,
        ssn_string,
};

__initfunc(static int io_init(void))
{
    unsigned char data;
    unsigned long flags;
    u16 addr;

    save_flags(flags);

    switch (chip_type->io_type) {
    case IO_TYPE_PCI:
	    pci_read_config_byte(chip_type->dev, 0x6d, &data);
	    if (debug > 1)
		    printk("serialnumber: read of address 0x%x returned 0x%x\n", 0x6d, data);

	    /* Set input mode on GPIO3 */
	    pci_read_config_byte(chip_type->dev, 
				 chip_type->direction_addr, &data);
	    if (debug > 1)
		    printk("serialnumber: read of address 0x%x returned 0x%x\n", chip_type->direction_addr, data);
	    if (data & chip_type->mask)
		    pci_write_config_byte(chip_type->dev, 
					  chip_type->direction_addr, 
					  data & ~chip_type->mask);
	    
	    /* Set the output value to be 0 */
	    pci_read_config_byte(chip_type->dev, chip_type->output_addr, 
				 &data);
	    if (debug > 1)
		    printk("serialnumber: read of address 0x%x returned 0x%x\n", chip_type->output_addr, data);
	    if (data & chip_type->mask)
		    pci_write_config_byte(chip_type->dev, 
					  chip_type->output_addr,
					  data & ~chip_type->mask);
	    break;
    case IO_TYPE_IO:
	    outb(SUPERIO_LOGICAL_DEV, SUPERIO_INDEX_PORT);
	    outb(SUPERIO_DEV_GPIO, SUPERIO_DATA_PORT);
	    outb(SUPERIO_ADDR_HIGH, SUPERIO_INDEX_PORT);
	    addr = inb(SUPERIO_DATA_PORT) << 8;
	    outb(SUPERIO_ADDR_LOW, SUPERIO_INDEX_PORT);
	    addr |= inb(SUPERIO_DATA_PORT);
	    if (addr) {
		    u8 val;

		    chip_type->direction_addr += addr;
		    chip_type->input_addr += addr;
		    chip_type->output_addr += addr;
		    
		    /* set output value to 0 */
		    val = inb(chip_type->direction_addr);
		    outb(val | chip_type->mask, chip_type->direction_addr);
		    data = inb(chip_type->output_addr);
		    if (data & chip_type->mask)
			    outb(data & ~chip_type->mask, 
				 chip_type->output_addr);
		    /* set to input */
		    outb(val & ~chip_type->mask, chip_type->direction_addr);
	    }
	    break;
    }
	    
    restore_flags(flags);

    /* Let things calm down */
    udelay(500);
    return 0;
}

__initfunc(static void io_write(int delay))
{
    unsigned char data, direction;

    /* Set output mode on GPIO3 */
    switch (chip_type->io_type) {
    case IO_TYPE_PCI:
	    pci_read_config_byte(chip_type->dev, chip_type->direction_addr, 
				 &data);
	    pci_write_config_byte(chip_type->dev, chip_type->direction_addr,
				  data | chip_type->mask);
	    udelay(delay);

	    /* Set input mode */
	    pci_write_config_byte(chip_type->dev, chip_type->direction_addr, 
				  data & ~chip_type->mask);
	    break;
    case IO_TYPE_IO:
	    direction = inb(chip_type->direction_addr); /* change to output */
	    outb(direction | chip_type->mask, chip_type->direction_addr);
	    udelay(delay);
	    outb(direction & ~chip_type->mask, chip_type->direction_addr);
	    break;
    }
}

__initfunc(static int io_read(void))
{
    unsigned char data = 0;

    /* Get the input value */
    switch (chip_type->io_type) {
    case IO_TYPE_PCI:
	    pci_read_config_byte(chip_type->dev, chip_type->input_addr, &data);
	    break;
    case IO_TYPE_IO:
	    data = inb(chip_type->input_addr);
	    break;
    }
    if (data & chip_type->mask)
	    return 1;
    return 0;
}

__initfunc(static void io_write_byte(unsigned char c))
{
    int i;
    unsigned long flags;

    save_flags(flags);

    for (i = 0; i < 8; i++, c >>= 1) {
        cli();
        if (c & 1) {
            /* Transmit a 1 */
            io_write(5);
            udelay(80);
        } else {
            /* Transmit a 0 */
            io_write(80);
            udelay(10);
        }
        restore_flags(flags);
    }
}

__initfunc(static int io_read_byte(void))
{
    int i;
    int c = 0;
    unsigned long flags;

    save_flags(flags);

    for (i = 0; i < 8; i++) {
        cli();
        io_write(1);    /* Start the read */
        udelay(2);
        if (io_read()) {
            c |= 1 << i;
        }
        udelay(60);
        restore_flags(flags);
    }

    return c;
}

__initfunc(static int get_ssn(unsigned char *buf))
{
    int i;
    unsigned long flags;

    save_flags(flags);

    cli();
    /* Master Reset Pulse */
    for (i = 0; i < 600; i += 30) {
        if (io_read())
            break;
    }

    if (i >= 600) {
        if (debug)
            printk("serialnumber: GPIO3 seems to be held low\n");
        return -ENXIO;
    }

    io_write(600);

    for (i = 0; i < 300; i += 15) {
        udelay(15);
        if (io_read() == 0) {
            /* We got a presence pulse */
            udelay(600);        /* Wait for things to quieten down */
            break;
        }
    }
    restore_flags(flags);

    if (i >= 300) {
        if (debug)      
            printk("serialnumber: No presence pulse detected\n");
        return -ENXIO;
    }

    io_write_byte(0x33);

    for (i = 0; i < 8; i++) {
        int rc;

        rc = io_read_byte();
        if (rc < 0)
            return rc;

        *buf++ = rc;
    }

    return 0;
}
 
__initfunc(int hostid_init(void))
{
    char *version = strchr("$Revision: 1.9 $", ':');
    char buff[32];
    char *p;
    char versionbuff[64];
    int rc;
    int i;
    struct _chip_type *ct;

    if (version) {
        strncpy(buff, version + 1, sizeof(buff));
        if ((p = strchr(buff, '$'))) {
            *p = '\0';
        } 
        sprintf(versionbuff, "Version%sinitialized. ", buff);
    } else {
	versionbuff[0] = '\0';
    }

    if (!pci_present()) 
        return -ENODEV;

    for (ct = chip_types; ct < chip_types + sizeof(chip_types) / sizeof(chip_types[0]); ct++) {
        ct->dev = pci_find_device(ct->vendor, ct->device, 0);
        if (ct->dev) {
            chip_type = ct;
            break;
        }
    }

    if (!chip_type) {
        return -ENXIO;
    }

    io_init();

    rc = get_ssn(ssn);
    if (rc)
        return rc;
    
    /* Convert to ssn_string */

    for (i = 7; i >= 0; i--) {
        sprintf(ssn_string + (7 - i) * 2, "%02x\n", ssn[i]);
    }

    memcpy(hostid, ssn + 1, sizeof(hostid));

#ifdef CONFIG_PROC_FS
    proc_register_dynamic(&proc_root, &hostid_entry);
    proc_register_dynamic(&proc_root, &ssn_entry);
#else
#error proc_fs must be enabled
#endif

    printk(KERN_INFO "serialnumber: %sSerial number=%.16s.\n", 
	versionbuff, ssn_string);

    printk(KERN_INFO "Copyright (c)1994-2000 Axent Technologies, Inc.\n");

    (void) rcsid;

    return 0;
}

#ifdef MODULE
int init_module(void)
{
    return hostid_init();
}

void cleanup_module(void)
{
#ifdef CONFIG_PROC_FS
    proc_unregister(&proc_root, hostid_entry.low_ino);
    proc_unregister(&proc_root, ssn_entry.low_ino);
#endif
}
#endif
