/* drivers/net/eepro100.c: An Intel i82557-559 Ethernet driver for Linux. */
/*
   NOTICE: this version of the driver is supposed to work with 2.2 kernels.
	Written 1996-1999 by Donald Becker.

	This software may be used and distributed according to the terms
	of the GNU Public License, incorporated herein by reference.

	This driver is for the Intel EtherExpress Pro100 (Speedo3) design.
	It should work with all i82557/558/559 boards.

	To use as a module, use the compile-command at the end of the file.

	The author may be reached as becker@CESDIS.usra.edu, or C/O
	Center of Excellence in Space Data and Information Sciences
	   Code 930.5, NASA Goddard Space Flight Center, Greenbelt MD 20771
	For updates see
		http://cesdis.gsfc.nasa.gov/linux/drivers/eepro100.html
	For installation instructions
		http://cesdis.gsfc.nasa.gov/linux/misc/modules.html
	There is a Majordomo mailing list based at
		linux-eepro100@cesdis.gsfc.nasa.gov
	
	The driver also contains updates by different kernel developers.
	This driver clone is maintained by Andrey V. Savochkin <saw@saw.sw.com.sg>.
	Please use this email address and linux-kernel mailing list for bug reports.
	
	Modification history:
	2000 Mar 24  Dragan Stancevic <visitor@valinux.com>
		Disabled FC and ER, to avoid lockups when when we get FCP interrupts.
	2000 May 27  Andrey Moruga <moruga@sw.com.sg>
		Code duplication for 82559ER support was removed.
		Accurate handling of all supported chips was implemented.
		Some fixes in 2.3 clone of the driver were ported.
	2000 May 30  Dragan Stancevic <visitor@valinux.com> and
				 Andrey Moruga <moruga@sw.com.sg>
		Honor PortReset timing specification.
*/

static const char *version =
"eepro100.c:v1.09j-t 9/29/99 Donald Becker http://cesdis.gsfc.nasa.gov/linux/drivers/eepro100.html\n"
"eepro100.c: $Revision: 1.19 $ 2000/05/31 Modified by Andrey V. Savochkin <saw@saw.sw.com.sg> and others\n";

/* A few user-configurable values that apply to all boards.
   First set is undocumented and spelled per Intel recommendations. */


static int congenb = 0;		/* Enable congestion control in the DP83840. */
static int txfifo = 8;		/* Tx FIFO threshold in 4 byte units, 0-15 */
static int rxfifo = 8;		/* Rx FIFO threshold, default 32 bytes. */
/* Tx/Rx DMA burst length, 0-127, 0 == no preemption, tx==128 -> disabled. */
static int txdmacount = 128;
static int rxdmacount = 0;

/* Maximum number of multicast addresses to filter (vs. rx-all-multicast) */
static int multicast_filter_limit = 64;

/* 'options' is used to pass a transceiver override or full-duplex flag
   e.g. "options=16" for FD, "options=32" for 100mbps-only. */
static int full_duplex[] = {-1, -1, -1, -1, -1, -1, -1, -1};
static int options[] = {-1, -1, -1, -1, -1, -1, -1, -1};
#ifdef MODULE
static int debug = -1;			/* The debug level */
#endif

/* A few values that may be tweaked. */
/* The ring sizes should be a power of two for efficiency. */
#define TX_RING_SIZE	64
#define RXFD_RING_SIZE	512	/* 100baseT, min pkts, 2048 == 10 ms */
#define RXBD_RING_SIZE	128	/* 100baseT, 1k pkts, 128 == 10 ms */

/* How much slots multicast filter setup may take.
   Do not descrease without changing set_rx_mode() implementaion. */
#define TX_MULTICAST_SIZE   2
#define TX_MULTICAST_RESERV (TX_MULTICAST_SIZE*2)
/* Actual number of TX packets queued, must be
   <= TX_RING_SIZE-TX_MULTICAST_RESERV. */
#define TX_QUEUE_LIMIT  (TX_RING_SIZE-TX_MULTICAST_RESERV)
/* Hysteresis marking queue as no longer full. */
#define TX_QUEUE_UNFULL (TX_QUEUE_LIMIT-4)

/* Operational parameters that usually are not changed. */

/* Time in jiffies before concluding the transmitter is hung. */
#define TX_TIMEOUT		(2*HZ)
/* Size of an pre-allocated Rx buffer: <Ethernet MTU> + slack.*/
#define PKT_BUF_SZ		1600

#if !defined(__OPTIMIZE__)  ||  !defined(__KERNEL__)
#warning  You must compile this file with the correct options!
#warning  See the last lines of the source file.
#error You must compile this driver with "-O".
#endif

#include <linux/version.h>
#include <linux/module.h>
#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <asm/spinlock.h>

#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/irq.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/delay.h>

#if defined(MODULE)
MODULE_AUTHOR("Maintainer: Andrey V. Savochkin <saw@saw.sw.com.sg>");
MODULE_DESCRIPTION("Intel i8255[789] PCI EtherExpressPro driver");
MODULE_PARM(debug, "i");
MODULE_PARM(options, "1-" __MODULE_STRING(8) "i");
MODULE_PARM(full_duplex, "1-" __MODULE_STRING(8) "i");
MODULE_PARM(congenb, "i");
MODULE_PARM(txfifo, "i");
MODULE_PARM(rxfifo, "i");
MODULE_PARM(txdmacount, "i");
MODULE_PARM(rxdmacount, "i");
MODULE_PARM(multicast_filter_limit, "i");
#endif

#define RUN_AT(x) (jiffies + (x))
/* Condensed bus+endian portability operations. */
#define virt_to_le32desc(addr)  cpu_to_le32(virt_to_bus(addr))
#define le32desc_to_virt(addr)  bus_to_virt(le32_to_cpu(addr))

#define net_device              device
#define pci_base_address(p, n)  (p)->base_address[n]

#define dev_free_skb(skb)       dev_kfree_skb(skb);
#define netif_wake_queue(dev)   do { \
									clear_bit(0, (void*)&dev->tbusy); \
									mark_bh(NET_BH); \
								} while(0)
#define netif_start_queue(dev)  clear_bit(0, (void*)&dev->tbusy)
#define netif_stop_queue(dev)   set_bit(0, (void*)&dev->tbusy)
#ifndef PCI_DEVICE_ID_INTEL_82559ER
#define PCI_DEVICE_ID_INTEL_82559ER 0x1209
#endif
#ifndef PCI_DEVICE_ID_INTEL_ID1029
#define PCI_DEVICE_ID_INTEL_ID1029 0x1029
#endif
#ifndef PCI_DEVICE_ID_INTEL_ID1030
#define PCI_DEVICE_ID_INTEL_ID1030 0x1030
#endif

/* The total I/O port extent of the board.
   The registers beyond 0x18 only exist on the i82558. */
#define SPEEDO3_TOTAL_SIZE 0x20

int speedo_debug = 0;

/*
				Theory of Operation

I. Board Compatibility

This device driver is designed for the Intel i82557 "Speedo3" chip, Intel's
single-chip fast Ethernet controller for PCI, as used on the Intel
EtherExpress Pro 100 adapter.

II. Board-specific settings

PCI bus devices are configured by the system at boot time, so no jumpers
need to be set on the board.  The system BIOS should be set to assign the
PCI INTA signal to an otherwise unused system IRQ line.  While it's
possible to share PCI interrupt lines, it negatively impacts performance and
only recent kernels support it.

III. Driver operation

IIIA. General
The Speedo3 is very similar to other Intel network chips, that is to say
"apparently designed on a different planet".  This chips retains the complex
Rx and Tx descriptors and multiple buffers pointers as previous chips, but
also has simplified Tx and Rx buffer modes.  This driver uses the "flexible"
Tx mode, but in a simplified lower-overhead manner: it associates only a
single buffer descriptor with each frame descriptor.

Despite the extra space overhead in each receive skbuff, the driver must use
the simplified Rx buffer mode to assure that only a single data buffer is
associated with each RxFD. The driver implements this by reserving space
for the Rx descriptor at the head of each Rx skbuff.

The Speedo-3 has receive and command unit base addresses that are added to
almost all descriptor pointers.  The driver sets these to zero, so that all
pointer fields are absolute addresses.

The System Control Block (SCB) of some previous Intel chips exists on the
chip in both PCI I/O and memory space.  This driver uses the I/O space
registers, but might switch to memory mapped mode to better support non-x86
processors.

IIIB. Transmit structure

The driver must use the complex Tx command+descriptor mode in order to
have a indirect pointer to the skbuff data section.  Each Tx command block
(TxCB) is associated with two immediately appended Tx Buffer Descriptor
(TxBD).  A fixed ring of these TxCB+TxBD pairs are kept as part of the
speedo_private data structure for each adapter instance.

The newer i82558 explicitly supports this structure, and can read the two
TxBDs in the same PCI burst as the TxCB.

This ring structure is used for all normal transmit packets, but the
transmit packet descriptors aren't long enough for most non-Tx commands such
as CmdConfigure.  This is complicated by the possibility that the chip has
already loaded the link address in the previous descriptor.  So for these
commands we convert the next free descriptor on the ring to a NoOp, and point
that descriptor's link to the complex command.

An additional complexity of these non-transmit commands are that they may be
added asynchronous to the normal transmit queue, so we disable interrupts
whenever the Tx descriptor ring is manipulated.

A notable aspect of these special configure commands is that they do
work with the normal Tx ring entry scavenge method.  The Tx ring scavenge
is done at interrupt time using the 'dirty_tx' index, and checking for the
command-complete bit.  While the setup frames may have the NoOp command on the
Tx ring marked as complete, but not have completed the setup command, this
is not a problem.  The tx_ring entry can be still safely reused, as the
tx_skbuff[] entry is always empty for config_cmd and mc_setup frames.

Commands may have bits set e.g. CmdSuspend in the command word to either
suspend or stop the transmit/command unit.  This driver always flags the last
command with CmdSuspend, erases the CmdSuspend in the previous command, and
then issues a CU_RESUME.
Note: Watch out for the potential race condition here: imagine
	erasing the previous suspend
		the chip processes the previous command
		the chip processes the final command, and suspends
	doing the CU_RESUME
		the chip processes the next-yet-valid post-final-command.
So blindly sending a CU_RESUME is only safe if we do it immediately after
after erasing the previous CmdSuspend, without the possibility of an
intervening delay.  Thus the resume command is always within the
interrupts-disabled region.  This is a timing dependence, but handling this
condition in a timing-independent way would considerably complicate the code.

Note: In previous generation Intel chips, restarting the command unit was a
notoriously slow process.  This is presumably no longer true.

IIIC. Receive structure

Because of the bus-master support on the Speedo3 this driver uses the new
SKBUFF_RX_COPYBREAK scheme, rather than a fixed intermediate receive buffer.
This scheme allocates full-sized skbuffs as receive buffers.  The value
SKBUFF_RX_COPYBREAK is used as the copying breakpoint: it is chosen to
trade-off the memory wasted by passing the full-sized skbuff to the queue
layer for all frames vs. the copying cost of copying a frame to a
correctly-sized skbuff.

For small frames the copying cost is negligible (esp. considering that we
are pre-loading the cache with immediately useful header information), so we
allocate a new, minimally-sized skbuff.  For large frames the copying cost
is non-trivial, and the larger copy might flush the cache of useful data, so
we pass up the skbuff the packet was received into.

IV. Notes

Thanks to Steve Williams of Intel for arranging the non-disclosure agreement
that stated that I could disclose the information.  But I still resent
having to sign an Intel NDA when I'm helping Intel sell their own product!

*/

/* This table drives the PCI probe routines. */
static struct net_device *speedo_found1(struct pci_dev *pdev, int pci_bus, 
										int pci_devfn, long ioaddr, 
										int chip_idx, int card_idx);

/* 
 * FIXME: I put this in as a temporary fix until Cobalt firmware fixes for 
 * MMIO are more prevalent - TH
 */
#ifndef CONFIG_EEXPRESS_PRO100_MMIO
#define USE_IO
#endif

#ifdef USE_IO
#define SPEEDO_IOTYPE   PCI_USES_MASTER|PCI_USES_IO|PCI_ADDR1
#define SPEEDO_SIZE		32
#else
#define SPEEDO_IOTYPE   PCI_USES_MASTER|PCI_USES_MEM|PCI_ADDR0
#define SPEEDO_SIZE		0x1000
#endif

enum pci_flags_bit {
	PCI_USES_IO=1, PCI_USES_MEM=2, PCI_USES_MASTER=4,
	PCI_ADDR0=0x10<<0, PCI_ADDR1=0x10<<1, PCI_ADDR2=0x10<<2, PCI_ADDR3=0x10<<3,
};
struct pci_id_info {
	const char *name;
	u16	vendor_id, device_id;
	int pci_index;
} static pci_tbl[] = {
	{ "Intel PCI EtherExpress Pro100 82557",
	  PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82557,
	  0
	},
	{ "Intel PCI EtherExpress Pro100 82559ER",
	  PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82559ER,
	  0
	},
	{ "Intel PCI EtherExpress Pro100 ID1029",
	  PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ID1029,
	  0 
	},
	{ "Intel Corporation 82559 InBusiness 10/100",
	  PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ID1030,
	  0 
	},
	{0,}						/* 0 terminated list. */
};

static inline unsigned int io_inw(unsigned long port)
{
	return inw(port);
}
static inline void io_outw(unsigned int val, unsigned long port)
{
	outw(val, port);
}

#ifndef USE_IO
#undef inb
#undef inw
#undef inl
#undef outb
#undef outw
#undef outl
#define inb readb
#define inw readw
#define inl readl
#define outb writeb
#define outw writew
#define outl writel
#endif

static void speedo_fasttimer(unsigned long data);
static struct timer_list fasttimer;	/* interupt prod timer. */
int fasttimer_numopens;

#define FASTTIMER_RATE 1
#define SLOWTIMER_RATE (2 * HZ)

/* How to wait for the command unit to accept a command.
   Typically this takes 0 ticks. */
static inline void wait_for_cmd_done(long cmd_ioaddr)
{
	int wait = 10000;
	do   ;
	while(inb(cmd_ioaddr) && --wait >= 0) ;
#ifndef final_version
	if (wait < 0 && speedo_debug > 3)
		printk(KERN_ALERT "eepro100: wait_for_cmd_done timeout!\n");
#endif
}

/* Offsets to the various registers.
   All accesses need not be longword aligned. */
enum speedo_offsets {
	SCBStatus = 0, SCBCmd = 2,	/* Rx/Command Unit command and status. */
	SCBIntmask = 3,
	SCBPointer = 4,				/* General purpose pointer. */
	SCBPort = 8,				/* Misc. commands and operands.  */
	SCBflash = 12, SCBeeprom = 14, /* EEPROM and flash memory control. */
	SCBCtrlMDI = 16,			/* MDI interface control. */
	SCBEarlyRx = 20,			/* Early receive byte count. */
};
/* Commands that can be put in a command list entry. */
enum commands {
	CmdNOp = 0, CmdIASetup = 0x10000, CmdConfigure = 0x20000,
	CmdMulticastList = 0x30000, CmdTx = 0x40000, CmdTDR = 0x50000,
	CmdDump = 0x60000, CmdDiagnose = 0x70000,
	CmdSuspend = 0x40000000,	/* Suspend after completion. */
	CmdIntr = 0x20000000,		/* Interrupt after completion. */
	CmdTxFlex = 0x00080000,		/* Use "Flexible mode" for CmdTx command. */
};
/* Clear CmdSuspend (1<<30) avoiding interference with the card access to the
   status bits.  Previous driver versions used separate 16 bit fields for
   commands and statuses.  --SAW
 */
#if defined(__LITTLE_ENDIAN)
#define clear_suspend(cmd)  ((__u16 *)&(cmd)->cmd_status)[1] &= ~0x4000
#elif defined(__BIG_ENDIAN)
#define clear_suspend(cmd)  ((__u16 *)&(cmd)->cmd_status)[1] &= ~0x0040
#else
#error Unsupported byteorder
#endif

enum SCBCmdBits {
     SCBMaskCmdDone=0x8000, SCBMaskRxDone=0x4000, SCBMaskCmdIdle=0x2000,
     SCBMaskRxSuspend=0x1000, SCBMaskEarlyRx=0x0800, SCBMaskFlowCtl=0x0400,
     SCBTriggerIntr=0x0200, SCBMaskAll=0x0100,
     /* The rest are Rx and Tx commands. */
     CUStart=0x0010, CUResume=0x0020, CUStatsAddr=0x0040, CUShowStats=0x0050,
     CUCmdBase=0x0060,  /* CU Base address (set to zero) . */
     CUDumpStats=0x0070, /* Dump then reset stats counters. */
     RxStart=0x0001, RxResume=0x0002, RxAbort=0x0004, RxAddrLoad=0x0006,
     RxResumeNoResources=0x0007,
};

/*
 * We wish to use the symbollic names for the interrupt mask bits in
 * circumstances where we will do a byte I/O to offset 3, instead of a
 * word access to offset 2 (the entire command register).  This macro
 * make the necessary translation of the enum's above.
 */
#define CMD_TO_MASK(x)	(x >> 8)

enum SCBPort_cmds {
	PortReset=0, PortSelfTest=1, PortPartialReset=2, PortDump=3,
};


/* The Speedo3 Rx and Tx frame/buffer descriptors. */
struct descriptor {			    /* A generic descriptor. */
	s32 cmd_status;				/* All command and status fields. */
	u32 link;				    /* struct descriptor *  */
	unsigned char params[0];
};

#define RXFD_HDR_SIZE 		\
	(256 - (sizeof(s32) + sizeof(u32) + sizeof(u32) + sizeof(u32)))
/* The Speedo3 Rx and Tx buffer descriptors. */
struct RxFD {					/* Receive frame descriptor. */
	s32 status;
	u32 link;					/* struct RxFD * */
	u32 rx_buf_addr;			/* void * */
	u32 count;
	char hdr[RXFD_HDR_SIZE];
};

struct RxBD {
	u32 status_framesz;
	u32 link;
	u32 rx_buf_addr;
	u32 end_bufsz;
};

/* Selected elements of the Tx/RxFD.status word. */
enum RxFD_bits {
	RxComplete=0x8000, RxOK=0x2000,
	RxErrCRC=0x0800, RxErrAlign=0x0400, RxErrTooBig=0x0200, RxErrSymbol=0x0010,
	RxEth2Type=0x0020, RxNoMatch=0x0004, RxNoIAMatch=0x0002,
	TxUnderrun=0x1000,  StatusComplete=0x8000,
};

#define RXF_NEEDSFILL 0x8FFF	/* status field flag when waiting for fill */
#define RBD_NEEDSFILL 0x8FFF	/* status_framesz flag when waiting for fill */

struct TxFD {					/* Transmit frame descriptor set. */
	s32 status;
	u32 link;					/* void * */
	u32 tx_desc_addr;			/* Always points to the tx_buf_addr element. */
	s32 count;					/* # of TBD (=1), Tx start thresh., etc. */
	/* This constitutes two "TBD" entries -- we only use one. */
	u32 tx_buf_addr0;			/* void *, frame to be transmitted.  */
	s32 tx_buf_size0;			/* Length of Tx frame. */
	u32 tx_buf_addr1;			/* void *, frame to be transmitted.  */
	s32 tx_buf_size1;			/* Length of Tx frame. */
};

/* Multicast filter setting block.  --SAW */
struct speedo_mc_block {
	struct speedo_mc_block *next;
	unsigned int tx;
	struct descriptor frame __attribute__ ((__aligned__(16)));
};

/* Elements of the dump_statistics block. This block must be lword aligned. */
struct speedo_stats {
	u32 tx_good_frames;
	u32 tx_coll16_errs;
	u32 tx_late_colls;
	u32 tx_underruns;
	u32 tx_lost_carrier;
	u32 tx_deferred;
	u32 tx_one_colls;
	u32 tx_multi_colls;
	u32 tx_total_colls;
	u32 rx_good_frames;
	u32 rx_crc_errs;
	u32 rx_align_errs;
	u32 rx_resource_errs;
	u32 rx_overrun_errs;
	u32 rx_colls_errs;
	u32 rx_runt_errs;
	u32 done_marker;
};

#define RXFD_HDR_SIZE_INIT cpu_to_le32(RXFD_HDR_SIZE << 16)

#define RXF_DESC_NONSTOP	cpu_to_le32(0x00080001)
#define RXF_DESC_STOP		cpu_to_le32(0xC0080002)

#define RXB_DESC_NONSTOP_MTULEN		\
			cpu_to_le32(0x0000 | (PKT_BUF_SZ - RXFD_HDR_SIZE))
#define RXB_DESC_STOP_MTULEN		\
			cpu_to_le32(0x8000 | (PKT_BUF_SZ - RXFD_HDR_SIZE))
#define RXB_DESC_STOP_NOLEN	cpu_to_le32(0x8000)

enum Rx_ring_state_bits {
	RrNoMem=1, RrPostponed=2, RrNoResources=4, RrOOMReported=8,
};

struct net_device *speedo_open_list;

/* Do not change the position (alignment) of the first few elements!
   The later elements are grouped for cache locality. */
struct speedo_private {
	struct TxFD	tx_ring[TX_RING_SIZE];	/* Commands and CmdTxPacket */
	struct RxFD *rx_ringp[RXFD_RING_SIZE];	/* Ring of Rx descriptor */
	struct sk_buff *rx_ringskb[RXFD_RING_SIZE];
	struct RxBD rx_buffs[RXBD_RING_SIZE];	/* Rx buffer descriptors */
	/* The addresses of a Tx/Rx-in-place packets/buffers. */
	struct sk_buff* tx_skbuff[TX_RING_SIZE];
	struct sk_buff* rx_skbuff[RXBD_RING_SIZE];
	struct descriptor  *last_cmd;	/* Last command sent. */
	unsigned int cur_tx, dirty_tx;	/* The ring entries to be free()ed. */
	spinlock_t lock;				/* Group with Tx control cache line. */
	u32 tx_threshold;					/* The value for txdesc.count. */
	unsigned int suspend_rxbd;		/* last Rx index with suspend */
	unsigned int cur_rxbd, dirty_rxbd;	/* The next free ring entry */
	unsigned int suspend_rxfd;		/* last Rx index with suspend */
	unsigned int cur_rxfd, dirty_rxfd;	/* The next free ring entry */
	unsigned long last_rx_time;	/* Last Rx, in jiffies, to handle Rx hang. */
	int resetting;
	const char *product_name;
	struct net_device *next_module;
	struct net_device *open_list;
	void *priv_addr;					/* Unaligned address for kfree */
	struct enet_statistics stats;
	struct speedo_stats lstats;
	int chip_id;
	unsigned char pci_bus, pci_devfn, acpi_pwr;
	struct timer_list timer;	/* Media selection timer. */

	struct speedo_mc_block *mc_setup_head;/* Multicast setup frame list head. */
	struct speedo_mc_block *mc_setup_tail;/* Multicast setup frame list tail. */
	int in_interrupt;					/* Word-aligned dev->interrupt */
	int intmask;
	char rx_mode;						/* Current PROMISC/ALLMULTI setting. */
	int rx_loopcnt;
	unsigned int tx_full:1;				/* The Tx queue is full. */
	unsigned int full_duplex:1;			/* Full-duplex operation requested. */
	unsigned int flow_ctrl:1;			/* Use 802.3x flow control. */
	unsigned int rx_bug:1;				/* Work around receiver hang errata. */
	unsigned int rx_bug10:1;			/* Receiver might hang at 10mbps. */
	unsigned int rx_bug100:1;			/* Receiver might hang at 100mbps. */
	unsigned int irq_shadow_slave:1;
	unsigned int irq_shadow_master:1;
	unsigned char default_port:8;		/* Last dev->if_port value. */
	unsigned char rx_ring_state;		/* RX ring status flags. */
	unsigned short phy[2];				/* PHY media interfaces available. */
	unsigned short advertising;			/* Current PHY advertised caps. */
	unsigned short partner;				/* Link partner caps. */
	unsigned int fasttimer_cnt;
};

/* The parameters for a CmdConfigure operation.
   There are so many options that it would be difficult to document each bit.
   We mostly use the default or recommended settings. */
const char i82557_config_cmd[22] = {
	22, 0x08, 0, 0,  0, 0, 0x32, 0x03,  1, /* 1=Use MII  0=Use AUI */
	0, 0x2E, 0,  0x60, 0,
	0xf2, 0x48,   0, 0x40, 0xf2, 0x80, 		/* 0x40=Force full-duplex */
	0x3f, 0x05, };
const char i82558_config_cmd[22] = {
	22, 0x08, 0, 1,  0, 0, 0x22, 0x03,  1, /* 1=Use MII  0=Use AUI */
	0, 0x2E, 0,  0x60, 0x08, 0x88,
	0x68, 0, 0x40, 0xf2, 0x84,		/* Disable FC */
	0x31, 0x05, };

/* PHY media interface chips. */
static const char *phys[] = {
	"None", "i82553-A/B", "i82553-C", "i82503",
	"DP83840", "80c240", "80c24", "i82555",
	"unknown-8", "unknown-9", "DP83840A", "unknown-11",
	"unknown-12", "unknown-13", "unknown-14", "unknown-15", };
enum phy_chips { NonSuchPhy=0, I82553AB, I82553C, I82503, DP83840, S80C240,
					 S80C24, I82555, DP83840A=10, };
static const char is_mii[] = { 0, 1, 1, 0, 1, 1, 0, 1 };
#define EE_READ_CMD		(6)

static int do_eeprom_cmd(long ioaddr, int cmd, int cmd_len);
static int mdio_read(long ioaddr, int phy_id, int location);
static int mdio_write(long ioaddr, int phy_id, int location, int value);
static int speedo_open(struct net_device *dev);
static void speedo_resume(struct net_device *dev);
static void speedo_timer(unsigned long data);
static void speedo_init_rx_ring(struct net_device *dev);
#if 0 /* debug */
static void speedo_check_rx_ring(struct net_device *dev);
#endif
static void speedo_tx_timeout(struct net_device *dev);
static int speedo_start_xmit(struct sk_buff *skb, struct net_device *dev);
static void speedo_refill_rx_buffers(struct net_device *dev);
static int speedo_rx(struct net_device *dev);
static int speedo_tx_buffer_gc(struct net_device *dev);
static void speedo_interrupt(int irq, void *dev_instance, struct pt_regs *regs);
static int speedo_close(struct net_device *dev);
static struct enet_statistics *speedo_get_stats(struct net_device *dev);
static int speedo_ioctl(struct net_device *dev, struct ifreq *rq, int cmd);
static void set_rx_mode(struct net_device *dev);
static void speedo_show_state(struct net_device *dev);
static void speedo_reset(struct net_device *dev);
static void speedo_freebufs(struct net_device *dev);

#ifdef CONFIG_X86_REMOTE_DEBUG
extern void breakpoint(void);
#endif

#if 0
#define SPEEDO_CHECK_RX_RING speedo_check_rx_ring
#else
#define SPEEDO_CHECK_RX_RING(t)
#endif



#ifdef honor_default_port
/* Optional driver feature to allow forcing the transceiver setting.
   Not recommended. */
static int mii_ctrl[8] = { 0x3300, 0x3100, 0x0000, 0x0100,
						   0x2000, 0x2100, 0x0400, 0x3100};
#endif

/* A list of all installed Speedo devices, for removing the driver module. */
static struct net_device *root_speedo_dev = NULL;

/* helper function to find relative first device in PCI space */
static int
find_first_pci(struct pci_dev *list[], int maxent)
{
	int i;
	int mindex = -1;

	for (i=0; i<maxent; i++) {
		if (!list[i] || i == mindex) {
			continue;
		}
           
		if (mindex < 0) {
			mindex = i;
			continue;
		}

		if (list[i]->bus->number < list[mindex]->bus->number) {
			mindex = i;
		} else if (list[i]->bus->number == list[mindex]->bus->number) {
			if (list[i]->devfn < list[mindex]->devfn) {
				mindex = i;
			}
		}
	}

	return mindex;
}
             

#define MAX_SPEEDOS    8

int eepro100_init(void)
{
	int cards_found = 0;
	int chip_idx;
	struct pci_dev *pdev;
	struct pci_dev *dev_list[MAX_SPEEDOS];
	int idx_list[MAX_SPEEDOS];
	static int list_idx = 0;
	int found_any = 0;
	unsigned char pci_bus, pci_device_fn;

	if (! pcibios_present())
		return cards_found;

	/* clear out the list */
	memset(&dev_list[0], 0, sizeof(void *) * MAX_SPEEDOS);
	memset(&idx_list[0], 0, sizeof(int) * MAX_SPEEDOS);

	/* search for all known devices using this driver */
	for (chip_idx = 0; pci_tbl[chip_idx].name; chip_idx++) {
		for (; pci_tbl[chip_idx].pci_index < 8; pci_tbl[chip_idx].pci_index++) {
			if (list_idx < MAX_SPEEDOS 
			 && !pcibios_find_device(pci_tbl[chip_idx].vendor_id,
									pci_tbl[chip_idx].device_id,
									pci_tbl[chip_idx].pci_index, &pci_bus,
									&pci_device_fn))
			{
				idx_list[list_idx] = chip_idx;
				dev_list[list_idx] = pci_find_slot(pci_bus, pci_device_fn);
				list_idx++;
			  found_any++;
			}
		}
	}
	/* close up early if we didn't find any */
	if (!found_any) {
		return cards_found;
	}

	while (1) {
		int i;
		unsigned char pci_latency;
		unsigned long pciaddr;
		long ioaddr;
		int irq;
		u16 pci_command, new_command;
		
		i = find_first_pci(dev_list, MAX_SPEEDOS);
		if (i < 0 || !dev_list[i]) {
			break;
		}

		pdev = dev_list[i];
		chip_idx = idx_list[i];

#ifdef USE_IO
		pciaddr = pci_base_address(pdev, 1);	/* Use [0] to mem-map */
#else
		pciaddr = pci_base_address(pdev, 0);
#endif
		irq = pdev->irq;
		pci_bus = pdev->bus->number;
		pci_device_fn = pdev->devfn;

		dev_list[i] = NULL;

		/* Remove I/O space marker in bit 0. */
		if (pciaddr & 1) {
			ioaddr = pciaddr & ~3UL;
			if (check_region(ioaddr, 32))
				continue;
		} else if ((ioaddr = (long)ioremap(pciaddr & ~0xfUL, 0x1000)) == 0) {
			printk(KERN_INFO "Failed to map PCI address %#lx.\n",
				   pciaddr);
			continue;
		}
		if (speedo_debug > 2)
			printk("Found Intel i8255[789] PCI Speedo at I/O %#lx, IRQ %d.\n",
				   ioaddr, irq);

		/* Get and check the bus-master and latency values. */
		pcibios_read_config_word(pci_bus, pci_device_fn,
								 PCI_COMMAND, &pci_command);
#ifdef USE_IO
		new_command = pci_command | PCI_COMMAND_MASTER|PCI_COMMAND_IO;
#else
		new_command = pci_command | PCI_COMMAND_MASTER|PCI_COMMAND_IO
			|PCI_COMMAND_MEMORY;
#endif
		if (pci_command != new_command) {
			printk(KERN_INFO "  The PCI BIOS has not enabled this"
				   " device!  Updating PCI command %4.4x->%4.4x.\n",
				   pci_command, new_command);
			pcibios_write_config_word(pci_bus, pci_device_fn,
									  PCI_COMMAND, new_command);
		}
		pcibios_read_config_byte(pci_bus, pci_device_fn,
								 PCI_LATENCY_TIMER, &pci_latency);
		if (pci_latency < 32) {
			printk("  PCI latency timer (CFLT) is unreasonably low at %d."
				   "  Setting to 32 clocks.\n", pci_latency);
			pcibios_write_config_byte(pci_bus, pci_device_fn,
									  PCI_LATENCY_TIMER, 32);
		} else if (speedo_debug > 1)
			printk("  PCI latency timer (CFLT) is %#x.\n", pci_latency);

		if (speedo_found1(pdev, pci_bus, pci_device_fn, ioaddr, chip_idx, cards_found))
			cards_found++;
	}

	return cards_found;
}

static struct net_device *speedo_found1(struct pci_dev *pdev, int pci_bus, 
										int pci_devfn, long ioaddr, 
										int chip_idx, int card_idx)
{
	struct net_device *dev;
	struct speedo_private *sp;
	const char *product;
	int i, option;
	u16 eeprom[0x100];
	int acpi_idle_state = 0;
#ifndef MODULE
	static int did_version = 0;			/* Already printed version info. */
	if (speedo_debug > 0  &&  did_version++ == 0)
		printk(version);
#endif

	dev = init_etherdev(NULL, sizeof(*sp));

	if (dev->mem_start > 0)
		option = dev->mem_start;
	else if (card_idx >= 0  &&  options[card_idx] >= 0)
		option = options[card_idx];
	else
		option = 0;

	/* Read the station address EEPROM before doing the reset.
	   Nominally his should even be done before accepting the device, but
	   then we wouldn't have a device name with which to report the error.
	   The size test is for 6 bit vs. 8 bit address serial EEPROMs.
	*/
	{
		unsigned long iobase;
		int read_cmd, ee_size;
		u16 sum;
		int j;

		/* Use IO only to avoid postponed writes and satisfy EEPROM timing
		   requirements. */
		iobase = pci_base_address(pdev, 1) & ~3UL;
		if ((do_eeprom_cmd(iobase, EE_READ_CMD << 24, 27) & 0xffe0000)
			== 0xffe0000) {
			ee_size = 0x100;
			read_cmd = EE_READ_CMD << 24;
		} else {
			ee_size = 0x40;
			read_cmd = EE_READ_CMD << 22;
		}

		for (j = 0, i = 0, sum = 0; i < ee_size; i++) {
			u16 value = do_eeprom_cmd(iobase, read_cmd | (i << 16), 27);
			eeprom[i] = value;
			sum += value;
			if (i < 3) {
				dev->dev_addr[j++] = value;
				dev->dev_addr[j++] = value >> 8;
			}
		}
		if (sum != 0xBABA)
			printk(KERN_WARNING "%s: Invalid EEPROM checksum %#4.4x, "
				   "check settings before activating this device!\n",
				   dev->name, sum);
		/* Don't  unregister_netdev(dev);  as the EEPro may actually be
		   usable, especially if the MAC address is set later.
		   On the other hand, it may be unusable if MDI data is corrupted. */
	}

	/* Reset the chip: stop Tx and Rx processes and clear counters.
	   This takes less than 10usec and will easily finish before the next
	   action. */
	outl(PortReset, ioaddr + SCBPort);
	/* Honor PortReset timing. */
	udelay(10);

	if (eeprom[3] & 0x0100)
		product = "OEM i8255[789] 10/100 Ethernet";
	else if (pci_tbl[chip_idx].name)
		product = pci_tbl[chip_idx].name;
	else 
		product = "Unknown i8255[789] 10/100 Ethernet";

	printk(KERN_INFO "%s: %s, ", dev->name, product);

	for (i = 0; i < 5; i++)
		printk("%2.2X:", dev->dev_addr[i]);
	printk("%2.2X, ", dev->dev_addr[i]);
#ifdef USE_IO
	printk("I/O at %#3lx, ", ioaddr);
#else
	printk("Memory at %#lx, ", ioaddr);
#endif
	printk("IRQ %d.\n", pdev->irq);

#if 1 || defined(kernel_bloat)
	/* OK, this is pure kernel bloat.  I don't like it when other drivers
	   waste non-pageable kernel space to emit similar messages, but I need
	   them for bug reports. */
	{
		const char *connectors[] = {" RJ45", " BNC", " AUI", " MII"};
		/* The self-test results must be paragraph aligned. */
		s32 str[6], *volatile self_test_results;
		int boguscnt = 16000;	/* Timeout for set-test. */
		if (eeprom[3] & 0x03)
			printk(KERN_INFO "  Receiver lock-up bug exists -- enabling"
				   " work-around.\n");
		printk(KERN_INFO "  Board assembly %4.4x%2.2x-%3.3d, Physical"
			   " connectors present:",
			   eeprom[8], eeprom[9]>>8, eeprom[9] & 0xff);
		for (i = 0; i < 4; i++)
			if (eeprom[5] & (1<<i))
				printk(connectors[i]);
		printk("\n"KERN_INFO"  Primary interface chip %s PHY #%d.\n",
			   phys[(eeprom[6]>>8)&15], eeprom[6] & 0x1f);
		if (eeprom[7] & 0x0700)
			printk(KERN_INFO "    Secondary interface chip %s.\n",
				   phys[(eeprom[7]>>8)&7]);
		if (((eeprom[6]>>8) & 0x3f) == DP83840
			||  ((eeprom[6]>>8) & 0x3f) == DP83840A) {
			int mdi_reg23 = mdio_read(ioaddr, eeprom[6] & 0x1f, 23) | 0x0422;
			if (congenb)
			  mdi_reg23 |= 0x0100;
			printk(KERN_INFO"  DP83840 specific setup, setting register 23 to %4.4x.\n",
				   mdi_reg23);
			mdio_write(ioaddr, eeprom[6] & 0x1f, 23, mdi_reg23);
		}
		if ((option >= 0) && (option & 0x70)) {
			printk(KERN_INFO "  Forcing %dMbs %s-duplex operation.\n",
				   (option & 0x20 ? 100 : 10),
				   (option & 0x10 ? "full" : "half"));
			mdio_write(ioaddr, eeprom[6] & 0x1f, 0,
					   ((option & 0x20) ? 0x2000 : 0) | 	/* 100mbps? */
					   ((option & 0x10) ? 0x0100 : 0)); /* Full duplex? */
		}

		/* Perform a system self-test. */
		self_test_results = (s32*) ((((long) str) + 15) & ~0xf);
		self_test_results[0] = 0;
		self_test_results[1] = -1;
		outl(virt_to_bus(self_test_results) | PortSelfTest, ioaddr + SCBPort);
		do {
			udelay(10);
		} while (self_test_results[1] == -1  &&  --boguscnt >= 0);

		if (boguscnt < 0) {		/* Test optimized out. */
			printk(KERN_ERR "Self test failed, status %8.8x:\n"
				   KERN_ERR " Failure to initialize the i8255[789].\n"
				   KERN_ERR " Verify that the card is a bus-master"
				   " capable slot.\n",
				   self_test_results[1]);
		} else
			printk(KERN_INFO "  General self-test: %s.\n"
				   KERN_INFO "  Serial sub-system self-test: %s.\n"
				   KERN_INFO "  Internal registers self-test: %s.\n"
				   KERN_INFO "  ROM checksum self-test: %s (%#8.8x).\n",
				   self_test_results[1] & 0x1000 ? "failed" : "passed",
				   self_test_results[1] & 0x0020 ? "failed" : "passed",
				   self_test_results[1] & 0x0008 ? "failed" : "passed",
				   self_test_results[1] & 0x0004 ? "failed" : "passed",
				   self_test_results[0]);
	}
#endif  /* kernel_bloat */

	outl(PortReset, ioaddr + SCBPort);
	/* Honor PortReset timing. */
	udelay(10);

	/* We do a request_region() only to register /proc/ioports info. */
	request_region(ioaddr, SPEEDO3_TOTAL_SIZE, "Intel Speedo3 Ethernet");

	dev->base_addr = ioaddr;
	dev->irq = pdev->irq;

	sp = dev->priv;
	if (dev->priv == NULL) {
		void *mem = kmalloc(sizeof(*sp), GFP_ATOMIC);

		if ( ! mem) {
		    printk("%s: kmalloc %d failed\n", dev->name, sizeof(*sp));
		    return 0;
		}
		dev->priv = sp = mem;		/* Cache align here if kmalloc does not. */
		sp->priv_addr = mem;
	}
	memset(sp, 0, sizeof(*sp));
	sp->next_module = root_speedo_dev;
	root_speedo_dev = dev;

	sp->pci_bus = pci_bus;
	sp->pci_devfn = pci_devfn;
	sp->chip_id = chip_idx;
	sp->acpi_pwr = acpi_idle_state;

	sp->full_duplex = option >= 0 && (option & 0x10) ? 1 : 0;
	if (card_idx >= 0) {
		if (full_duplex[card_idx] >= 0)
			sp->full_duplex = full_duplex[card_idx];
	}
	sp->default_port = option >= 0 ? (option & 0x0f) : 0;

	sp->phy[0] = eeprom[6];
	sp->phy[1] = eeprom[7];
	sp->rx_bug = (eeprom[3] & 0x03) == 3 ? 0 : 1;

	if (sp->rx_bug)
		printk(KERN_INFO "  Receiver lock-up workaround activated.\n");

	/* The Speedo-specific entries in the device structure. */
	dev->open = &speedo_open;
	dev->hard_start_xmit = &speedo_start_xmit;
#if defined(HAS_NETIF_QUEUE)
	dev->tx_timeout = &speedo_tx_timeout;
	dev->watchdog_timeo = TX_TIMEOUT;
#endif
	dev->stop = &speedo_close;
	dev->get_stats = &speedo_get_stats;
	dev->set_multicast_list = &set_rx_mode;
	dev->do_ioctl = &speedo_ioctl;

	return dev;
}

/* Serial EEPROM section.
   A "bit" grungy, but we work our way through bit-by-bit :->. */
/*  EEPROM_Ctrl bits. */
#define EE_SHIFT_CLK	0x01	/* EEPROM shift clock. */
#define EE_CS			0x02	/* EEPROM chip select. */
#define EE_DATA_WRITE	0x04	/* EEPROM chip data in. */
#define EE_DATA_READ	0x08	/* EEPROM chip data out. */
#define EE_ENB			(0x4800 | EE_CS)
#define EE_WRITE_0		0x4802
#define EE_WRITE_1		0x4806
#define EE_OFFSET		SCBeeprom

/* The fixes for the code were kindly provided by Dragan Stancevic
   <visitor@valinux.com> to strictly follow Intel specifications of EEPROM
   access timing.
   The publicly available sheet 64486302 (sec. 3.1) specifies 1us access
   interval for serial EEPROM.  However, it looks like that there is an
   additional requirement dictating larger udelay's in the code below.
   2000/05/24  SAW */
static int do_eeprom_cmd(long ioaddr, int cmd, int cmd_len)
{
	unsigned retval = 0;
	long ee_addr = ioaddr + SCBeeprom;

	io_outw(EE_ENB, ee_addr); udelay(2);
	io_outw(EE_ENB | EE_SHIFT_CLK, ee_addr); udelay(2);

	/* Shift the command bits out. */
	do {
		short dataval = (cmd & (1 << cmd_len)) ? EE_WRITE_1 : EE_WRITE_0;
		io_outw(dataval, ee_addr); udelay(2);
		io_outw(dataval | EE_SHIFT_CLK, ee_addr); udelay(2);
		retval = (retval << 1) | ((io_inw(ee_addr) & EE_DATA_READ) ? 1 : 0);
	} while (--cmd_len >= 0);
	io_outw(EE_ENB, ee_addr); udelay(2);

	/* Terminate the EEPROM access. */
	io_outw(EE_ENB & ~EE_CS, ee_addr);
	return retval;
}

static int mdio_read(long ioaddr, int phy_id, int location)
{
	int val, boguscnt = 64*10;		/* <64 usec. to complete, typ 27 ticks */
	outl(0x08000000 | (location<<16) | (phy_id<<21), ioaddr + SCBCtrlMDI);
	do {
		val = inl(ioaddr + SCBCtrlMDI);
		if (--boguscnt < 0) {
			printk(KERN_ERR " mdio_read() timed out with val = %8.8x.\n", val);
			break;
		}
	} while (! (val & 0x10000000));
	return val & 0xffff;
}

static int mdio_write(long ioaddr, int phy_id, int location, int value)
{
	int val, boguscnt = 64*10;		/* <64 usec. to complete, typ 27 ticks */
	outl(0x04000000 | (location<<16) | (phy_id<<21) | value,
		 ioaddr + SCBCtrlMDI);
	do {
		val = inl(ioaddr + SCBCtrlMDI);
		if (--boguscnt < 0) {
			printk(KERN_ERR" mdio_write() timed out with val = %8.8x.\n", val);
			break;
		}
	} while (! (val & 0x10000000));
	return val & 0xffff;
}



static int
speedo_open(struct net_device *dev)
{
	struct speedo_private *sp = (struct speedo_private *)dev->priv;
	long ioaddr = dev->base_addr;
	int irqflag;

	if (speedo_debug > 1)
		printk(KERN_DEBUG "%s: speedo_open() irq %d.\n", dev->name, dev->irq);

	MOD_INC_USE_COUNT;

	/* Set up the Tx queue early.. */
	sp->cur_tx = 0;
	sp->dirty_tx = 0;
	sp->last_cmd = 0;
	sp->tx_full = 0;
	sp->lock = (spinlock_t) SPIN_LOCK_UNLOCKED;
	sp->in_interrupt = 0;

	sp->open_list = speedo_open_list;
	sp->last_rx_time = jiffies;
	irqflag = SA_SHIRQ;
	speedo_open_list = dev;

	/*
	 * If this is not the first device, join the existing
	 * IRQ group.
	 */
	if (sp->open_list) {
		struct net_device *tdev;
		struct speedo_private *tsp;

		/*
		 * If there is another eepro on this IRQ, tell the
		 * interrupt layer not to bother calling us both
		 * on every interrupt.
		 */
		for (tdev = speedo_open_list; tdev; tdev = tsp->open_list) {
			tsp = (struct speedo_private *) tdev->priv;

			if (tdev != dev && tdev->irq == dev->irq) {
				sp->irq_shadow_slave = 1;
				break;
			}
		}
	}

	if ( ! sp->irq_shadow_slave)
	    sp->irq_shadow_master = 1;

	/* .. we can safely take handler calls during init. */
	if (request_irq_class(dev->irq, &speedo_interrupt, irqflag, dev->name,
	    dev, "splnet")) {
		struct net_device **tdevpp;

		for (tdevpp = &speedo_open_list; *tdevpp && *tdevpp != dev; ) {
			tdevpp = &(((struct speedo_private *)(*tdevpp)
				    ->priv)->open_list);
		}
		if (*tdevpp) 
			*tdevpp = ((struct speedo_private *)(*tdevpp)
				    ->priv)->open_list;

		MOD_DEC_USE_COUNT;
		return -EAGAIN;
	}

	dev->if_port = sp->default_port;

#ifdef oh_no_you_dont_unless_you_honour_the_options_passed_in_to_us
	/* Retrigger negotiation to reset previous errors. */
	if ((sp->phy[0] & 0x8000) == 0) {
		int phy_addr = sp->phy[0] & 0x1f ;
		/* Use 0x3300 for restarting NWay, other values to force xcvr:
		   0x0000 10-HD
		   0x0100 10-FD
		   0x2000 100-HD
		   0x2100 100-FD
		*/
#ifdef honor_default_port
		mdio_write(ioaddr, phy_addr, 0, mii_ctrl[dev->default_port & 7]);
#else
		mdio_write(ioaddr, phy_addr, 0, 0x3300);
#endif
	}
#endif

	speedo_init_rx_ring(dev);

	/* Fire up the hardware. */
	outw(SCBMaskAll, ioaddr + SCBCmd);
	speedo_resume(dev);

	dev->interrupt = 0;
	dev->start = 1;
	netif_start_queue(dev);

	/* Setup the chip and configure the multicast list. */
	sp->mc_setup_head = NULL;
	sp->mc_setup_tail = NULL;
	sp->flow_ctrl = sp->partner = 0;
	sp->rx_mode = -1;			/* Invalid -> always reset the mode. */
	set_rx_mode(dev);
	if ((sp->phy[0] & 0x8000) == 0)
		sp->advertising = mdio_read(ioaddr, sp->phy[0] & 0x1f, 4);

	if (speedo_debug > 2) {
		printk(KERN_DEBUG "%s: Done speedo_open(), status %8.8x.\n",
			   dev->name, inw(ioaddr + SCBStatus));
	}

	/* Set the timer.  The timer serves a dual purpose:
	   1) to monitor the media interface (e.g. link beat) and perhaps switch
	   to an alternate media type
	   2) to monitor Rx activity, and restart the Rx process if the receiver
	   hangs. */
	init_timer(&sp->timer);
	sp->timer.expires = RUN_AT(SLOWTIMER_RATE);
	sp->timer.data = (unsigned long)dev;
	sp->timer.function = &speedo_timer;					/* timer handler */
	add_timer(&sp->timer);

	/*
	 * We run a faster timer to babysit the interrupt routine under
	 * heavy load the chip seems to get unhappy and stops generating
	 * interrupts (even setting the S/W interrupt flag fails).
	 */
	if ( ! fasttimer_numopens) {
		init_timer(&fasttimer);
		fasttimer.expires = RUN_AT(FASTTIMER_RATE);
		fasttimer.data = (unsigned long)42;
		fasttimer.function = &speedo_fasttimer;
		add_timer(&fasttimer);

	}
	fasttimer_numopens++;;

	/* No need to wait for the command unit to accept here. */
	if ((sp->phy[0] & 0x8000) == 0)
		mdio_read(ioaddr, sp->phy[0] & 0x1f, 0);

	return 0;
}

/* Start the chip hardware after a full reset. */
static void speedo_resume(struct net_device *dev)
{
	struct speedo_private *sp = (struct speedo_private *)dev->priv;
	long ioaddr = dev->base_addr;

	/* Start with a Tx threshold of 256 (0x..20.... 8 byte units). */
	sp->tx_threshold = 0x01208000;

	/* Set the segment registers to '0'. */
	wait_for_cmd_done(ioaddr + SCBCmd);
	outl(0, ioaddr + SCBPointer);
	outb(RxAddrLoad, ioaddr + SCBCmd);
	wait_for_cmd_done(ioaddr + SCBCmd);
	outb(CUCmdBase, ioaddr + SCBCmd);
	wait_for_cmd_done(ioaddr + SCBCmd);

	/* Load the statistics block and rx ring addresses. */
	outl(virt_to_bus(&sp->lstats), ioaddr + SCBPointer);
	outb(CUStatsAddr, ioaddr + SCBCmd);
	sp->lstats.done_marker = 0;
	wait_for_cmd_done(ioaddr + SCBCmd);
	dev->trans_start = jiffies;

	sp->rx_ringp[sp->cur_rxfd % RXFD_RING_SIZE]->rx_buf_addr =
		virt_to_le32desc(&sp->rx_buffs[sp->cur_rxbd % RXBD_RING_SIZE]);
	outl(virt_to_bus(sp->rx_ringp[sp->cur_rxfd % RXFD_RING_SIZE]),
		ioaddr + SCBPointer);
	outb(RxStart, ioaddr + SCBCmd);
	wait_for_cmd_done(ioaddr + SCBCmd);

	outb(CUDumpStats, ioaddr + SCBCmd);

	/* Fill the first command with our physical address. */
	{
		struct descriptor *ias_cmd;

		ias_cmd =
			(struct descriptor *)&sp->tx_ring[sp->cur_tx++ % TX_RING_SIZE];
		/* Avoid a bug(?!) here by marking the command already completed. */
		ias_cmd->cmd_status = cpu_to_le32((CmdSuspend | CmdIASetup) | 0xa000);
		ias_cmd->link =
			virt_to_le32desc(&sp->tx_ring[sp->cur_tx % TX_RING_SIZE]);
		memcpy(ias_cmd->params, dev->dev_addr, 6);
		sp->last_cmd = ias_cmd;
	}

	/* Start the chip's Tx process and unmask interrupts. */
	wait_for_cmd_done(ioaddr + SCBCmd);
	outl(virt_to_bus(&sp->tx_ring[sp->dirty_tx % TX_RING_SIZE]),
		 ioaddr + SCBPointer);
	/* We are not ACK-ing FCP and ER in the interrupt handler yet so they should
	   remain masked --Dragan */
	outw(CUStart | SCBMaskEarlyRx | SCBMaskFlowCtl, ioaddr + SCBCmd);
}

static void
speedo_rx_soft_reset(struct net_device *dev)
{
	struct speedo_private *sp = dev->priv;
	struct RxBD *rxb;
	struct RxFD *rfd;
	long ioaddr;

	ioaddr = dev->base_addr;
	sp->intmask &= ~CMD_TO_MASK(SCBMaskRxDone);
	wait_for_cmd_done(ioaddr + SCBCmd);
	if (inb(ioaddr + SCBCmd) != 0) {
		printk("%s: previous command stalled\n", dev->name);
		return;
	}
	/*
	* Put the hardware into a known state.
	*/
	outb(RxAbort, ioaddr + SCBCmd);

	rxb = &sp->rx_buffs[sp->cur_rxbd % RXBD_RING_SIZE];
	rfd = sp->rx_ringp[sp->cur_rxfd % RXFD_RING_SIZE];

	rfd->rx_buf_addr = virt_to_le32desc(rxb);

	wait_for_cmd_done(ioaddr + SCBCmd);

	if (inb(ioaddr + SCBCmd) != 0) {
		printk("%s: RxAbort command stalled\n", dev->name);
		return;
	}
	outl(virt_to_bus(rfd), ioaddr + SCBPointer);
	outb(RxStart, ioaddr + SCBCmd);
	outb(sp->intmask, ioaddr + SCBIntmask);
}

/* Media monitoring and control. */
static void
speedo_timer(unsigned long data)
{
	struct net_device *dev = (struct net_device *)data;
	struct speedo_private *sp = (struct speedo_private *)dev->priv;
	long ioaddr = dev->base_addr;
	int phy_num = sp->phy[0] & 0x1f;
	unsigned long flags;

	/* We have MII and lost link beat. */
	if ((sp->phy[0] & 0x8000) == 0) {
		int partner = mdio_read(ioaddr, phy_num, 5);
		if (partner != sp->partner) {
			int flow_ctrl = sp->advertising & partner & 0x0400 ? 1 : 0;
			if (speedo_debug > 2) {
				printk(KERN_DEBUG "%s: Link status change.\n", dev->name);
				printk(KERN_DEBUG "%s: Old partner %x, new %x, adv %x.\n",
					   dev->name, sp->partner, partner, sp->advertising);
			}
			sp->partner = partner;
			if (flow_ctrl != sp->flow_ctrl) {
				sp->flow_ctrl = flow_ctrl;
				sp->rx_mode = -1;	/* Trigger a reload. */
			}
			/* Clear sticky bit. */
			mdio_read(ioaddr, phy_num, 1);
			/* If link beat has returned... */
			if (mdio_read(ioaddr, phy_num, 1) & 0x0004)
				dev->flags |= IFF_RUNNING;
			else
				dev->flags &= ~IFF_RUNNING;
		}
	}
	if (speedo_debug > 3) {
		printk(KERN_DEBUG "%s: Media control tick, status %4.4x.\n",
			   dev->name, inw(ioaddr + SCBStatus));
	}
	if (sp->rx_mode < 0  ||
		(sp->rx_bug  && jiffies - sp->last_rx_time > 2*HZ)) {
		/* We haven't received a packet in a Long Time.  We might have been
		   bitten by the receiver hang bug.  This can be cleared by sending
		   a set multicast list command. */
		if (speedo_debug > 2)
			printk(KERN_DEBUG "%s: Sending a multicast list set command"
				   " from a timer routine.\n", dev->name);
		set_rx_mode(dev);
		spin_lock_irqsave(&sp->lock, flags);
		speedo_rx_soft_reset(dev);
		sp->last_rx_time = jiffies;
		spin_unlock_irqrestore(&sp->lock, flags);
	}

	/* We must continue to monitor the media. */
	mod_timer(&sp->timer, RUN_AT(SLOWTIMER_RATE));
}

/* Prod routine for interrupt */
static void
speedo_fasttimer(unsigned long data)
{
	struct net_device *dev;
	struct speedo_private *sp;
	unsigned long flags;
	unsigned char intmask;

	/* 
	 * If we haven't had an interrupt since the last
	 * timer tick, set the software interrupt flag. 
	 */
	for (dev = speedo_open_list; dev; dev = sp->open_list) {
		sp = (struct speedo_private *) dev->priv;
		__cli();
		sp->intmask &= ~CMD_TO_MASK(SCBMaskRxDone);
		outb(sp->intmask, dev->base_addr + SCBIntmask);
		__sti();
		if (sp->fasttimer_cnt == 1) {
		    spin_lock_irqsave(&sp->lock, flags);
		    intmask = inb(dev->base_addr + SCBIntmask);
		    intmask |= CMD_TO_MASK(SCBTriggerIntr);
		    outb(intmask, dev->base_addr + SCBIntmask);
		    spin_unlock_irqrestore(&sp->lock, flags);
		}
	}

	/*
	 * If we have missed two ticks, then the software interrupt
	 * didn't fire, so we probably need to reset the chip.  Call
	 * into the interrupt routine to force the issue.
	 */
	for (dev = speedo_open_list; dev; dev = sp->open_list) {
		sp = (struct speedo_private *) dev->priv;
		if (sp->fasttimer_cnt > 1) {
			disable_irq_nosync(dev->irq);
			speedo_interrupt(dev->irq, dev, NULL);
			sp->fasttimer_cnt = 2; /* dont reset yet */
			enable_irq(dev->irq);
		}
		sp->fasttimer_cnt++;
	}

	mod_timer(&fasttimer, RUN_AT(FASTTIMER_RATE));
}

static void speedo_show_state(struct net_device *dev)
{
	struct speedo_private *sp = (struct speedo_private *)dev->priv;
	int i;

	/* Print a few items for debugging. */
	if (speedo_debug > 0) {
		printk(KERN_DEBUG "%s: Tx ring dump,  Tx queue %u / %u:\n", dev->name,
			   sp->cur_tx, sp->dirty_tx);
		for (i = 0; i < TX_RING_SIZE; i++)
			printk(KERN_DEBUG "%s:  %c%c%2d %8.8x.\n", dev->name,
				   i == sp->dirty_tx % TX_RING_SIZE ? '*' : ' ',
				   i == sp->cur_tx % TX_RING_SIZE ? '=' : ' ',
				   i, sp->tx_ring[i].status);
		printk(KERN_DEBUG "%s: Printing Rx ring"
			   " (next to receive into %u, dirty index %u).\n",
			   dev->name, sp->cur_rxfd, sp->dirty_rxfd);

		for (i = 0; i < RXFD_RING_SIZE; i++)
			printk(KERN_DEBUG "%s: %c%c%c%2d %8.8x.\n", dev->name,
			   i == sp->suspend_rxfd % RXFD_RING_SIZE ? 'l' : ' ',
			   i == sp->dirty_rxfd % RXFD_RING_SIZE ? '*' : ' ',
			   i == sp->cur_rxfd % RXFD_RING_SIZE ? '=' : ' ',
			   i, (unsigned)sp->rx_ringp[i]->status);
	}

#if 0
	for (i = 0; i < 16; i++) {
		/* FIXME: what does it mean?  --SAW */
		if (i == 6) i = 21;
		printk(KERN_DEBUG "%s:  PHY index %d register %d is %4.4x.\n",
			   dev->name, phy_num, i, mdio_read(ioaddr, phy_num, i));
	}
#endif

}

#if 0
static void
speedo_check_rx_ring(struct net_device *dev)
{
    struct speedo_private *sp = (struct speedo_private *)dev->priv;
    int rxf_need_deliver = 0, rxf_num_need_delivers = 0, rxf_pktwait = 0,
		rxf_num_pktwait = 0;
    int rxb_need_deliver = 0, rxb_num_need_delivers = 0, rxb_pktwait = 0,
		rxb_num_pktwait = 0;
    unsigned int i;
    int entry;
    unsigned int rbd_status;
    unsigned int rxf_status;
    int errs = 0;
    struct RxFD *rxf;
    struct RxBD *rxb;
    unsigned long flags;

    spin_lock_irqsave(&sp->lock, flags);
    for (i = 0; i < RXFD_RING_SIZE; i++) {
	rxf = sp->rx_ringp[i];

	if (virt_to_le32desc(sp->rx_ringp[(i + 1) % RXFD_RING_SIZE])
		!= rxf->link) {
	    printk("%s: bad RXF link at entry %d\n", dev->name, i);
	    errs++;
	}
    }

    for (i = 0; i < RXBD_RING_SIZE; i++) {
	rxb = &sp->rx_buffs[i];

	if (virt_to_le32desc(&sp->rx_buffs[(i + 1) % RXBD_RING_SIZE])
		!= rxb->link) {
	    printk("%s: bad RXBD link at entry %d\n", dev->name, i);
	    errs++;
	}

	if (sp->rx_skbuff[i] && rxb->rx_buf_addr
		!= virt_to_le32desc(sp->rx_skbuff[i]->tail + RXFD_HDR_SIZE)) {
	    printk("%s: bad RXBD skt buf_addr at entry %d\n", dev->name, i);
	    errs++;
	}
    }

    for (i = sp->cur_rxfd; i < sp->dirty_rxfd + RXFD_RING_SIZE; i++) {

	entry = i % RXFD_RING_SIZE;
	rxf = sp->rx_ringp[entry];
	rxf_status = cpu_to_le32(rxf->status);

	if ((rxf_status & 0xFFFF) == RXF_NEEDSFILL) {
	    printk("%s: found needsfill in Rx frame %d\n", dev->name, i);
	    errs++;
	}
	else if (rxf_status & RxComplete) {
	    if ( ! rxf_need_deliver) {
		rxf_need_deliver = 1;
		rxf_num_need_delivers++;
	    }
	    rxf_pktwait = 0;
	}
	else {
	    if ( ! rxf_pktwait) {
		rxf_pktwait = 1;
		rxf_num_pktwait++;
	    }
	    rxf_need_deliver = 0;
	}
    }

    for (i = sp->cur_rxbd; i < sp->dirty_rxbd + RXBD_RING_SIZE; i++) {
	entry = i % RXBD_RING_SIZE;
	rxb = &sp->rx_buffs[entry];
	rbd_status = cpu_to_le32(rxb->status_framesz);

	if ((rbd_status & 0xFFFF) == RBD_NEEDSFILL) {
	    printk("%s: found needsfill in Rx buffer %d\n", dev->name, i);
	    errs++;
	}
	else if (rbd_status & 0x8000) {
	    if ( ! rxb_need_deliver) {
		rxb_need_deliver = 1;
		rxb_num_need_delivers++;
	    }
	    rxb_pktwait = 0;
	}
	else {
	    if ( ! rxb_pktwait) {
		rxb_pktwait = 1;
		rxb_num_pktwait++;
	    }
	    rxb_need_deliver = 0;
	}
    }

    /*
     * This is being loosened up to check for >2.  It is possible for the
     * CPU and this code to race, and if they race correctly it can count
     * two of these, but never more than that.  In fact, 2 is rare.
     */
    if (rxf_num_need_delivers > 2) {
	printk("%s: found too many need_delivers frames\n", dev->name);
	errs++;
    }

    if (rxb_num_need_delivers > 1) {
	printk("%s: found too many need_delivers buffers\n", dev->name);
	errs++;
    }

    for (i = sp->dirty_rxfd; i < sp->cur_rxfd; i++) {
	/*
	 * From dirty up to current, there should be a
	 * block of buffers that have been filled and delivered,
	 * but now need to be refilled
	 */
	entry = i % RXFD_RING_SIZE;
	rxf = sp->rx_ringp[entry];
	rxf_status = cpu_to_le32(rxf->status);

	if ((rxf_status & 0xFFFF) != RXF_NEEDSFILL) {
	    printk("%s: dirty RXFD w/o needsfill, index %d\n", dev->name, i);
	    errs++;
	}
    }


    for (i = sp->dirty_rxbd; i < sp->cur_rxbd; i++) {
	/*
	 * As above
	 */
	entry = i % RXBD_RING_SIZE;
	rxb = &sp->rx_buffs[entry];
	rbd_status = cpu_to_le32(rxb->status_framesz);

	if ((rbd_status & 0xFFFF) != RBD_NEEDSFILL) {
	    printk("%s: dirty RXBD w/o needsfill, index %d\n", dev->name, i);
	    errs++;
	}
    }

    if (errs) {
	printk("%s: speedo_check_rx_ring found errors\n", dev->name);
#ifdef CONFIG_X86_REMOTE_DEBUG
	breakpoint();
#endif
    }
    spin_unlock_irqrestore(&sp->lock, flags);
}
#endif


/* Initialize the Rx and Tx rings, along with various 'dev' bits. */
static void
speedo_init_rx_ring(struct net_device *dev)
{
	struct speedo_private *sp = (struct speedo_private *)dev->priv;
	struct RxFD *rxf = NULL;
	struct RxBD *rxb = NULL;
	struct sk_buff *skb;
	int i;

	memset(&sp->rx_ringp[0], 0, sizeof(sp->rx_ringp));
	memset(&sp->rx_ringskb[0], 0, sizeof(sp->rx_ringskb));
	memset(&sp->rx_buffs[0], 0, sizeof(sp->rx_buffs));

	for (i = 0; i < RXFD_RING_SIZE; i++) {
		skb = dev_alloc_skb(sizeof(*rxf));
		if ( ! skb)
		    break;

		skb->dev = dev;
		sp->rx_ringskb[i] = skb;

		rxf = (struct RxFD *) skb->tail;
		sp->rx_ringp[i] = rxf;
		rxf->rx_buf_addr = 0;
		rxf->status = RXF_DESC_NONSTOP;
		rxf->count = RXFD_HDR_SIZE_INIT;
		if (i)
			sp->rx_ringp[i - 1]->link = virt_to_le32desc(rxf);
	}

	if (i == RXFD_RING_SIZE) {
		rxf = sp->rx_ringp[RXFD_RING_SIZE - 1];
		rxf->link = virt_to_le32desc(sp->rx_ringp[0]);
	}

	if (i) {
		rxf->status = RXF_DESC_STOP;
		rxf->count = RXFD_HDR_SIZE_INIT;
		rxf->rx_buf_addr = 0;
	}

	sp->suspend_rxfd = i - 1;
	sp->dirty_rxfd = 0;
	sp->cur_rxfd = 0;

	for (i = 0; i < RXBD_RING_SIZE - 1; i++) {
		rxb = &sp->rx_buffs[i];

		rxb->link = virt_to_le32desc(&sp->rx_buffs[i + 1]);
	}

	rxb = &sp->rx_buffs[RXBD_RING_SIZE - 1];
	rxb->link = virt_to_le32desc(&sp->rx_buffs[0]);

	if (sp->rx_ringp[0])
		sp->rx_ringp[0]->rx_buf_addr
			= virt_to_le32desc(&sp->rx_buffs[0]);

	for (i = 0; i < RXBD_RING_SIZE; i++) {

		skb = sp->rx_skbuff[i];
		if ( ! skb) {
			skb = dev_alloc_skb(PKT_BUF_SZ);
			sp->rx_skbuff[i] = skb;
		}
		if ( ! skb)
			break;

		sp->suspend_rxbd = i;
		skb->dev = dev;

		rxb = &sp->rx_buffs[i];
		rxb->rx_buf_addr = virt_to_le32desc(skb->tail + RXFD_HDR_SIZE);
		rxb->end_bufsz = RXB_DESC_NONSTOP_MTULEN;
	}

	sp->dirty_rxbd = (unsigned int)(i - RXBD_RING_SIZE);
	sp->cur_rxbd = 0;

	/* 
	 * There is always a buffer associated with suspend_rx
	 * unless the very first dev_alloc_skb failed.  If nothing
	 * was allocated, we set len == 0 and the end of list bit.
	 */
	if (sp->rx_skbuff[sp->suspend_rxbd])
		sp->rx_buffs[sp->suspend_rxbd].end_bufsz = RXB_DESC_STOP_MTULEN;
	else sp->rx_buffs[sp->suspend_rxbd].end_bufsz = RXB_DESC_STOP_NOLEN;
}

static void speedo_purge_tx(struct net_device *dev)
{
	struct speedo_private *sp = (struct speedo_private *)dev->priv;
	int entry;

	while ((int)(sp->cur_tx - sp->dirty_tx) > 0) {
		entry = sp->dirty_tx % TX_RING_SIZE;
		if (sp->tx_skbuff[entry]) {
			sp->stats.tx_errors++;
			dev_free_skb(sp->tx_skbuff[entry]);
			sp->tx_skbuff[entry] = 0;
		}
		sp->dirty_tx++;
	}
	while (sp->mc_setup_head != NULL) {
		struct speedo_mc_block *t;
		if (speedo_debug > 1)
			printk(KERN_DEBUG "%s: freeing mc frame.\n", dev->name);
		t = sp->mc_setup_head->next;
		kfree(sp->mc_setup_head);
		sp->mc_setup_head = t;
	}
	sp->mc_setup_tail = NULL;
	sp->tx_full = 0;
	dev->trans_start = jiffies;
	netif_wake_queue(dev);
}

static void reset_mii(struct net_device *dev)
{
	struct speedo_private *sp = (struct speedo_private *)dev->priv;
	long ioaddr = dev->base_addr;
	/* Reset the MII transceiver, suggested by Fred Young @ scalable.com. */
	if ((sp->phy[0] & 0x8000) == 0) {
		int phy_addr = sp->phy[0] & 0x1f;
		int advertising = mdio_read(ioaddr, phy_addr, 4);
		int mii_bmcr = mdio_read(ioaddr, phy_addr, 0);
		mdio_write(ioaddr, phy_addr, 0, 0x0400);
		mdio_write(ioaddr, phy_addr, 1, 0x0000);
		mdio_write(ioaddr, phy_addr, 4, 0x0000);
		mdio_write(ioaddr, phy_addr, 0, 0x8000);
#ifdef honor_default_port
		mdio_write(ioaddr, phy_addr, 0, mii_ctrl[dev->default_port & 7]);
#else
		mdio_read(ioaddr, phy_addr, 0);
		mdio_write(ioaddr, phy_addr, 0, mii_bmcr);
		mdio_write(ioaddr, phy_addr, 4, advertising);
#endif
	}
}

static void speedo_tx_timeout(struct net_device *dev)
{
	struct speedo_private *sp = (struct speedo_private *)dev->priv;
	long ioaddr = dev->base_addr;
	int status = inw(ioaddr + SCBStatus);

	SPEEDO_CHECK_RX_RING(dev);

	printk(KERN_WARNING "%s: Transmit timed out: status %4.4x "
		   " %4.4x at %d/%d command %8.8x.\n",
		   dev->name, status, inw(ioaddr + SCBCmd),
		   sp->dirty_tx, sp->cur_tx,
		   sp->tx_ring[sp->dirty_tx % TX_RING_SIZE].status);

	/* Trigger a stats dump to give time before the reset. */
	speedo_get_stats(dev);

	speedo_show_state(dev);
#if 0
	if ((status & 0x00C0) != 0x0080
		&&  (status & 0x003C) == 0x0010) {
		/* Only the command unit has stopped. */
		printk(KERN_WARNING "%s: Trying to restart the transmitter...\n",
			   dev->name);
		outl(virt_to_bus(&sp->tx_ring[sp->dirty_tx % TX_RING_SIZE]),
			 ioaddr + SCBPointer);
		outb(CUStart, ioaddr + SCBCmd);
		reset_mii(dev);
	} else
#endif
	{
	speedo_reset(dev);
	}
	SPEEDO_CHECK_RX_RING(dev);
}

static void
speedo_reset(struct net_device *dev)
{
	struct speedo_private *sp = (struct speedo_private *)dev->priv;
	long ioaddr = dev->base_addr;
	unsigned long flags;

	start_bh_atomic();
	/* Ensure that timer routine doesn't run! */
	del_timer(&sp->timer);
	sp->resetting = 1;
	end_bh_atomic();
	/*
	 * disable the transmit/receive logic
	 */
	outb(RxAbort, ioaddr + SCBCmd);
	wait_for_cmd_done(ioaddr + SCBCmd);
	/* Reset the Tx and Rx units. */
	outl(PortPartialReset, ioaddr + SCBPort);

	{
	    int i = 10000;
	    while (inw(ioaddr + SCBPort) && i-- > 0)
		    ;

	    if (inw(ioaddr + SCBPort)) {
		printk("%s: partial reset timed out\n", dev->name);
		udelay(10);
	    }
	}

	outl(PortReset, ioaddr + SCBPort);
	/* We may get spurious interrupts here.  But I don't think that they
	   may do much harm.  1999/12/09 SAW */
	udelay(10);
	/* Disable interrupts. */
	outw(SCBMaskAll, ioaddr + SCBCmd);
	synchronize_irq();
	speedo_tx_buffer_gc(dev);
	/* Free as much as possible.
	   It helps to recover from a hang because of out-of-memory.
	   It also simplifies speedo_resume() in case TX ring is full or
	   close-to-be full. */
	speedo_purge_tx(dev);
	speedo_freebufs(dev);

	speedo_init_rx_ring(dev);
	spin_lock_irqsave(&sp->lock, flags);
	speedo_resume(dev);
	sp->rx_mode = -1;
	dev->trans_start = jiffies;
	spin_unlock_irqrestore(&sp->lock, flags);
	sp->resetting = 0;
	set_rx_mode(dev); /* it takes the spinlock itself --SAW */
	/* Reset MII transceiver.  Do it before starting the timer to serialize
	   mdio_xxx operations.  Yes, it's a paranoya :-)  2000/05/09 SAW */
	reset_mii(dev);
	sp->timer.expires = RUN_AT(SLOWTIMER_RATE);
	add_timer(&sp->timer);

	return;
}

/*
 * The right way to avoid race conditions vs. hardware.  Don't
 * guess, don't bump up your priority (like the hardware state
 * machine _cares_ what the cpu priority is).  Don't put the
 * instructions "close together" - hasn't anyone heard of bus
 * contention or TLB misses?
 *
 * This is pretty simple - if the command unit is stopped, and
 * it isn't stopped at the end of the list, then it must be
 * stopped elsewhere. So we give it a resume.
 */
static inline void
speedo_tx_clr_resume(struct net_device *dev,
	struct speedo_private *sp, struct descriptor *new_end)
{
	long ioaddr = dev->base_addr;

	SPEEDO_CHECK_RX_RING(dev);
	clear_suspend(sp->last_cmd);
	sp->last_cmd = new_end;

	switch ((inw(ioaddr + SCBStatus) >> 6) & 0x3) {
	case 0:	/* Idle */
	case 2: /* Active */
	case 3: /* Reserved */
		break;

	case 1: /* Suspended */
		if ( ! (le32_to_cpu(new_end->cmd_status) & StatusComplete)) {
			wait_for_cmd_done(ioaddr + SCBCmd);

			if (inb(ioaddr + SCBCmd) == 0)
				outb(CUResume, ioaddr + SCBCmd);
			else printk("%s: clr_resume wait_for_cmd_done failed\n",
				dev->name);
		}
	}
}


static int
speedo_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct speedo_private *sp = (struct speedo_private *)dev->priv;
	long ioaddr = dev->base_addr;
	int entry;

	SPEEDO_CHECK_RX_RING(dev);
#if ! defined(HAS_NETIF_QUEUE)
	if (test_bit(0, (void*)&dev->tbusy) != 0) {
		int tickssofar = jiffies - dev->trans_start;
		if (tickssofar < TX_TIMEOUT - 2)
			return 1;
		if (tickssofar < TX_TIMEOUT) {
			/* Reap sent packets from the full Tx queue. */
			unsigned long flags;
			/* Take a spinlock to make wait_for_cmd_done and sending the
			command atomic.  --SAW */
			spin_lock_irqsave(&sp->lock, flags);
			wait_for_cmd_done(ioaddr + SCBCmd);
			outw(SCBTriggerIntr, ioaddr + SCBCmd);
			spin_unlock_irqrestore(&sp->lock, flags);
			return 1;
		}
		speedo_tx_timeout(dev);
		return 1;
	}
#endif

	{	/* Prevent interrupts from changing the Tx ring from underneath us. */
		unsigned long flags;

		spin_lock_irqsave(&sp->lock, flags);

		/* Check if there are enough space. */
		if ((int)(sp->cur_tx - sp->dirty_tx) >= TX_QUEUE_LIMIT) {
			printk(KERN_ERR "%s: incorrect tbusy state, fixed.\n", dev->name);
			dev->trans_start = jiffies;
			netif_stop_queue(dev);
			sp->tx_full = 1;
			spin_unlock_irqrestore(&sp->lock, flags);
			return 1;
		}

		/* Calculate the Tx descriptor entry. */
		entry = sp->cur_tx++ % TX_RING_SIZE;

		sp->tx_skbuff[entry] = skb;
		sp->tx_ring[entry].status =
			cpu_to_le32(CmdSuspend | CmdTx | CmdTxFlex);
		if (!(entry & ((TX_RING_SIZE>>2)-1)))
			sp->tx_ring[entry].status |= cpu_to_le32(CmdIntr);
		sp->tx_ring[entry].link =
			virt_to_le32desc(&sp->tx_ring[sp->cur_tx % TX_RING_SIZE]);
		sp->tx_ring[entry].tx_desc_addr =
			virt_to_le32desc(&sp->tx_ring[entry].tx_buf_addr0);
		/* The data region is always in one buffer descriptor. */
		sp->tx_ring[entry].count = cpu_to_le32(sp->tx_threshold);
		sp->tx_ring[entry].tx_buf_addr0 = virt_to_le32desc(skb->data);
		sp->tx_ring[entry].tx_buf_size0 = cpu_to_le32(skb->len);

		speedo_tx_clr_resume(dev, sp,
			(struct descriptor *) &sp->tx_ring[entry]);

		/* Leave room for set_rx_mode(). If there is no more space than reserved
		   for multicast filter mark the ring as full. */
		if ((int)(sp->cur_tx - sp->dirty_tx) >= TX_QUEUE_LIMIT) {
			netif_stop_queue(dev);
			sp->tx_full = 1;
		}

		spin_unlock_irqrestore(&sp->lock, flags);
	}

	dev->trans_start = jiffies;

	SPEEDO_CHECK_RX_RING(dev);
	return 0;
}

static int speedo_tx_buffer_gc(struct net_device *dev)
{
	unsigned int dirty_tx;
	struct speedo_private *sp = (struct speedo_private *)dev->priv;
	int npkts = 0;

	SPEEDO_CHECK_RX_RING(dev);
	dirty_tx = sp->dirty_tx;
	while ((int)(sp->cur_tx - dirty_tx) > 0) {
		int entry = dirty_tx % TX_RING_SIZE;
		int status = le32_to_cpu(sp->tx_ring[entry].status);

		if (speedo_debug > 5)
			printk(KERN_DEBUG " scavenge candidate %d status %4.4x.\n",
				   entry, status);
		if ((status & StatusComplete) == 0)
			break;			/* It still hasn't been processed. */
		npkts++;
		if (status & TxUnderrun)
			if (sp->tx_threshold < 0x01e08000) {
				if (speedo_debug > 2)
					printk(KERN_DEBUG "%s: TX underrun, threshold adjusted.\n",
						   dev->name);
				sp->tx_threshold += 0x00040000;
			}
		/* Free the original skb. */
		if (sp->tx_skbuff[entry]) {
			sp->stats.tx_packets++;	/* Count only user packets. */
			sp->stats.tx_bytes += sp->tx_skbuff[entry]->len;
			dev_free_skb(sp->tx_skbuff[entry]);
			sp->tx_skbuff[entry] = 0;
		}
		dirty_tx++;
	}

	if (speedo_debug && (int)(sp->cur_tx - dirty_tx) > TX_RING_SIZE) {
		printk(KERN_ERR "out-of-sync dirty pointer, %d vs. %d,"
			   " full=%d.\n",
			   dirty_tx, sp->cur_tx, sp->tx_full);
		dirty_tx += TX_RING_SIZE;
	}

	while (sp->mc_setup_head != NULL
		   && (int)(dirty_tx - sp->mc_setup_head->tx - 1) > 0) {
		struct speedo_mc_block *t;
		if (speedo_debug > 1)
			printk(KERN_DEBUG "%s: freeing mc frame.\n", dev->name);
		t = sp->mc_setup_head->next;
		kfree(sp->mc_setup_head);
		sp->mc_setup_head = t;
		npkts++;
	}
	if (sp->mc_setup_head == NULL)
		sp->mc_setup_tail = NULL;

	sp->dirty_tx = dirty_tx;
	SPEEDO_CHECK_RX_RING(dev);

	return npkts;
}

/*
 * Under very high load (pounding hard on 4 interfaces) we don't
 * always get to the watchdog refresh routine.  Re-arm it here
 * just to be safe.
 */
#include <linux/cobalt_wdt.h>

#define COBALT_WDT_CLEARDOG cobalt_wdt_cleardog()

static struct net_device *interrupt_last_dev;

int in_speedo_intr = 0;
/* The interrupt handler does all of the Rx thread work and cleans up
   after the Tx thread. */
static void
speedo_interrupt(int irq, void *dev_instance, struct pt_regs *regs)
{
    struct net_device *dev = (struct net_device *)dev_instance;
    struct speedo_private *sp;
    long ioaddr;
    unsigned short status;
    int cur_txcnt;
    int cur_rxcnt;

#ifndef final_version
    if (dev_instance == NULL) {
	printk(KERN_ERR "speedo_interrupt: irq %d for unknown device.\n", irq);
	return;
    }
#endif
	if (++in_speedo_intr > 1) {
		printk("Warning: eepro100: speedo_interrupt recursion: %d entries!\n", in_speedo_intr);
	}

	ioaddr = dev->base_addr;
	sp = (struct speedo_private *)dev->priv;
	sp->fasttimer_cnt = 0;

	COBALT_WDT_CLEARDOG;

	if (sp->resetting)
		goto out;

	if (jiffies - sp->last_rx_time > SLOWTIMER_RATE) {
	    /*
	     * If things are too quiet, we may need to re-init the receiver
	     */
	    set_rx_mode(dev);
	    speedo_rx_soft_reset(dev);
	    sp->last_rx_time = jiffies;
	}

	/* 
	 * Get pending status/interrupts and clear interrupts
	 */
	status = inw(ioaddr + SCBStatus);
	outw(status & 0xfc00, ioaddr + SCBStatus);

	if (speedo_debug > 4)
	    printk(KERN_DEBUG "%s: interrupt status=%#4.4x.\n",
		    dev->name, status);

	dev->trans_start = jiffies;

	/*
	 * If we are masking RX, continue to do so, unless the
	 * receive logic is actually blocked.  This enables
	 * output messages to take priority under heavy load.
	 */
	if ( ! (sp->intmask & CMD_TO_MASK(SCBMaskRxDone))
	  || (status & SCBMaskRxSuspend)) {

	    sp->intmask &= ~CMD_TO_MASK(SCBMaskRxDone);

	    cur_rxcnt = speedo_rx(dev);
	    sp->rx_loopcnt = cur_rxcnt;	/* per device per interrupt */

	    if (cur_rxcnt)
		sp->last_rx_time = jiffies;

	    if (cur_rxcnt)
		dev->trans_start = jiffies;

	    /*
	     * If the sum of all packets on this device for this interrupt
	     * exceeds a full RXFD ring, there is too high a load incoming.  
	     * We set a mask flag, and disable the RX interrupt so that we
	     * have time to respond to the packets that we received.
	     *
	     * The stabilizing effects of this can be seen on bidirection
	     * TCP thruput measurements with ttcp across multiple NICs.
	     * (I used 4.)  In that case, without this, you see the distribution
	     * spread of the different ttcp sessions get much wider.  The
	     * percentage of variation in individual session times quadruples!
	     */
	    if (sp->rx_loopcnt > RXFD_RING_SIZE)
		sp->intmask |= CMD_TO_MASK(SCBMaskRxDone);

	    sp->intmask &= ~CMD_TO_MASK(SCBMaskRxSuspend);
	    speedo_refill_rx_buffers(dev);
	}

	switch ((status >> 2) & 0xf) {
	case 0:	/* Idle */
	    break;
	case 4:	/* Ready */
	    break;
	case 1:	 /* Suspended */
	case 2:	 /* No resources (RxFDs) */
	case 9:	 /* Suspended with no more RBDs */
	case 10: /* No resources due to no RBDs */
	case 12: /* Ready with no RBDs */
	    speedo_rx_soft_reset(dev);
	    break;
	case 3: case 5: case 6: case 7: case 8:
	case 11: case 13: case 14: case 15:
	    /* These are all reserved values */
	    break;
	}

	/* User interrupt, Command/Tx unit interrupt or CU not active. */
	spin_lock(&sp->lock);
	cur_txcnt = speedo_tx_buffer_gc(dev);

	if (cur_txcnt)
	    dev->trans_start = jiffies;

	if (sp->tx_full && (int)(sp->cur_tx - sp->dirty_tx) < TX_QUEUE_UNFULL) {
	    /* The ring is no longer full. */
	    sp->tx_full = 0;
	    dev->trans_start = jiffies;
	    netif_wake_queue(dev); /* Attention: under a spinlock.  --SAW */
	}
	spin_unlock(&sp->lock);

out:
    if (speedo_debug > 3)
	printk(KERN_DEBUG "%s: exiting interrupt, status=%#4.4x.\n",
		dev->name, inw(ioaddr + SCBStatus));

    __cli();
    if (!sp->resetting) {
	dev->interrupt = 0;
	outb(sp->intmask, dev->base_addr + SCBIntmask);
    }
    __sti();

    in_speedo_intr--;
    if (in_speedo_intr < 0) {
	/* Should not ever happen... */
	in_speedo_intr = 0;
    }
    return;
}

static int
compare_and_swap(int *addr, int new, int old)
{
	asm volatile ("movl %1, %%eax; lock ; cmpxchgl %2,%0 ; movl %%eax, %1"
		: "=m" (*addr), "+r" (old)
		: "r" (new)
		: "%eax", "cc");

	return old;
}

static void speedo_refill_rx_buffers(struct net_device *dev)
{
    struct speedo_private *sp = (struct speedo_private *)dev->priv;
    int entry;
    struct RxFD *rxf = NULL;
    struct RxBD *rxb = NULL;
    struct sk_buff *skb;

    /*
    * We leave one buffer gap because we don't trust the
    * documentation.  The CU says a resume makes it reread
    * the TxFD, and the RU says it doesn't.  Rather than
    * risking a typo, we make a point of saving the space
    * where our suspend command is hiding.
    */
    while (sp->cur_rxbd - sp->dirty_rxbd > 1) {
	entry = sp->dirty_rxbd % RXBD_RING_SIZE;

	if ( ! sp->rx_skbuff[entry]) {
	    skb = dev_alloc_skb(PKT_BUF_SZ);
	    if (skb == NULL)
		break;

	    skb->dev = dev;
	    sp->rx_skbuff[entry] = skb;
	}
	else skb = sp->rx_skbuff[entry];

	rxb = &sp->rx_buffs[entry];

	if (sp->dirty_rxbd == sp->suspend_rxbd) {
		if (speedo_debug > 3)
			printk("%s: refill overran RBD suspend\n", dev->name);
	}
	rxb->rx_buf_addr = virt_to_bus(skb->tail + RXFD_HDR_SIZE);
	rxb->link = virt_to_le32desc(&sp->rx_buffs[(entry + 1)
		% RXBD_RING_SIZE]);
	rxb->end_bufsz = RXB_DESC_NONSTOP_MTULEN;
	rxb->status_framesz = 0;

	sp->dirty_rxbd++;
    }

    if (rxb) {
	rxb->end_bufsz = RXB_DESC_STOP_MTULEN;
	/*
	 * We ignore the result of the CAS because we can't do anything one
	 * way or the other.  We just use the CAS to avoid overwriting the
	 * rxfd status that the chip may have written.  The original driver
	 * used a "&=", which is hardly atomic.
	 */
	rxb = &sp->rx_buffs[sp->suspend_rxbd % RXBD_RING_SIZE];
	(void) compare_and_swap(&rxb->end_bufsz,
		RXB_DESC_NONSTOP_MTULEN, RXB_DESC_STOP_MTULEN);

	/* 
	* There are two cases here.
	*	1) We cleared the suspend bit before the hardware read it.
	*	2) The hardware got there first
	*
	* Case 1 is what we hope for - this lets us pipeline receive without
	* any stall because of the suspend.
	*
	* Case 2 happens because of the race condition between the hardware
	* and software.  There is no way to avoid the race without forcing
	* stop/start for every suspend, and risking dropped packets.  During
	* the interrupt, we check and see if the RU is suspended.  If it is,
	* we see if the chip has processed the buffer at sp->suspend_rx.
	* If so, we cannot blindly issue a resume - we need more buffers first.
	*/
	sp->suspend_rxbd = sp->dirty_rxbd - 1;
    }

    while (sp->cur_rxfd - sp->dirty_rxfd > 1) {
	entry = sp->dirty_rxfd % RXFD_RING_SIZE;
	rxf = sp->rx_ringp[entry];

	rxf->rx_buf_addr = 0xffffffff;
	rxf->count = RXFD_HDR_SIZE_INIT;
	rxf->status = RXF_DESC_NONSTOP;
	rxf->link = virt_to_le32desc(sp->rx_ringp[(entry + 1)
			    % RXFD_RING_SIZE]);
	sp->dirty_rxfd++;
    }

    if (rxf) {
	/*
	 * This is the same as the big comment above, but for the RFDs.
	 */
	rxf->status = RXF_DESC_STOP;
	rxf = sp->rx_ringp[sp->suspend_rxfd % RXFD_RING_SIZE];
	(void) compare_and_swap(&rxf->status, RXF_DESC_NONSTOP, RXF_DESC_STOP);
	sp->suspend_rxfd = sp->dirty_rxfd - 1;
    }

    sp->rx_ring_state &= ~(RrNoMem|RrOOMReported); /* Mark the progress. */
    return;
}

static int
speedo_rx(struct net_device *dev)
{
    struct speedo_private *sp = (struct speedo_private *)dev->priv;
    volatile struct RxFD *rfd;
    volatile struct RxBD *rbd;
    int entry, rxb_entry;
    int status;
    int pkt_len;
    int npkts = 0;
    struct sk_buff *skb;

    if (speedo_debug > 4)
	printk(KERN_DEBUG " In speedo_rx().\n");
    /* 
     * Loop delivering messages until we reach the dirty_rx buffer
     * or get to a buffer that has not been filled yet.
     */
    SPEEDO_CHECK_RX_RING(dev);
    while (1) {
	entry = sp->cur_rxfd % RXFD_RING_SIZE;
	rfd = sp->rx_ringp[entry];

	if ( ! rfd || npkts > RXFD_RING_SIZE / 2)
	    break;

	status = le32_to_cpu(rfd->status);

	if ( ! (status & RxComplete)
		|| sp->cur_rxfd == sp->dirty_rxfd + RXFD_RING_SIZE)
	    break;

	rxb_entry = sp->cur_rxbd % RXBD_RING_SIZE;
	rbd = &sp->rx_buffs[rxb_entry];

	/*
	 * At one point, it seemed as if RxComplete was being set before
	 * the rfd->count bits were.  It ended up not being reproducible,
	 * but is worth mentioning--it might be possible under high load...
	 */
	if ((le32_to_cpu(rfd->count) & 0xC000) == 0xC000) {
	    /*
	     * This packet fit entirely in the RFD
	     */
	    pkt_len = le32_to_cpu(rfd->count) & 0x3fff;

	    if ((skb = dev_alloc_skb(pkt_len + 2)) == 0) {
		/*
		 * No buffers - give up
		 */
		break;
	    }
	    sp->cur_rxfd++;
	    rfd->status = le32_to_cpu(RXF_NEEDSFILL);

	    skb->dev = dev;
	    skb_reserve(skb, 2);	/* Align IP to 16 bit */
#if !defined(__alpha__)
	    /* Packet is in one chunk -- we can copy + cksum. */
	    eth_copy_and_sum(skb, (void *)&rfd->hdr[0], pkt_len, 0);
	    skb_put(skb, pkt_len);
#else
	    memcpy(skb_put(skb, pkt_len), &rfd->hdr[0], pkt_len);
#endif
	}
	else {
	    sp->cur_rxfd++;
	    rfd->status = le32_to_cpu(RXF_NEEDSFILL);
	    /*
	     * This packet ran over into the RxBD buffer.
	     */
	    if ((le32_to_cpu(rbd->status_framesz) & 0xC000) != 0xC000) {

		speedo_reset(dev);
		return npkts;
	    }

	    sp->cur_rxbd++;
	    skb = sp->rx_skbuff[rxb_entry];
	    if (skb == NULL) {
		printk(KERN_ERR "%s: Inconsistent Rx descriptor chain.\n",
			dev->name);
		sp->stats.rx_errors++;
		continue;
	    }
	    sp->rx_skbuff[rxb_entry] = 0;
	    memcpy(skb->tail, (void *)rfd->hdr, RXFD_HDR_SIZE);
	    pkt_len = le32_to_cpu(rbd->status_framesz) & 0x3fff;
	    pkt_len += RXFD_HDR_SIZE;
	    rbd->status_framesz = le32_to_cpu(RBD_NEEDSFILL);
	    rbd->rx_buf_addr = 0xffffffff;
	    rbd->end_bufsz = 0;
	    skb_put(skb, pkt_len);
	}

	/*
	 * at this point, skb points to the buffer
	 * and pkt_len is the length of the transfer.
	 */
	if (speedo_debug > 4)
	    printk(KERN_DEBUG " speedo_rx() status %8.8x len %d.\n",
		    status, pkt_len);

	if ((status & (RxErrTooBig|RxOK|0x0f90)) != RxOK) {
	    sp->stats.rx_errors++;
	    if (status & RxErrTooBig) {
		if (speedo_debug > 3)
		    printk(KERN_ERR "%s: Ethernet frame overran the Rx buffer,"
			    " status %8.8x!\n", dev->name, status);
	    }
	    else if (! (status & RxOK)) {
		/* There was a fatal error.  This *should* be impossible. */
		printk(KERN_ERR "%s: Anomalous event in speedo_rx(), "
			"status %8.8x.\n", dev->name, status);
	    }
	    dev_free_skb(skb);
	    continue;
	}

	skb->protocol = eth_type_trans(skb, dev);
	netif_rx(skb);
	sp->stats.rx_packets++;
	sp->stats.rx_bytes += pkt_len;

	sp->rx_ring_state &= ~RrPostponed;

	if (++npkts % (RXFD_RING_SIZE / 4) == 0) {
	    speedo_refill_rx_buffers(dev);
	}
    }

    SPEEDO_CHECK_RX_RING(dev);
    return npkts;
}

static void
speedo_freebufs(struct net_device *dev)
{
	struct speedo_private *sp = (struct speedo_private *)dev->priv;
	struct sk_buff *skb;
	int i;

	/* Free all the skbuffs in the Rx and Tx queues. */
	for (i = 0; i < RXBD_RING_SIZE; i++) {
		skb = sp->rx_skbuff[i];
		sp->rx_skbuff[i] = 0;
		/* Clear the Rx descriptors. */
		if (skb)
			dev_free_skb(skb);
	}

	for (i = 0; i < RXFD_RING_SIZE; i++) {
		skb = sp->rx_ringskb[i];
		sp->rx_ringskb[i] = 0;
		/* Clear the Rx descriptors. */
		if (skb)
			dev_free_skb(skb);
	}

	for (i = 0; i < TX_RING_SIZE; i++) {
		skb = sp->tx_skbuff[i];
		sp->tx_skbuff[i] = 0;
		/* Clear the Tx descriptors. */
		if (skb)
			dev_free_skb(skb);
	}
}

static int
speedo_close(struct net_device *dev)
{
	long ioaddr = dev->base_addr;
	struct speedo_private *sp = (struct speedo_private *)dev->priv;
	int i;
	struct net_device **tdevpp, *tdev;
	struct speedo_private *tsp;

	dev->start = 0;
	netif_stop_queue(dev);

	if (speedo_debug > 1)
		printk(KERN_DEBUG "%s: Shutting down ethercard, status was %4.4x.\n",
			   dev->name, inw(ioaddr + SCBStatus));

	/* Shut off the media monitoring timer. */
	start_bh_atomic();
	del_timer(&sp->timer);

	if ( --fasttimer_numopens == 0) {
		del_timer(&fasttimer);
	}

	end_bh_atomic();
	/* Shutting down the chip nicely fails to disable flow control. So.. */
	outl(PortPartialReset, ioaddr + SCBPort);

	i = 10000;
	while (inw(ioaddr + SCBPort) && i-- > 0)
		;

	if (inw(ioaddr + SCBPort)) {
	    printk("%s: partial reset timed out\n", dev->name);
	    udelay(10);
	}

	outl(PortReset, ioaddr + SCBPort);
	/* We may get spurious interrupts here.  But I don't think that they
	   may do much harm.  1999/12/09 SAW */
	udelay(10);
	/* Disable interrupts. */

	/*
	 * Before we nuke the interrupt, we have to be sure
	 * we reenable the IRQ for any shared devices.
	 */
	if (sp->irq_shadow_master) {
	    for (tdev = speedo_open_list; tdev; tdev = tsp->open_list ) {
		tsp = (struct speedo_private *) tdev->priv;
		if (dev != tdev && tdev->irq == dev->irq) {
		    tsp->irq_shadow_master = 1;
		    tsp->irq_shadow_slave = 0;
		    break;
		}
	    }
	}
	sp->resetting = 1;

	free_irq(dev->irq, dev);

	for (tdevpp = &speedo_open_list; *tdevpp && *tdevpp != dev; ) {
	    tdevpp = &(((struct speedo_private *)(*tdevpp)->priv)->open_list);
	}

	if (*tdevpp) 
	    *tdevpp = ((struct speedo_private *)(*tdevpp)->priv)->open_list;

	interrupt_last_dev = 0;
	sp->open_list = 0;

	/* Print a few items for debugging. */
	if (speedo_debug > 3)
		speedo_show_state(dev);

	speedo_freebufs(dev);

	/* Free multicast setting blocks. */
	for (i = 0; sp->mc_setup_head != NULL; i++) {
		struct speedo_mc_block *t;
		t = sp->mc_setup_head->next;
		kfree(sp->mc_setup_head);
		sp->mc_setup_head = t;
	}
	sp->mc_setup_tail = NULL;
	if (speedo_debug > 0)
		printk(KERN_DEBUG "%s: %d multicast blocks dropped.\n", dev->name, i);

	sp->resetting = 0;
	MOD_DEC_USE_COUNT;

	return 0;
}

/* The Speedo-3 has an especially awkward and unusable method of getting
   statistics out of the chip.  It takes an unpredictable length of time
   for the dump-stats command to complete.  To avoid a busy-wait loop we
   update the stats with the previous dump results, and then trigger a
   new dump.

   These problems are mitigated by the current /proc implementation, which
   calls this routine first to judge the output length, and then to emit the
   output.

   Oh, and incoming frames are dropped while executing dump-stats!
   */
static struct enet_statistics *
speedo_get_stats(struct net_device *dev)
{
	struct speedo_private *sp = (struct speedo_private *)dev->priv;
	long ioaddr = dev->base_addr;

	/* Update only if the previous dump finished. */
	if (sp->lstats.done_marker == le32_to_cpu(0xA007)) {
		sp->stats.tx_aborted_errors += le32_to_cpu(sp->lstats.tx_coll16_errs);
		sp->stats.tx_window_errors += le32_to_cpu(sp->lstats.tx_late_colls);
		sp->stats.tx_fifo_errors += le32_to_cpu(sp->lstats.tx_underruns);
		sp->stats.tx_fifo_errors += le32_to_cpu(sp->lstats.tx_lost_carrier);
		/*sp->stats.tx_deferred += le32_to_cpu(sp->lstats.tx_deferred);*/
		sp->stats.collisions += le32_to_cpu(sp->lstats.tx_total_colls);
		sp->stats.rx_crc_errors += le32_to_cpu(sp->lstats.rx_crc_errs);
		sp->stats.rx_frame_errors += le32_to_cpu(sp->lstats.rx_align_errs);
		sp->stats.rx_over_errors += le32_to_cpu(sp->lstats.rx_resource_errs);
		sp->stats.rx_fifo_errors += le32_to_cpu(sp->lstats.rx_overrun_errs);
		sp->stats.rx_length_errors += le32_to_cpu(sp->lstats.rx_runt_errs);
		sp->lstats.done_marker = 0x0000;
		if (dev->start) {
			unsigned long flags;
			/* Take a spinlock to make wait_for_cmd_done and sending the
			   command atomic.  --SAW */
			spin_lock_irqsave(&sp->lock, flags);
			wait_for_cmd_done(ioaddr + SCBCmd);
			outb(CUDumpStats, ioaddr + SCBCmd);
			spin_unlock_irqrestore(&sp->lock, flags);
		}
	}
	return &sp->stats;
}

static int speedo_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
	struct speedo_private *sp = (struct speedo_private *)dev->priv;
	long ioaddr = dev->base_addr;
	u16 *data = (u16 *)&rq->ifr_data;
	int phy = sp->phy[0] & 0x1f;

    switch(cmd) {
	case SIOCDEVPRIVATE:		/* Get the address of the PHY in use. */
		data[0] = phy;
	case SIOCDEVPRIVATE+1:		/* Read the specified MII register. */
		/* FIXME: these operations need to be serialized with MDIO
		   access from the timeout handler.
		   They are currently serialized only with MDIO access from the
		   timer routine.  2000/05/09 SAW */
		start_bh_atomic();
		data[3] = mdio_read(ioaddr, data[0], data[1]);
		end_bh_atomic();
		return 0;
	case SIOCDEVPRIVATE+2:		/* Write the specified MII register */
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		start_bh_atomic();
		mdio_write(ioaddr, data[0], data[1], data[2]);
		end_bh_atomic();
		return 0;
	default:
		return -EOPNOTSUPP;
	}
}

/* Set or clear the multicast filter for this adaptor.
   This is very ugly with Intel chips -- we usually have to execute an
   entire configuration command, plus process a multicast command.
   This is complicated.  We must put a large configuration command and
   an arbitrarily-sized multicast command in the transmit list.
   To minimize the disruption -- the previous command might have already
   loaded the link -- we convert the current command block, normally a Tx
   command, into a no-op and link it to the new command.
*/
static void set_rx_mode(struct net_device *dev)
{
	struct speedo_private *sp = (struct speedo_private *)dev->priv;
	char new_rx_mode;
	unsigned long flags;
	int entry, i;

	if (dev->flags & IFF_PROMISC) {			/* Set promiscuous. */
		new_rx_mode = 3;
	} else if ((dev->flags & IFF_ALLMULTI)  ||
			   dev->mc_count > multicast_filter_limit) {
		new_rx_mode = 1;
	} else
		new_rx_mode = 0;

	if (speedo_debug > 3)
		printk(KERN_DEBUG "%s: set_rx_mode %d -> %d\n", dev->name,
				sp->rx_mode, new_rx_mode);

	if ((int)(sp->cur_tx - sp->dirty_tx) > TX_RING_SIZE - TX_MULTICAST_SIZE) {
	    /* The Tx ring is full -- don't add anything!  Hope the mode will be
		 * set again later. */
		sp->rx_mode = -1;
		return;
	}

	if (new_rx_mode != sp->rx_mode) {
		u8 *config_cmd_data;

		spin_lock_irqsave(&sp->lock, flags);
		entry = sp->cur_tx++ % TX_RING_SIZE;

		sp->tx_skbuff[entry] = 0;			/* Redundant. */
		sp->tx_ring[entry].status = cpu_to_le32(CmdSuspend | CmdConfigure);
		sp->tx_ring[entry].link =
			virt_to_le32desc(&sp->tx_ring[(entry + 1) % TX_RING_SIZE]);
		config_cmd_data = (void *)&sp->tx_ring[entry].tx_desc_addr;
		/* Construct a full CmdConfig frame. */
		memcpy(config_cmd_data, i82558_config_cmd, sizeof(i82558_config_cmd));
		config_cmd_data[1] = (txfifo << 4) | rxfifo;
		config_cmd_data[4] = rxdmacount;
		config_cmd_data[5] = txdmacount + 0x80;
		config_cmd_data[15] |= (new_rx_mode & 2) ? 1 : 0;
		/* 0x80 doesn't disable FC 0x84 does.
		   Disable Flow control since we are not ACK-ing any FC interrupts
		   for now. --Dragan */
		config_cmd_data[19] = 0x84;
		config_cmd_data[19] |= sp->full_duplex ? 0x40 : 0;
		config_cmd_data[21] = (new_rx_mode & 1) ? 0x0D : 0x05;
		if (sp->phy[0] & 0x8000) {			/* Use the AUI port instead. */
			config_cmd_data[15] |= 0x80;
			config_cmd_data[8] = 0;
		}

		speedo_tx_clr_resume(dev, sp,
			(struct descriptor *) &sp->tx_ring[entry]);

		if ((int)(sp->cur_tx - sp->dirty_tx) >= TX_QUEUE_LIMIT) {
			dev->trans_start = jiffies;
			netif_stop_queue(dev);
			sp->tx_full = 1;
		}
		spin_unlock_irqrestore(&sp->lock, flags);
	}

	if (new_rx_mode == 0  &&  dev->mc_count < 4) {
		/* The simple case of 0-3 multicast list entries occurs often, and
		   fits within one tx_ring[] entry. */
		struct dev_mc_list *mclist;
		u16 *setup_params, *eaddrs;

		spin_lock_irqsave(&sp->lock, flags);
		entry = sp->cur_tx++ % TX_RING_SIZE;

		sp->tx_skbuff[entry] = 0;
		sp->tx_ring[entry].status = cpu_to_le32(CmdSuspend | CmdMulticastList);
		sp->tx_ring[entry].link =
			virt_to_le32desc(&sp->tx_ring[(entry + 1) % TX_RING_SIZE]);
		sp->tx_ring[entry].tx_desc_addr = 0; /* Really MC list count. */
		setup_params = (u16 *)&sp->tx_ring[entry].tx_desc_addr;
		*setup_params++ = cpu_to_le16(dev->mc_count*6);
		/* Fill in the multicast addresses. */
		for (i = 0, mclist = dev->mc_list; i < dev->mc_count;
			 i++, mclist = mclist->next) {
			eaddrs = (u16 *)mclist->dmi_addr;
			*setup_params++ = *eaddrs++;
			*setup_params++ = *eaddrs++;
			*setup_params++ = *eaddrs++;
		}

		speedo_tx_clr_resume(dev, sp,
			(struct descriptor *) &sp->tx_ring[entry]);

		if ((int)(sp->cur_tx - sp->dirty_tx) >= TX_QUEUE_LIMIT) {
			dev->trans_start = jiffies;
			netif_stop_queue(dev);
			sp->tx_full = 1;
		}
		spin_unlock_irqrestore(&sp->lock, flags);
	} else if (new_rx_mode == 0) {
		struct dev_mc_list *mclist;
		u16 *setup_params, *eaddrs;
		struct speedo_mc_block *mc_blk;
		struct descriptor *mc_setup_frm;
		int i;

		mc_blk = kmalloc(sizeof(*mc_blk) + 2 + multicast_filter_limit*6,
						 GFP_ATOMIC);
		if (mc_blk == NULL) {
			printk(KERN_ERR "%s: Failed to allocate a setup frame.\n",
				   dev->name);
			sp->rx_mode = -1; /* We failed, try again. */
			return;
		}
		mc_blk->next = NULL;
		mc_setup_frm = &mc_blk->frame;

		/* Fill the setup frame. */
		if (speedo_debug > 1)
			printk(KERN_DEBUG "%s: Constructing a setup frame at %p.\n",
				   dev->name, mc_setup_frm);
		mc_setup_frm->cmd_status =
			cpu_to_le32(CmdSuspend | CmdIntr | CmdMulticastList);
		/* Link set below. */
		setup_params = (u16 *)&mc_setup_frm->params;
		*setup_params++ = cpu_to_le16(dev->mc_count*6);
		/* Fill in the multicast addresses. */
		for (i = 0, mclist = dev->mc_list; i < dev->mc_count;
			 i++, mclist = mclist->next) {
			eaddrs = (u16 *)mclist->dmi_addr;
			*setup_params++ = *eaddrs++;
			*setup_params++ = *eaddrs++;
			*setup_params++ = *eaddrs++;
		}

		/* Disable interrupts while playing with the Tx Cmd list. */
		spin_lock_irqsave(&sp->lock, flags);

		if (sp->mc_setup_tail)
			sp->mc_setup_tail->next = mc_blk;
		else
			sp->mc_setup_head = mc_blk;
		sp->mc_setup_tail = mc_blk;
		mc_blk->tx = sp->cur_tx;

		entry = sp->cur_tx++ % TX_RING_SIZE;

		/* Change the command to a NoOp, pointing to the CmdMulti command. */
		sp->tx_skbuff[entry] = 0;
		sp->tx_ring[entry].status = cpu_to_le32(CmdNOp);
		sp->tx_ring[entry].link = virt_to_le32desc(mc_setup_frm);

		/* Set the link in the setup frame. */
		mc_setup_frm->link =
			virt_to_le32desc(&(sp->tx_ring[(entry+1) % TX_RING_SIZE]));

		speedo_tx_clr_resume(dev, sp, mc_setup_frm);

		if ((int)(sp->cur_tx - sp->dirty_tx) >= TX_QUEUE_LIMIT) {
			dev->trans_start = jiffies;
			netif_stop_queue(dev);
			sp->tx_full = 1;
		}
		spin_unlock_irqrestore(&sp->lock, flags);

		if (speedo_debug > 5)
			printk(" CmdMCSetup frame length %d in entry %d.\n",
				   dev->mc_count, entry);
	}

	sp->rx_mode = new_rx_mode;
}


#ifdef MODULE

int init_module(void)
{
	int cards_found;

	if (debug >= 0 && speedo_debug != debug)
		printk(KERN_INFO "eepro100.c: Debug level is %d.\n", debug);
	if (debug >= 0)
		speedo_debug = debug;
	/* Always emit the version message. */
	if (speedo_debug)
		printk(KERN_INFO "%s", version);

	cards_found = eepro100_init();
	if (cards_found <= 0) {
		printk(KERN_INFO "eepro100: No cards found, driver not installed.\n");
		return -ENODEV;
	}
	return 0;
}

void
cleanup_module(void)
{
	struct net_device *next_dev;

	/* No need to check MOD_IN_USE, as sys_delete_module() checks. */
	while (root_speedo_dev) {
		struct speedo_private *sp = (void *)root_speedo_dev->priv;
		unregister_netdev(root_speedo_dev);
		release_region(root_speedo_dev->base_addr, SPEEDO3_TOTAL_SIZE);
#ifndef USE_IO
		iounmap((char *)root_speedo_dev->base_addr);
#endif
		next_dev = sp->next_module;
		if (sp->priv_addr)
			kfree(sp->priv_addr);
		kfree(root_speedo_dev);
		root_speedo_dev = next_dev;
	}
}

#else   /* not MODULE */

int eepro100_probe(void)
{
	int cards_found = 0;

	cards_found = eepro100_init();

	if (speedo_debug > 0  &&  cards_found)
		printk(version);

	return cards_found ? 0 : -ENODEV;
}
#endif  /* MODULE */

/*
 * Local variables:
 *  compile-command: "gcc -DMODULE -D__KERNEL__ -I/usr/src/linux/net/inet -Wall -Wstrict-prototypes -O6 -c eepro100.c `[ -f /usr/include/linux/modversions.h ] && echo -DMODVERSIONS`"
 *  SMP-compile-command: "gcc -D__SMP__ -DMODULE -D__KERNEL__ -I/usr/src/linux/net/inet -Wall -Wstrict-prototypes -O6 -c eepro100.c `[ -f /usr/include/linux/modversions.h ] && echo -DMODVERSIONS`"
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 4
 * End:
 */
